# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# This Source Code Form is "Incompatible With Secondary Licenses", as
# defined by the Mozilla Public License, v. 2.0.

package Bugzilla::Auth::Verify::RedHat;

use 5.10.1;
use strict;
use warnings;

use parent qw(Bugzilla::Auth::Verify);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util;
use Bugzilla::Constants;
use Bugzilla::Token;
use Bugzilla::Util;
use Bugzilla::User;
use Authen::Radius;

use HTTP::Request;
use LWP::UserAgent;
use List::Util qw(any);
use Data::Dumper;
use JSON;

# Bug 1262651 - Radius authentication for admin users.
#
# The purpose of this is to force users with administrative permissions to log
# in using PrivacyIdea (pin+otp) authentication.

sub check_credentials {
  my ($self, $params) = @_;

  my $username = $params->{username};
  my $user     = Bugzilla::User->new({name => $username});

  return {failure => AUTH_NO_SUCH_USER} unless $user;

  # List of groups that require RADIUS auth.
  my $group_str          = Bugzilla->params->{'rh_radius_groups'} // '';
  my @groups             = split(/,/, $group_str);
  my $in_sensitive_group = any { $user->in_group($_) } @groups;

  if ($user->is_bot()) {
    return ({
      failure => AUTH_ERROR, user_error => 'no_login_for_bots', details => {},
    });
  }

  ## REDHAT EXTENSION START 2072826
  my $res;

  if ($user->can_use_password()) {
    $res = $self->check_credentials_password($params);

    # if the password matches but it's not good enough, bail
    if ( $res
      && $res->{failure}
      && $res->{user_error}
      && $res->{user_error} ne 'passwords_dont_match')
    {
      return $res;
    }
  }

  # These users have to use PrivacyIdea, but they supplied a password
  if ($res && !$res->{failure} && $in_sensitive_group) {
    return {failure => AUTH_RH_RADIUS_LOGINFAILED};
  }

  if ((!$res || $res->{failure}) && $user->can_use_privacyidea()) {
    $res = $self->check_credentials_privacyidea($params, $user);
  }

  # Can't use password or PrivacyIdea, this is run after SSO, so use that.
  if (!$res) {
    $res = {
      failure    => AUTH_ERROR,
      user_error => 'no_password_user',
      details    => {no_back => 1}
    };
  }

  ## REDHAT EXTENSION END 2072826

  return $res;
}

sub check_credentials_privacyidea {
  my ($self, $params, $user) = @_;
  my $api_url        = Bugzilla->params->{'PrivacyIDEA_API_URL'};
  my $address_suffix = Bugzilla->params->{'RADIUS_email_suffix'};
  my $username       = $params->{username};

  Bugzilla->logger->debug("Logging in using PrivacyIDEA");

  my $uid = $user->get_uid();
  Bugzilla->logger->debug("check_credentials_privacyidea uid: $uid");

  my %request_data = (user => $uid, pass => $params->{password},);

  my $ua = LWP::UserAgent->new();
  $ua->agent("RedHatBugzilla/" . BUGZILLA_VERSION);
  my $request = HTTP::Request::Common::POST($api_url, [%request_data]);
  my $resp    = $ua->request($request);

  if (!$resp->is_success) {
    Bugzilla->logger->info("PrivacyIDEA Auth failure: non-success response");
    return {failure => AUTH_RH_RADIUS_LOGINFAILED};
  }
  my $message = from_json($resp->decoded_content);

  # This returns whether the request was succesful
  if ($message->{result}->{status} != 1) {
    Bugzilla->logger->info("PrivacyIDEA Auth failure: non-success status value");
    return {failure => AUTH_RH_RADIUS_LOGINFAILED};
  }

  # This returns whether the actual token was valid
  if ($message->{result}->{value} != 1) {
    Bugzilla->logger->info("PrivacyIDEA Auth failure: token invalid (or reused)");
    return {failure => AUTH_RH_RADIUS_LOGINFAILED};
  }

  $params->{bz_username} = $username;

  return $params;
}

sub check_credentials_password {
  my ($self, $login_data) = @_;
  my $dbh = Bugzilla->dbh;

  my $username = $login_data->{username};
  my $user     = new Bugzilla::User({name => $username});

  return {failure => AUTH_NO_SUCH_USER} unless $user;

  $login_data->{user}        = $user;
  $login_data->{bz_username} = $user->login;

  if ($user->account_is_locked_out) {
    return {failure => AUTH_LOCKOUT, user => $user};
  }

  my $password              = $login_data->{password};
  my $real_password_crypted = $user->cryptpassword;

  # Using the internal crypted password as the salt,
  # crypt the password the user entered.
  my $entered_password_crypted = bz_crypt($password, $real_password_crypted);

  if ($entered_password_crypted ne $real_password_crypted) {

    # Record the login failure
    $user->note_login_failure();

    # Immediately check if we are locked out
    if ($user->account_is_locked_out) {
      my $ip = remote_ip();
      trick_taint($ip);

      my $count = Bugzilla->dbh->selectrow_array(
        "SELECT count(*)
           FROM login_failure
          WHERE user_id = ?
            AND ip_addr = ?", undef, $user->id, remote_ip()
      );

      return {
        failure         => AUTH_LOCKOUT,
        user            => $user,
        just_locked_out => $count == MAX_LOGIN_ATTEMPTS,
      };
    }

    return {
      failure       => AUTH_LOGINFAILED,
      failure_count => scalar(@{$user->account_ip_login_failures}),
    };
  }

  # Force the user to change their password if it does not meet the current
  # criteria. This should usually only happen if the criteria has changed.
  if (Bugzilla->usage_mode == USAGE_MODE_BROWSER
    && Bugzilla->params->{password_check_on_login})
  {
    my $check = validate_password_check($password);
    if ($check) {
      return {
        failure    => AUTH_ERROR,
        user_error => $check,
        details    => {locked_user => $user}
      };
    }
  }

  # The user's credentials are okay, so delete any outstanding
  # password tokens or login failures they may have generated.
  Bugzilla::Token::DeletePasswordTokens($user->id, "user_logged_in");
  if ($user->account_is_locked_out) {
    Bugzilla->logger->error("Successful login for locked account " . $user->login);
  }
  $user->clear_login_failures();

  my $update_password = 0;

  # If their old password was using crypt() or some different hash
  # than we're using now, convert the stored password to using
  # whatever hashing system we're using now.
  my $current_algorithm = PASSWORD_DIGEST_ALGORITHM;
  $update_password = 1 if ($real_password_crypted !~ /{\Q$current_algorithm\E}$/);

  # If their old password was using a different length salt than what
  # we're using now, update the password to use the new salt length.
  if ($real_password_crypted =~ /^([^,]+),/) {
    $update_password = 1 if (length($1) != PASSWORD_SALT_LENGTH);
  }

  # If needed, update the user's password.
  if ($update_password) {

    # We can't call $user->set_password because we don't want the password
    # complexity rules to apply here.
    $user->{cryptpassword} = bz_crypt($password);
    $user->update();
  }

  return $login_data;
}

sub change_password {
  my ($self, $user, $password) = @_;
  my $dbh           = Bugzilla->dbh;
  my $cryptpassword = bz_crypt($password);
  $dbh->do("UPDATE profiles SET cryptpassword = ? WHERE userid = ?",
    undef, $cryptpassword, $user->id);
  Bugzilla->memcached->clear({table => 'profiles', id => $user->id});
}

1;
