#!/usr/bin/perl -T
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

use strict;
use warnings;
use 5.10.1;
use lib qw(. lib);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Release;
use Bugzilla::Token;

my $cgi      = Bugzilla->cgi;
my $dbh      = Bugzilla->dbh;
my $template = Bugzilla->template;
my $vars     = {};

# There is only one section about release in the documentation,
# so all actions point to the same page.
$vars->{'doc_section'} = 'release.html';

#
# Preliminary checks:
#

my $user = Bugzilla->login(LOGIN_REQUIRED);

print $cgi->header();

$user->in_group('editcomponents')
  || scalar(@{$user->get_products_by_permission('editcomponents')})
  || ThrowUserError("auth_failure",
  {group => "editcomponents", action => "edit", object => "release"});

#
# often used variables
#
my $product_name = trim($cgi->param('product') || '');
my $release_name = trim($cgi->param('release') || '');
my $sortkey      = trim($cgi->param('sortkey') || 0);
my $action       = trim($cgi->param('action')  || '');
my $showbugcounts = (defined $cgi->param('showbugcounts'));
my $token         = $cgi->param('token');
my $isactive      = $cgi->param('isactive');

#
# product = '' -> Show nice list of products
#

unless ($product_name) {
  my $selectable_products = $user->get_selectable_products;

  # If the user has editcomponents privs for some products only,
  # we have to restrict the list of products to display.
  unless ($user->in_group('editcomponents')) {
    $selectable_products = $user->get_products_by_permission('editcomponents');
  }
  $vars->{'products'}      = $selectable_products;
  $vars->{'showbugcounts'} = $showbugcounts;

  $template->process("admin/releases/select-product.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

my $product = $user->check_can_admin_product($product_name);

#
# action='' -> Show nice list of releases
#

unless ($action) {

  $vars->{'showbugcounts'} = $showbugcounts;
  $vars->{'product'}       = $product;
  $template->process("admin/releases/list.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

#
# action='add' -> present form for parameters for new release
#
# (next action will be 'new')
#

if ($action eq 'add') {
  $vars->{'token'}   = issue_session_token('add_release');
  $vars->{'product'} = $product;
  $template->process("admin/releases/create.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

#
# action='new' -> add release entered in the 'action=add' screen
#

if ($action eq 'new') {
  check_token_data($token, 'add_release');
  my $release
    = Bugzilla::Release->create({
    value => $release_name, product => $product, sortkey => $sortkey
    });
  delete_token($token);

  $vars->{'message'} = 'release_created';
  $vars->{'release'} = $release;
  $vars->{'product'} = $product;
  $template->process("admin/releases/list.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

#
# action='del' -> ask if user really wants to delete
#
# (next action would be 'delete')
#

if ($action eq 'del') {
  my $release
    = Bugzilla::Release->check({product => $product, name => $release_name});

  $vars->{'release'} = $release;
  $vars->{'product'} = $product;

  # The default release cannot be deleted.
  if ($product->default_release eq $release->name) {
    ThrowUserError("release_is_default", {release => $release});
  }
  $vars->{'token'} = issue_session_token('delete_release');

  $template->process("admin/releases/confirm-delete.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

#
# action='delete' -> really delete the release
#

if ($action eq 'delete') {
  check_token_data($token, 'delete_release');
  my $release
    = Bugzilla::Release->check({product => $product, name => $release_name});
  $release->remove_from_db;
  delete_token($token);

  $vars->{'message'}              = 'release_deleted';
  $vars->{'release'}              = $release;
  $vars->{'product'}              = $product;
  $vars->{'no_edit_release_link'} = 1;

  $template->process("admin/releases/list.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

#
# action='edit' -> present the edit release form
#
# (next action would be 'update')
#

if ($action eq 'edit') {

  my $release
    = Bugzilla::Release->check({product => $product, name => $release_name});

  $vars->{'release'} = $release;
  $vars->{'product'} = $product;
  $vars->{'token'}   = issue_session_token('edit_release');

  $template->process("admin/releases/edit.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

#
# action='update' -> update the release
#

if ($action eq 'update') {
  check_token_data($token, 'edit_release');
  my $release_old_name = trim($cgi->param('releaseold') || '');
  my $release
    = Bugzilla::Release->check({product => $product, name => $release_old_name});

  $release->set_name($release_name);
  $release->set_sortkey($sortkey);
  $release->set_is_active($isactive);
  my $changes = $release->update();

  delete_token($token);

  $vars->{'message'} = 'release_updated';
  $vars->{'release'} = $release;
  $vars->{'product'} = $product;
  $vars->{'changes'} = $changes;
  $template->process("admin/releases/list.html.tmpl", $vars)
    || ThrowTemplateError($template->error());
  exit;
}

# No valid action found
ThrowUserError('unknown_action', {action => $action});
