# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2012 Jolla Ltd.
# Contact: Pami Ketolainen <pami.ketolainen@jollamobile.com>

=head1 NAME

Bugzilla::Extension::AgileTools::WebService::Team - Team manipulation WS methods

=head1 DESCRIPTION

Web service methods available under namespace 'Agile.Team'.

=cut

use strict;
use warnings;
use 5.10.1;

package Bugzilla::Extension::AgileTools::WebService::Team;

use parent qw(Bugzilla::WebService);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(trick_taint);

use Bugzilla::Extension::AgileTools::Team;

use Bugzilla::Extension::AgileTools::Util;
use Bugzilla::Extension::AgileTools::WebService::Util;


# Use the _bug_to_hash method from Bugzilla::WebService::Bug
use Bugzilla::WebService::Bug;

BEGIN {
  *_bug_to_hash = \&Bugzilla::WebService::Bug::_bug_to_hash;
  if (Bugzilla::WebService::Bug->can('_flag_to_hash')) {
    *_flag_to_hash = \&Bugzilla::WebService::Bug::_flag_to_hash;
  }
  if (Bugzilla::WebService::Bug->can('_user_to_hash')) {
    *_user_to_hash = \&Bugzilla::WebService::Bug::_user_to_hash;
  }
  if (Bugzilla::WebService::Bug->can('_translate_comment')) {
    *_translate_comment = \&Bugzilla::WebService::Bug::_translate_comment;
  }
  if (Bugzilla::WebService::Bug->can('_attachment_to_hash')) {
    *_attachment_to_hash = \&Bugzilla::WebService::Bug::_attachment_to_hash;
  }
}

# Webservice field type mapping
use constant FIELD_TYPES =>
  {id => "int", name => "string", group_id => "int", process_id => "int",};

use constant PUBLIC_METHODS => qw(
  update
  add_member
  remove_member
  add_member_role
  remove_member_role
  unprioritized_items
  create
  add_view_group
  remove_view_group
  add_edit_group
  remove_edit_group
  get
  edit_member_note
  get_visible
  get_editable
);

=head1 METHODS

=over

=item C<update>

    Description: Updates team details
    Params:      id - Team id
                 name - (optional) Change team name
    Returns:     Changes hash like from L<Bugzilla::Object::update>

=cut

sub update {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.update', param => 'id'})
    unless defined $params->{id};

  my $team = get_team(delete $params->{id}, 1);
  $team->set_all($params);
  my $changes = $team->update();
  return changes_to_hash($self, $changes, FIELD_TYPES);
}

=item C<add_member>

    Description: Add a new team member
    Params:      id - Team ID
                 user - User login name or id
                 note - A note about the user
    Returns:     The new list of team members

=cut

sub add_member {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);

  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_member', param => 'id'})
    unless defined $params->{id};

  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_member', param => 'user'})
    unless defined $params->{user};

  my $team = get_team($params->{id}, 1);
  $team->add_member($params->{user});

  if ($params->{note}) {
    my $note = $params->{note};
    trick_taint($note);
    $team->edit_member_note($params->{user}, $note);
  }

  # TODO: WebService type the result
  return $team->members;
}

=item C<remove_member>

    Description: Remove a team member
    Params:      id - Team ID
                 user - User login name or id
    Returns:     The new list of team members

=cut

sub remove_member {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_member', param => 'id'})
    unless defined $params->{id};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_member', param => 'user'})
    unless defined $params->{user};

  my $team = get_team($params->{id}, 1);
  $team->remove_member($params->{user});

  # TODO: WebService type the result
  return $team->members;
}

=item C<add_member_role>

    Description: Add new role to a team member
    Params:      id - Team id
                 user - Team member user id or login name
                 role - New role id or name
                 note - Update the team note for the user
    Returns:     The new role or empty object if user already had that role

=cut

sub add_member_role {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_member_role', param => 'id'})
    unless defined $params->{id};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_member_role', param => 'user'})
    unless defined $params->{user};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_member_role', param => 'role'})
    unless defined $params->{role};
  my $team = get_team($params->{id}, 1);
  my $role = get_role($params->{role});
  my $user = get_user($params->{user});

  if ($params->{note}) {
    my $note = $params->{note};
    trick_taint($note);
    $team->edit_member_note($params->{user}, $note);
  }

  # TODO: WebService type the result
  if ($role->add_user_role($team, $user)) {
    return {userid => $user->id, role => $role};
  }

  return {};
}

=item C<remove_member_role>

    Description: Remove role from a team member
    Params:      id - Team id
                 user - Team member user id or login name
                 role - New role id or name
                 note - Update the team note for the user
    Returns:     The role that was removed or empty object if user did not have
                 that role

=cut

sub remove_member_role {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_member_role', param => 'id'})
    unless defined $params->{id};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_member_role', param => 'user'})
    unless defined $params->{user};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_member_role', param => 'role'})
    unless defined $params->{role};
  my $team = get_team($params->{id}, 1);
  my $role = get_role($params->{role});
  my $user = get_user($params->{user});

  if ($params->{note}) {
    my $note = $params->{note};
    trick_taint($note);
    $team->edit_member_note($params->{user}, $note);
  }

  # TODO: WebService type the result
  if ($role->remove_user_role($team, $user)) {
    return {userid => $user->id, role => $role};
  }

  return {};
}

=item C<unprioritized_items>

    Description: Get unprioritized bugs in teams responsibilites
    Params:      id - team id
                 include - Resposibilities to include
                           { type: [ IDs ], }
    Returns:     { bugs: [list of bugs,... ] }

=cut

sub unprioritized_items {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.unprioritized_items', param => 'id'})
    unless defined $params->{id};

  my $team = get_team($params->{id});
  my @bugs;
  foreach my $bug (@{$team->unprioritized_items($params->{include})}) {
    my $bug_hash = $self->_bug_to_hash($bug, $params);
    push(@bugs, $bug_hash);
  }

  return {bugs => \@bugs};
}

## RED HAT EXTENSION START 1466075

=item C<create>

    Description: Create new team
    Params:      name                 - Name to call team
                 process_id           - (optional) Change the uses, currently there is only scrum. Defaults to 1 (scrum).
                 is_active            - (optional) Change active state. Defaults to 1, active.
                 product              - (optional) Limit the Unprioritized items to unpooled new and assigned bugs in the specified product.
                 create_backlog       - (optional) Create a new backlog for this team. Defaults to 1.
                 responsibility_query - (optional) Supply custom query to limit Unprioritized items. Defaults to "?f1=bug_agile_pool.pool_id&v1=-1&o1=equals&resolution=---"
    Returns:     Created team object

=cut

sub create {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);

  ThrowUserError("agile_team_manage_denied") unless user_in_agiletools_group();

  ThrowCodeError('param_required',
    {function => 'Agile.Team.create', param => 'name'})
    unless defined $params->{name};

  $params->{process_id}     //= 1;
  $params->{is_active}      //= 1;
  $params->{create_backlog} //= 1;

  my $create_backlog = delete $params->{create_backlog};
  if ($params->{product}) {
    my $prod = delete $params->{product};
    $params->{responsibility_query}
      = qq{?bug_status=NEW&bug_status=ASSIGNED&product=$prod&query_format=advanced&f1=bug_agile_pool.pool_id&v1=-1&o1=equals};
  }

  my $team = Bugzilla::Extension::AgileTools::Team->create($params);

  if ($create_backlog) {
    Bugzilla::Extension::AgileTools::Backlog->create({
      name => $team->name . " backlog", team_id => $team->id,
    });
  }

  return object_to_hash($self, $team, FIELD_TYPES);
}
## RED HAT EXTENSION END 1466075

## RED HAT EXTENSION START 1323078

=item C<add_view_group>

    Description: Add a view group to a team
    Params:      id - Team id
                 group - Group id or name
    Returns:     Nothing

=cut

sub add_view_group {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_view_group', param => 'id'})
    unless defined $params->{id};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_view_group', param => 'group'})
    unless defined $params->{group};

  my $team = get_team($params->{id}, 1);
  my $user = get_group($params->{group});

  $team->add_view_group($params->{group});

  return;
}

=item C<remove_view_group>

    Description: Remove a view group from a team
    Params:      id - Team id
                 group - Group id or name
    Returns:     Nothing

=cut

sub remove_view_group {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_view_group', param => 'id'})
    unless defined $params->{id};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_view_group', param => 'group'})
    unless defined $params->{group};

  my $team = get_team($params->{id}, 1);
  my $user = get_group($params->{group});

  $team->remove_view_group($params->{group});

  return;
}

=item C<add_edit_group>

    Description: Add an edit group to a team
    Params:      id - Team id
                 group - Group id or name
    Returns:     Nothing

=cut

sub add_edit_group {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_edit_group', param => 'id'})
    unless defined $params->{id};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.add_edit_group', param => 'group'})
    unless defined $params->{group};

  my $team = get_team($params->{id}, 1);
  my $user = get_group($params->{group});

  $team->add_edit_group($params->{group});

  return;
}

=item C<remove_edit_group>

    Description: Remove an edit group from a team
    Params:      id - Team id
                 group - Group id or name
    Returns:     Nothing

=cut

sub remove_edit_group {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_edit_group', param => 'id'})
    unless defined $params->{id};
  ThrowCodeError('param_required',
    {function => 'Agile.Team.remove_edit_group', param => 'group'})
    unless defined $params->{group};

  my $team = get_team($params->{id}, 1);
  my $user = get_group($params->{group});

  $team->remove_edit_group($params->{group});

  return;
}
## RED HAT EXTENSION END 1323078

## RED HAT EXTENSION START 1211439

=item C<get>

    Description: Get a team
    Params:      team - Team id or name
    Returns:     Team id, name, members, backlogs

=cut

sub get {
  my ($self, $params) = @_;
  Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);

  ThrowCodeError('param_required',
    {function => 'Agile.Team.get', param => 'team'})
    unless defined $params->{team};

  my $team = get_team($params->{team});

  my $tobj = object_to_hash($self, $team, FIELD_TYPES);
  ## RED HAT EXTENSION START 1613309
  $tobj->{members}  = $team->members();
  $tobj->{backlogs} = $team->backlogs();
  $team->roles();
  ## RED HAT EXTENSION END 1613309

  return $tobj;
}

## RED HAT EXTENSION END 1211439

## RED HAT EXTENSION START 2061215

=item C<get_visible>

    Description: Get all teams a user can view
    Params:      None
    Returns:     An array of hashes with id and name

=cut

sub get_visible {
  my ($self, $params) = @_;
  my $user = Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);

  my @teams;

  foreach my $team (@{$user->visible_agile_teams}) {
    push(@teams, {id => $team->id, name => $team->name});
  }

  return {teams => \@teams};
}

=item C<get_editable>

    Description: Get all teams the user can edit
    Params:      None
    Returns:     An array of hashes with id and name

=cut

sub get_editable {
  my ($self, $params) = @_;
  my $user = Bugzilla->login(LOGIN_REQUIRED);
  user_in_agiletools_group(1);

  my @teams;

  foreach my $team (@{$user->editable_agile_teams}) {
    push(@teams, {id => $team->id, name => $team->name});
  }

  return {teams => \@teams};
}
## RED HAT EXTENSION END 2061215

## REDHAT EXTENSION START 1828943

=item C<edit_member_note>

    Description: Add new role to a team member
    Params:      id - Team id
                 user - Team member user id or login name
                 role -> New role id or name
    Returns:     The new role or empty object if user already had that role

=cut

sub edit_member_note {
  my ($self, $params) = @_;

  Bugzilla->login(LOGIN_REQUIRED);

  user_in_agiletools_group(1);

  ThrowCodeError('param_required',
    {function => 'Agile.Team.edit_member_note', param => 'id'})
    unless defined $params->{id};

  ThrowCodeError('param_required',
    {function => 'Agile.Team.edit_member_note', param => 'user'})
    unless defined $params->{user};

  ThrowCodeError('param_required',
    {function => 'Agile.Team.edit_member_note', param => 'note'})
    unless defined $params->{note};

  my $team         = get_team($params->{id}, 1);
  my $user_to_edit = get_user($params->{user});
  my $note         = $params->{note};
  trick_taint($note);

  $team->edit_member_note($user_to_edit, $note);

  return {};
}
## REDHAT EXTENSION END 1828943

1;

__END__

=back

=head1 SEE ALSO

L<Bugzilla::WebService>

