# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::AuthJWT::Login;
use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::Auth::Login);

use Bugzilla;
use Bugzilla::Constants
  qw(AUTH_NODATA AUTH_ERROR USAGE_MODE_BROWSER AUTH_NO_SUCH_USER AUTH_LOCKOUT AUTH_LOGINFAILED);
use Bugzilla::Util;

use Bugzilla::Error;
use Bugzilla::Token;

use Crypt::JWT qw(decode_jwt);
use Bugzilla::Extension::AuthJWT::Util qw(sources_kids user_can_use filter_mail);
use Data::Dumper;
use File::Basename;

use constant {
  requires_verification   => 0,
  user_can_create_account => 0,
  extern_id_used          => 0,
  can_change_email        => 0,
  requires_persistence    => 0
};

# This method is only available to web services. A JWT key can never
# be used to authenticate a Web request.
sub get_login_info {
  my ($self) = @_;
  my $params = Bugzilla->input_params;
  my $cgi    = Bugzilla->cgi;
  my $is_internal   = $cgi->http('X-Bugzilla-Internal');

  my ($jwt);

  if (!i_am_webservice() || $is_internal) {
    return {failure => AUTH_NODATA};
  }

  # Only allow JWTs on JSONRPC
  unless (basename($0) eq 'jsonrpc.cgi') {
    return {failure => AUTH_NODATA};
  }

  if ($ENV{MOD_PERL}) {

    # use require as we may not be running in mod_perl mode.
    require Apache2::RequestUtil;
    require Apache2::Connection;

    my $req = Apache2::RequestUtil->request;
    $jwt = $req->headers_in->{'Authorization'};
  }
  else {
    # You need to expose this header in Apache
    # RewriteRule ^jsonrpc.cgi - [E=HTTP_AUTHORIZATION:%{HTTP:Authorization}]
    $jwt = $ENV{HTTP_AUTHORIZATION};
  }

  my $origin = $cgi->http('Origin');

  # If we aren't using a JWT, or there is no origin, bail out
  if (!$jwt || !$origin || $origin eq '') {
    return {failure => AUTH_NODATA};
  }

  # If we are an XHR and there is no JWT explode
  if ("$origin/" ne Bugzilla->params->{'urlbase'} && !$jwt) {
    return ({
      failure    => AUTH_ERROR,
      user_error => 'authjwt_invalid_xhr',
      details    => {origin => $origin}
    });
  }

  $jwt =~ s/^Bearer\s*//;

  my $kid_keys = sources_kids();

  my ($header, $token);
  eval {
    ($header, $token)
      = decode_jwt(token => $jwt, kid_keys => $kid_keys, decode_header => 1);
  };

  if ($@ || !$header->{kid}) {
    return ({
      failure    => AUTH_ERROR,
      user_error => 'authjwt_invalid_jwt',
      details    => {msg => $@ // 'No JWT KID header detected'}
    });
  }

## BUGBUG some JWTs have a valid-for origin, should we validate $origin against that?

  my $field = 'email';
  my $val   = $token->{$field};
  unless ($val && $val ne '') {
    return ({
      failure    => AUTH_ERROR,
      user_error => 'authjwt_missing_mapfield',
      details    => {field => $field}
    });
  }

  $val = filter_mail($header->{kid}, $val);

  my $user;
  $user = Bugzilla::User->new({name => $val});
  unless ($user) {
    return {
      failure    => AUTH_ERROR,
      user_error => 'authjwt_no_such_user',
      details    => {email => $val}
    };
  }

  if ($user->account_is_locked_out) {
    return {failure => AUTH_LOCKOUT, user => $user};
  }

  # enforce group restrictions
  unless (user_can_use($header->{kid}, $user)) {
    return {failure => AUTH_ERROR, user_error => 'authjwt_user_not_allowed'};
  }

  return {user_id => $user->id};
}

1;

