# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2012 Jolla Ltd.
# Contact: Pami Ketolainen <pami.ketolainen@jollamobile.com>
#
# The Initial Developer of the Original Code is Nokia. Portions created by the
# Initial Developer are Copyright (C) 2012 Nokia. All Rights Reserved.
#
# Contributor(s):
#   David Wilson <ext-david.3.wilson@nokia.com>

package Bugzilla::Extension::BayotBase;

use strict;
use warnings;
use 5.10.1;
use JSON;
use base qw(Bugzilla::Extension);

use Bugzilla::Hook;
use Bugzilla::Install::Filesystem;
use Bugzilla::Extension::BayotBase::Util;

use Digest::MD5 qw(md5_hex);

our $VERSION = '0.01';


# Invoke the bb_common_links hook, aggregating returned menu items into a
# bb_common_links variable, for hook/global/common-links-link-row.html.tmpl
sub _build_common_links {
  my ($args, $vars, $file) = @_;

  my %links;
  Bugzilla::Hook::process('bb_common_links',
    {args => $args, vars => $vars, file => $file, links => \%links});

  my @items;
  while (my ($section, $section_items) = each(%links)) {
    push @items, @$section_items;
  }

  $vars->{bb_common_links} = [
    sort {
           ($a->{priority} || 999) <=> ($b->{priority} || 999)
        or ($a->{text} cmp $b->{text})
    } @items
  ];

  return;
}

# Invoke bb_group_params and add group identifiers on the admin group listing
# page
sub _group_identifiers {
  my $vars         = shift;
  my $group_params = [];
  Bugzilla::Hook::process('bb_group_params', {group_params => $group_params});
  $vars->{overrides}->{action}->{name} ||= {};
  my $overrides = $vars->{overrides}->{action}->{name};
  for my $param (@$group_params) {
    my $groups = Bugzilla->params->{$param};
    next unless $groups;
    $groups = [$groups] if (ref $groups ne 'ARRAY');
    for my $group (@$groups) {
      $overrides->{$group} ||= {override_content => 1, content => ''};
      $overrides->{$group}->{content} .= ', ' if $overrides->{$group}->{content};
      $overrides->{$group}->{content} .= $param;
    }
  }

  return;
}

# Create a JSON object containing various useful Bugzilla runtime information.
sub _make_bb_config {
  my ($args, $vars, $file) = @_;

  my $user   = Bugzilla->user;
  my $config = {};
  if ($user->id) {
    $config->{user} = {
      logged_in => JSON::true,
      login     => $user->login,
      real_name => $user->name,
      email     => $user->email,
      id        => $user->id,
## REDHAT EXTENSION START 1298042
      groups => [sort map { $_->{name} } @{$user->groups}],

#            enterable_products => [
#                sort map { $_->{name} } @{$user->get_enterable_products}
#            ],
      enterable_products => [],
## REDHAT EXTENSION END 1298042
    };
  }
  else {
    $config->{user}
      = {logged_in => JSON::false, groups => [], enterable_products => [],};
  }

## REDHAT EXTENSION START 1145408
  my $dt   = DateTime->now();
  my $tz   = $user->timezone;
  $config->{user}->{tz}   = $tz->name;
  $config->{user}->{tzsn} = $tz->short_name_for_datetime($dt);
  $config->{user}->{tzos} = $tz->offset_for_datetime($dt);
## REDHAT EXTENSION END 1145408

  $config->{defaults} = {
    bugentry_fields => [split(/\s/, Bugzilla->params->{'bb_bug_entry_fields'})],
    priority        => Bugzilla->params->{'defaultpriority'},
    severity        => Bugzilla->params->{'defaultseverity'},
    platform        => Bugzilla->params->{'defaultplatform'},
    op_sys          => Bugzilla->params->{'defaultopsys'},
  };

  $vars->{bb_config} = _detag(JSON->new->encode($config));

  return;
}

sub _detag {
  my $var = shift;
  $var =~ s/&/&amp;/g;
  $var =~ s/</&lt;/g;
  $var =~ s/>/&gt;/g;
  $var =~ s/(?![\t\r\n])[[:cntrl:]]//g;
  $var =~ tr/\x{202a}-\x{202e}//d;
  return $var;
}

sub install_filesystem {
  my ($self, $args) = @_;
  my $cache = cache_base_dir();
  $args->{create_dirs}->{$cache}  = Bugzilla::Install::Filesystem::DIR_CGI_WRITE;
  $args->{recurse_dirs}->{$cache} = {
    dirs  => Bugzilla::Install::Filesystem::DIR_CGI_WRITE,
    files => Bugzilla::Install::Filesystem::CGI_WRITE,
  };

  return;
}

sub _validate_fielddefs_cache {
  my $obj = shift;
  if ( $obj->isa('Bugzilla::Field')
    || $obj->isa('Bugzilla::Field::ChoiceInterface')
    || $obj->isa('Bugzilla::Keyword')
    || $obj->isa('Bugzilla::Version')
    || $obj->isa('Bugzilla::Milestone')
    || $obj->isa('Bugzilla::Product')
    || $obj->isa('Bugzilla::Component')
    || $obj->isa('Bugzilla::Group'))
  {
    ## REDHAT EXTENSION START 1999932
    my $dbh = Bugzilla->dbh;
    my $sql = q{UPDATE params SET value = NOW() WHERE name = 'BayotBase_cache_ts'};
    $dbh->do($sql);
    ## REDHAT EXTENSION END 1999932
  }

  return;
}

sub object_end_of_create {
  my ($self, $args) = @_;
  _validate_fielddefs_cache($args->{object});

  return;
}

sub object_end_of_update {
  my ($self, $args) = @_;
  _validate_fielddefs_cache($args->{object});

  return;
}

sub page_before_template {
  my ($self, $args) = @_;
  if ( $args->{page_id} eq 'bayotbase/fielddefs.js'
    || $args->{page_id} eq 'bayotbase/publicfielddefs.js')
  {
    ## Allow using 2 paths for 1 template
    if ($args->{page_id} eq 'bayotbase/publicfielddefs.js') {
      my $page_name = $args->{page_name};
      $$page_name = 'bayotbase/fielddefs';
    }

    print Bugzilla->cgi->header(-expires => '+1d',
      -Content_Type => 'text/javascript');
## REDHAT EXTENSION START 1775514
    my ($fields, $relations, $all_classifications, $teams, $field_order,
      $advanced_fields, $selectable_products)
      = get_field_defs(as_json => 1);
    $args->{vars}->{fielddefs_json}      = $fields;
    $args->{vars}->{relations}           = $relations;
    $args->{vars}->{all_classifications} = $all_classifications;
    $args->{vars}->{teams}               = $teams;
    $args->{vars}->{field_order}         = $field_order;
    $args->{vars}->{advanced_fields}     = $advanced_fields;
    $args->{vars}->{selectable_products} = $selectable_products;
    $args->{vars}->{field_map}           = Bugzilla->INT2EXT_FIELD_MAP;
    $args->{vars}->{ext_field_map}       = Bugzilla->EXT2INT_FIELD_MAP;
## REDHAT EXTENSION END 1775514
  }

  return;
}

sub template_before_process {
  my ($self, $args) = @_;

  my $vars = $args->{vars};
  my $file = $args->{file};
  if ($file eq 'global/common-links.html.tmpl') {
    _build_common_links($args, $vars, $file);
  }
  elsif ($file eq 'global/header.html.tmpl') {
    $vars->{bb_field_cache_ts} = cache_timestamp();
    _make_bb_config($args, $vars, $file);
  }
  elsif ($file eq 'admin/table.html.tmpl'
    && $vars->{template}->name eq 'admin/groups/list.html.tmpl')
  {
    _group_identifiers($vars);
  }

  return;
}

sub template_before_create {
  my ($self, $params) = @_;

  # TODO: document remove_query_param template filter
  $params->{config}->{FILTERS}->{remove_query_param} = [
    sub {
      my ($context, @args) = @_;
      return sub {
        my $query   = shift;
        my ($param) = @args;
        my @parts   = grep($_ !~ /^$param=/, split('&', $query));
        return join('&', @parts);
      }
    },
    1
  ];

  $params->{config}->{FILTERS}->{remove_private_query_params} = [
    sub {
      my ($context, @args) = @_;
      return sub {
        my $url = shift;
        return clean_params($url);
      }
    },
    1
  ];

  $params->{config}->{FILTERS}->{'md5_hex'} = [
    sub {
      my ($context, @args) = @_;
      return sub {
        my $data = shift;
        return md5_hex($data);
      }
    },
    1
  ];

  return;
}

sub config_add_panels {
  my ($self, $args) = @_;
  my $modules = $args->{panel_modules};
  $modules->{BayotBase} = "Bugzilla::Extension::BayotBase::Config";

  return;
}

## REDHAT EXTENSION START 1999932
sub install_update_db {
  my ($self, $args) = @_;
  my $dbh = Bugzilla->dbh;
  my $count = Bugzilla->dbh->selectrow_array(q{SELECT COUNT(*) FROM params WHERE name = 'BayotBase_cache_ts'});

  if($count == 0) {
    my $sql = q{INSERT INTO params (name, value, isactive, inui) VALUES ('BayotBase_cache_ts', NOW(), 1, 0)};
    $dbh->do($sql);
  }

  return;
}
## REDHAT EXTENSION END 1999932

__PACKAGE__->NAME;
