# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (C) 2013 Jolla Ltd.
# Contact: Pami Ketolainen <pami.ketolainen@jollamobile.com>

package Bugzilla::Extension::BayotBase::Util;
use strict;
use warnings;
use 5.10.1;

use Bugzilla::Error;
use open ':encoding(utf8)';

use base qw(Exporter);

our @EXPORT = qw(
  cache_base_dir
  cache_timestamp
  cache_user_file
  get_field_defs
  clean_params
);

use Bugzilla::Constants;
use Bugzilla::Component;
use Bugzilla::Keyword;
use Bugzilla::Milestone;
use Bugzilla::Status;
use Bugzilla::Util;

use Scalar::Util qw(blessed);
use List::MoreUtils qw(none);

use constant PRODUCT_SPECIFIC =>
  {version => 1, target_milestone => 1, component => 1, target_release => 1,};

# fake types user and bool.
use constant FAKE_TYPES => {
  user => FIELD_TYPE_HIGHEST_PLUS_ONE,
  bool => FIELD_TYPE_HIGHEST_PLUS_ONE + 1,
};

use constant FIELD_OVERRIDES => {
  alias       => {type => 1,                  is_on_bug_entry => 1,},
  assigned_to => {type => FAKE_TYPES->{user}, is_on_bug_entry => 1,},
  blocked     =>
    {name => 'blocks', type => 6, multivalue => 1, is_on_bug_entry => 1,},
  cc => {type => FAKE_TYPES->{user}, multivalue => 1, is_on_bug_entry => 1,},
  classification => {is_on_bug_entry => 0,},
  component      => {is_on_bug_entry => 1, value_field     => 'product',},
  longdesc       => {name => 'comment',    type => 4, is_on_bug_entry => 1,},
  creation_ts    => {
    name            => 'creation_time',
    immutable       => 1,
    is_on_bug_entry => 1,

    # It's a datetime, but as immutable the type doesn't matter
  },
  reporter  => {name => 'creator', type => FAKE_TYPES->{user}, immutable => 1,},
  deadline  => {is_on_bug_entry => 1,},
  dependson =>
    {name => 'depends_on', type => 6, is_on_bug_entry => 1, multivalue => 1,},
  dependson_all =>
    {name => 'depends_on_all', type => 6, is_on_bug_entry => 0, multivalue => 1,},
  estimated_time => {is_on_bug_entry => 1,},
  bug_group      => {
    name            => 'groups',
    type            => 3,
    multivalue      => 1,
    is_on_bug_entry => 1,
    value_field     => 'product',
  },
  bug_id            => {name => 'id', type => 6, immutable => 1,},
  cclist_accessible => {
    name            => 'is_cc_accessible',
    type            => FAKE_TYPES->{bool},
    is_on_bug_entry => 1,
    immutable       => 1,
  },
  reporter_accessible => {
    name            => 'is_creator_accessible',
    type            => FAKE_TYPES->{bool},
    is_on_bug_entry => 1,
    immutable       => 1,
  },
  keywords => {multivalue => 1,                  is_on_bug_entry => 1,},
  delta_ts => {name       => 'last_change_time', immutable       => 1,},
  op_sys   => {
    is_on_bug_entry => 1,
    is_mandatory    => Bugzilla->params->{defaultopsys} ? 0 : 1,
  },
  rep_platform => {
    name            => 'platform',
    is_on_bug_entry => 1,
    is_mandatory    => Bugzilla->params->{defaultplatform} ? 0 : 1,
  },
  priority   => {is_on_bug_entry => 1,},
  product    => {is_on_bug_entry => 1,},
  qa_contact => {type            => FAKE_TYPES->{user}, is_on_bug_entry => 1,},
  remaining_time => {is_on_bug_entry  => 1,},
  resolution     => {visibility_field => 'status',},
  see_also       => {multivalue       => 1,},
  bug_severity   => {name             => 'severity', is_on_bug_entry => 1,},
  bug_status => {name => 'status', is_on_bug_entry => 1, value_field => 'status'},
  short_desc => {name => 'summary', is_on_bug_entry => 1,},
  target_milestone =>
    {is_on_bug_entry => 1, type => 2, value_field => 'product',},
  target_release =>
    {is_on_bug_entry => 1, type => 2, value_field => 'product',},
  bug_file_loc => {name => 'url', is_on_bug_entry => 1,},
  version      => {
    type            => 2,
    is_mandatory    => 1,
    is_on_bug_entry => 1,
    value_field     => 'product',
  },
  status_whiteboard => {name => 'whiteboard', is_on_bug_entry => 1,},
  work_time         => {},
  'flagtypes.name' => {
    is_on_bug_entry  => 1,
    visibility_field => 'product',
    type             => FAKE_TYPES->{bool},
  },
  rh_sub_components =>
    {is_on_bug_entry => 1, value_field => 'component', type => 2,},
  dependent_products => {is_on_bug_entry => 1, value_field => 'product',},
  'agile_pool.name' => {
    is_on_bug_entry => 1,
##    value_field     => 'component',
    type            => 2,
  },
  owner_idle_time => {
    immutable       => 1,
  },
  cf_last_closed => {immutable => 1,},
  "attachments.description"       => {immutable => 1,},
  "attachments.filename"          => {immutable => 1,},
  "attachments.mimetype"          => {immutable => 1,},
  "attachments.ispatch"           => {immutable => 1,},
  "view"                          => {immutable => 1,},
  "attachments.isobsolete"        => {immutable => 1,},
  "attachments.isprivate"         => {immutable => 1,},
  "days_elapsed"                  => {immutable => 1,},
  "comment"                       => {immutable => 1,},
  "everconfirmed"                 => {immutable => 1,},
  "requestees.login_name"         => {immutable => 1,},
  "setters.login_name"            => {immutable => 1,},
  "content"                       => {immutable => 1,},
  "external_bugzilla.url"         => {immutable => 1,},
  "external_bugzilla.description" => {immutable => 1,},
  "ext_bz_bug_map.ext_bz_bug_id"  => {immutable => 1,},
  "attachments.submitter"         => {immutable => 1,},
  "longdescs.isprivate"           => {immutable => 1,},
  "commenter"                     => {immutable => 1,},
  "attach_data.thedata"           => {immutable => 1,},
  "owner_idle_time"               => {immutable => 1,},
  "extra_components"              => {immutable => 1,},
  "extra_versions"                => {immutable => 1,},
  "cf_clone_of"                   => {immutable => 1,},
  "cf_last_closed"                => {immutable => 1,},
  "longdescs.count"               => {immutable => 1,},
  "assigned_to_realname"          => {immutable => 1,},
  "reporter_realname"             => {immutable => 1,},
  "qa_contact_realname"           => {immutable => 1,},
  "docs_contact_realname"         => {immutable => 1,},
  "rh_rule"                       => {immutable => 1,},
  "ext_bz_bug_map.ext_status"     => {immutable => 1,},
  "comment_tag"                   => {immutable => 1,},
  "agile_pool.id"                 => {immutable => 1,},
  "bug_agile_pool.pool_order"     => {immutable => 1,},
  "agile_team.name"               => {immutable => 1,},
  "longdescs.extra_groups"        => {immutable => 1,},
  "ext_bz_bug_map.ext_priority"     => {immutable => 1,},
  "ext_bz_bug_map.ext_description"  => {immutable => 1,},
  "ext_bz_bug_map.ext_last_updated" => {immutable => 1,},
  'locked'                          => {type => FAKE_TYPES->{bool},},
};

use constant EXTRA_FIELDS => (

# Not in fields() result
  {
    id               => 0,
    name             => 'dupe_of',
    internal_name    => 'dup_id',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    type             => 6,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'is_confirmed',
    internal_name    => undef,
    display_name     => 'Confirmed',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => FAKE_TYPES->{bool},
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'is_open',
    internal_name    => undef,
    display_name     => 'Open',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => FAKE_TYPES->{bool},
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'update_token',
    internal_name    => undef,
    display_name     => 'Update Token',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 0,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'actual_time',
    internal_name    => undef,
    display_name     => 'Actual Hours',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 0,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
  },
  {
    id               => 0,
    name             => 'blocks_all',
    internal_name    => undef,
    display_name     => 'Blocks (full)',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 6,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
    multivalue       => 1,
  },
  {
    id               => 0,
    name             => 'depends_on_all',
    internal_name    => undef,
    display_name     => 'Depends On (full)',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 6,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
    multivalue       => 1,
  },
  {
    id               => 0,
    name             => 'relevance',
    internal_name    => undef,
    display_name     => 'Relevance',
    is_custom        => 0,
    is_mandatory     => 0,
    is_on_bug_entry  => 0,
    immutable        => 1,
    type             => 0,
    visibility_field => undef,
    visiblity_values => [],
    values           => [],
    buglist          => 0,
  },);

## REDHAT EXTENSION START 1775514
# Add relationships to user specific JS to allow client side field mappings

sub _add_related_value {
  my ($rel_fields, $changed_field, $cid, $affected_field, $aid) = @_;

  $rel_fields->{$changed_field} = {}
    unless (exists $rel_fields->{$changed_field});
  $rel_fields->{$changed_field}->{'values'} = {}
    unless (exists $rel_fields->{$changed_field}->{'values'});
  $rel_fields->{$changed_field}->{'values'}->{$cid} = {}
    unless (exists $rel_fields->{$changed_field}->{'values'}->{$cid});
  $rel_fields->{$changed_field}->{'values'}->{$cid}->{$affected_field} = []
    unless (
    exists $rel_fields->{$changed_field}->{'values'}->{$cid}->{$affected_field});

  push(
    @{$rel_fields->{$changed_field}->{'values'}->{$cid}->{$affected_field}},
    $aid
  );

  return;
}

sub _add_related_field {
  my ($rel_fields, $changed_field, $affected_field, $ids) = @_;

  $rel_fields->{$changed_field} = {}
    unless (exists $rel_fields->{$changed_field});
  $rel_fields->{$changed_field}->{'fields'} = {}
    unless (exists $rel_fields->{$changed_field}->{'fields'});

  $rel_fields->{$changed_field}->{'fields'}->{$affected_field} = $ids;

  return;
}
## REDHAT EXTENSION END 1775514

sub _generate_field_defs {
  my ($user, $user_values) = @_;
  $user ||= Bugzilla->user;

  my %fields;
  my %related_fields;
  my %all_classifications;
  my $field_descs = template_var('field_descs');
  my %selectable_products
    = map { $_->id => $_ } @{$user->get_selectable_products};

  foreach my $field (@{Bugzilla->fields}) {
    next unless $field->custom or defined FIELD_OVERRIDES->{$field->name};

    my @values;
    if ($field->name eq 'product') {
      ## REDHAT EXTENSION START 1775514
      foreach my $product (values %selectable_products) {
        $all_classifications{$product->name} = $product->classification->name;
      }
      ## REDHAT EXTENSION END 1775514
    }
    elsif ($field->name eq 'keywords') {
      foreach my $keyword (Bugzilla::Keyword->get_all) {
        push @values,
          {
          name              => $keyword->name,
          sort_key          => 0,
          visibility_values => [],
          is_default        => 0,
          };
      }
    }
    elsif (PRODUCT_SPECIFIC->{$field->name}) {
      my @list;
      if ($field->name eq 'version') {
        @list = Bugzilla::Version->get_all;
      }
      elsif ($field->name eq 'component') {
        ## REDHAT EXTENSION START 1298042
        #@list = Bugzilla::Component->get_all;
        ## REDHAT EXTENSION END 1298042
      }
      elsif ($field->name eq 'target_milestone') {
        @list = Bugzilla::Milestone->get_all;
      }
      elsif ($field->name eq 'target_release') {
        @list = Bugzilla::Release->get_all;
      }

      foreach my $value (@list) {
        my $product = $selectable_products{$value->product_id};
        next unless defined $product;

        ## REDHAT EXTENSION START 1317777
        next if (exists($value->{isactive}) && (!$value->{isactive}));
        ## REDHAT EXTENSION END 1317777

        my $sortkey = $value->can('sortkey') ? $value->sortkey : 0;
        push @values,
          {
          name              => $value->name,
          sort_key          => $sortkey || 0,
          visibility_values => [$product->name],
          is_default        => 0,
          };

        _add_related_value(\%related_fields, 'product', $product->name, $field->name,
          $value->name);
      }
    }
    elsif ($field->name eq 'bug_group') {
      my $list = Bugzilla::Group->match({isactive => 1, isbuggroup => 1});

      foreach my $group (@$list) {

        #my @prods = map {$_->{product}->name} @{$group->products};
        push @values, {
          name     => $group->name,
          id       => $group->id,
          sort_key => 0,
          title    => $group->description,
          category => $group->category,
          #visibility_values => \@prods,
          visibility_values => [],
          is_default        => 0,
          insider_group      => none {$_->id == $group->id} @{Bugzilla->non_insider_groups}
        };

        foreach my $product (@{$group->products}) {
          if ( $product->{controls}->{membercontrol} > 0
            || $product->{controls}->{othercontrol} > 0)
          {
            _add_related_value(\%related_fields, 'product', $product->{product}->name,
              $field->name, $group->name);
          }
        }
      }
    }
    elsif ($field->name eq 'agile_pool.name') {
      foreach my $team (@{$user->visible_agile_teams}) {
        foreach my $pool (@{$team->pools(1)}) {
          push @values, {
              name     => $pool->name,
              sort_key => 0,
              visibility_values => [],
              is_default        => 0,
            };

            #my @prods = map { $_->{product} } @{$group->products};
            #foreach my $product (@prods) {
            #  _add_related_value(\%related_fields, 'product', $product->name, $field->name,
            #    $group->name);
            #}
        }
      }
    }
    elsif ($field->name eq 'flagtypes.name') {
      my %flags;

      foreach my $product (values %selectable_products) {
        foreach my $flag (
          @{
            Bugzilla::FlagType::match(
              {product_id => $product->id, target_type => 'bug', is_active => 1}
            )
          }
          )
        {
          if ($user->can_request_flag($flag) && $flag->name !~ m/^approvals_request/) {
            if (defined $flags{$flag->id}) {
              push @{$flags{$flag->id}->{visibility_values}}, $product->name;
            }
            else {
              $flags{$flag->id} = {
                name              => $flag->name,
                sort_key          => $flag->sortkey,
                visibility_values => [$product->name],
                is_default        => 0,
                id                => $flag->id,
                is_requesteeble   => $flag->is_requesteeble,
              };
            }
          }
        }
      }

      foreach my $flag_id (
        sort {
               $flags{$a}->{sort_key} <=> $flags{$b}->{sort_key}
            || $flags{$a}->{name} cmp $flags{$b}->{name}
        } keys %flags
        )
      {
        push @values, $flags{$flag_id};
      }
    }
    elsif ($field->name eq 'bug_status') {
      my %statuses;

      # BUGBUG this limits selections to valid status changes, noice
      # we should use it ...
      foreach my $value (@{$field->legal_values}) {
        foreach (@{$value->can_change_to}) {
          $statuses{$_->name} ||= {visibility_values => []};
          push(@{$statuses{$_->name}->{visibility_values}}, $value->name);
        }
        $statuses{$value->name} ||= {visibility_values => []};
        $statuses{$value->name}->{name}       = $value->name;
        $statuses{$value->name}->{sort_key}   = $value->sortkey;
        $statuses{$value->name}->{is_default} = 0;
      }
      @values = values %statuses;
    }
    elsif ($field->name eq 'dependent_products'){
      my %processed = ();

      foreach my $product (values %selectable_products) {
        foreach my $dp (@{$product->dependent_products()}) {
          if(!defined $processed{$dp->name}){
            $processed{$dp->name} = 1;
            push @values, {
              name     => $dp->name,
              sort_key => 0,
              visibility_values => [],
              is_default        => 0,
            };
          }

          _add_related_value(\%related_fields, 'product', $product->name, $field->name,
            $dp->name);
        }      
      }
    }
    elsif ($field->is_select) {
      my @for_all;
      my %limited;

      foreach my $value (@{$field->legal_values}) {
        my $vis_val = $value->visibility_value;
        push @values,
          {
          name              => $value->name,
          sort_key          => $value->sortkey || 0,
          visibility_values => [defined $vis_val ? $vis_val->name : ()],
          is_default        => $value->is_default,
          };

        if ($vis_val) {
          _add_related_value(\%related_fields, $field->value_field->name,
            $vis_val->name, $field->name, $value->name);
          $limited{$vis_val->name} = 1;
        }
        elsif ($field->visibility_field) {
          push(@for_all, $value->name);
        }
      }

      # If a field can have values filtered then unfiltered values need to be
      # added to every visibility value that has anything set
      foreach my $val (@for_all) {
        foreach my $visval (keys %limited) {
          _add_related_value(\%related_fields, $field->value_field->name,
            $visval, $field->name, $val);
        }
      }
    }

    @values
      = sort { $a->{sort_key} <=> $b->{sort_key} || $a->{name} cmp $b->{name} }
      @values
      if (@values);

    my $visibility_field
      = $field->visibility_field ? $field->visibility_field->name : undef;
    my $value_field = $field->value_field ? $field->value_field->name : undef;
    my @visibility_values;
    if ($field->name eq 'resolution') {
      @visibility_values = map { $_->name } Bugzilla::Status::closed_bug_statuses();
      $visibility_field  = 'bug_status';
    }
    else {
      @visibility_values = map { $_->name } @{$field->visibility_values};
    }

    my $name         = FIELD_OVERRIDES->{$field->name}->{name} || $field->name;
    my $display_name = $field_descs->{$field->name} || $field->description;

    $fields{$name} = {
      name              => $name,
      internal_name     => $field->name,
      sortkey           => $field->sortkey,
      display_name      => $display_name,
      id                => scalar $field->id,
      is_custom         => $field->custom ? 1 : 0,
      values            => \@values,
      visibility_values => \@visibility_values,
      long_desc         => Bugzilla->safe_html($field->long_desc),

      # OVERRIDABLE
      type         => FIELD_OVERRIDES->{$field->name}->{type} || $field->type,
      is_mandatory => FIELD_OVERRIDES->{$field->name}->{is_mandatory}
        || $field->is_mandatory ? 1 : 0,
      is_on_bug_entry => FIELD_OVERRIDES->{$field->name}->{is_on_bug_entry}
        || $field->enter_bug ? 1 : 0,
      value_field => FIELD_OVERRIDES->{$field->name}->{value_field} || $value_field,
      visibility_field => FIELD_OVERRIDES->{$field->name}->{visibility_field}
        || $visibility_field,
      buglist => FIELD_OVERRIDES->{$field->name}->{buglist}
        || $field->buglist ? 1 : 0,

      # EXTRA
      immutable  => FIELD_OVERRIDES->{$field->name}->{immutable} || 0,
      multivalue => FIELD_OVERRIDES->{$field->name}->{multivalue}
        || $field->type == FIELD_TYPE_MULTI_SELECT,
    };

    _add_related_field(\%related_fields, $visibility_field, $field->name,
      \@visibility_values)
      if (@visibility_values);    # REDHAT EXTENSION 1775514
  }
  foreach (EXTRA_FIELDS) {
    $fields{$_->{name}} = $_;
  }

  my %teams;
  foreach my $team (@{$user->visible_agile_teams}) {
    my $new_team = {name => $team->name, id => $team->id,};

    my $pools;
    foreach my $pool (@{$team->pools(1)}) {
      my $new_pool = {
        name       => $pool->name,
        id         => $pool->id,
        is_sprint  => $pool->is_sprint,
        is_backlog => $pool->is_backlog,
      };
      $pools->{$pool->name} = $new_pool;
    }

    $new_team->{pools} = $pools;
    $teams{$team->name} = $new_team;
  }

  $user_values->{fields}              = \%fields;
  $user_values->{relations}           = \%related_fields;
  $user_values->{all_classifications} = \%all_classifications;
  $user_values->{teams}               = \%teams;

  my @order = map { FIELD_OVERRIDES->{$_->name}->{name} || $_->name } 
              grep {
                 (!FIELD_OVERRIDES->{$_->name}->{immutable}) &&
                 ($_->name !~ m/^(?:see_also)$/) &&
                 $_->obsolete == 0
              }
              sort { $a->sortkey <=> $b->sortkey || lc($a->name) cmp lc($b->name) }
              @{Bugzilla->fields};

  $user_values->{field_order} = \@order;

  my %products;

  foreach my $product (values %selectable_products) {
    $all_classifications{$product->name} = $product->classification->name;
    my @pcgs = map { $_->name } @{$product->private_comment_groups};
    $products{$product->name} = {
      name                   => $product->name,
      id                     => $product->id,
      private_comment_groups => \@pcgs,
      sort_key               => 0,
      visibility_values      => [],
      is_default             => 0,
      allows_unconfirmed     => $product->allows_unconfirmed ? 1 : 0,
    };
  }

  $user_values->{selectable_products} = \%products;

  return;    # REDHAT EXTENSION 1775514
}

sub cache_base_dir {
  return bz_locations()->{'datadir'} . '/extensions/bayotbase_cache/';
}

sub cache_user_file {
  my $uid  = shift;
  my $name = shift // '.json';    # REDHAT EXTENSION 1775514

  $uid = $uid->id if (blessed $uid);
  $uid ||= Bugzilla->user->id || 0;
  return cache_base_dir() . "$uid$name";
}

## REDHAT EXTENSION START 1999932
sub cache_timestamp {
  my ($timestamp)
    = Bugzilla->dbh->selectrow_array(
    q{SELECT extract(epoch from value::timestamp) FROM params WHERE name = 'BayotBase_cache_ts'});

  return $timestamp;
}
## REDHAT EXTENSION END 1999932

sub get_field_defs {
  my %args     = @_;
  my $cache_ts = cache_timestamp();
  my $as_json  = $args{as_json} || 0;
  my $user     = $args{user}    || Bugzilla->user;

  my $json_string     = "{}";
  my $user_cache_file = cache_user_file($user, '');
  my $user_cache_ts   = (stat($user_cache_file))[9] || 0;
  my $pretty          = Bugzilla->params->{bb_use_debug_js};
  my $json            = JSON->new->allow_blessed->allow_nonref;

  my $advanced_fields = '#' . join(', #', @{$user->advanced_fields});
  $advanced_fields = '' if ($advanced_fields eq '#');

## REDHAT EXTENSION START 1775514
  my %user_files = (
    fields              => '',
    relations           => '.rel',
    all_classifications => '.ac',
    teams               => '.teams',
    field_order         => '.fldo',
    selectable_products => '.slprods',
  );

  my %user_values;
## REDHAT EXTENSION END 1775514
  my ($fh);
  if ($cache_ts > $user_cache_ts) {
    _generate_field_defs($user, \%user_values);
    $json_string = $json->pretty($pretty)->encode($user_values{fields});

    foreach my $type (keys %user_files) {
      my $file_name = cache_user_file($user, $user_files{$type});
      open($fh, ">", $file_name) or warn "Failed to write $file_name: $!";
      if (defined $fh) {
        print $fh $json->encode($user_values{$type});
        close $fh;
      }
    }
  }
  else {
    eval {
      foreach my $type (keys %user_files) {
        my $file_name = cache_user_file($user, $user_files{$type});
        open($fh, "<", $file_name) or die "Failed to read $file_name: $!";
        if (defined $fh) {
            my $input = join('', <$fh>);
            $user_values{$type} = $json->decode($input);
            $json_string = $input if ($type eq 'fields');
            close $fh ;
        }
      }
    };

    if ($@) {
      _generate_field_defs($user, \%user_values);
      $json_string = $json->pretty($pretty)->encode($user_values{fields}) if $as_json;
    }
  }

  return (
    $as_json ? $json_string : $user_values{fields},
    $json->pretty($pretty)->encode($user_values{relations}),
    $user_values{all_classifications},
    $json->pretty($pretty)->canonical->encode($user_values{teams}),
    $json->pretty($pretty)->encode($user_values{field_order}),
    $advanced_fields,
    $json->pretty($pretty)->encode($user_values{selectable_products}),
  );
}

sub clean_params {
  my $url = shift;

  my $query = $url;
  my $start = '';

  # peel off start of URL if it's included
  if ($url =~ /^(.*\?)(.*)$/) {
    $start = $1;
    $query = $2;
  }
  my @params = qw/
    Bugzilla_api_key
    Bugzilla_api_token
    Bugzilla_login
    Bugzilla_login_token
    Bugzilla_password
    GoAheadAndLogIn
    api_key
    api_token
    password
    target
    token
    /;

  my @parts = grep {
    my $param = $_;
    none { $param =~ /^$_=/ } @params
  } split('&', $query);

  return $start . join('&', @parts);
}

1;
