# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::BotWatching;
use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::Extension);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::User;
use Bugzilla::User::Setting;

use List::MoreUtils qw(any uniq);

our $VERSION = '1';

use constant REL_BOT_WATCHER => 20;

BEGIN {
  *Bugzilla::User::watched_bots          = \&_user_watched_bots;
  *Bugzilla::User::set_bot_mail_defaults = \&_user_set_bot_mail_defaults;
}

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  $args->{'schema'}->{'bot_watching'} = {
    FIELDS => [
      id      => {TYPE => 'MEDIUMSERIAL', NOTNULL => 1, PRIMARYKEY => 1,},
      user_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE',}
      },
      bot_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid', DELETE => 'CASCADE',}
      },
      squelch          => {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 'FALSE'},
      no_minor_updates => {TYPE => 'BOOLEAN', NOTNULL => 1, DEFAULT => 'FALSE'},
    ],
  };

  return;
}

sub install_update_db {
  my $dbh = Bugzilla->dbh;

  my $sql   = 'SELECT count(*) FROM email_setting WHERE relationship = ?';
  my $count = $dbh->selectrow_array($sql, undef, REL_BOT_WATCHER);
  if ($count == 0 && lc(Bugzilla->localconfig->{db_driver}) ne 'sqlite') {
    foreach my $event (POS_EVENTS) {
      next if ($event == EVT_ADDED_REMOVED);
      $dbh->do("INSERT INTO email_setting(SELECT userid, ?, ? FROM profiles)",
        undef, REL_BOT_WATCHER, $event);
    }
  }

  return;
}

sub template_before_create {
  my ($self, $args) = @_;
  my $config    = $args->{config};
  my $constants = $config->{VARIABLES}->{constants};
  $constants->{REL_BOT_WATCHER} = REL_BOT_WATCHER;

  return;
}

sub user_preferences {
  my ($self, $args) = @_;

  my $tab = $args->{'current_tab'};
  return unless $tab eq 'bot_watching';

  my $handled = $args->{'handled'};
  my $input   = Bugzilla->input_params;
  my $save    = $args->{'save_changes'};
  my $vars    = $args->{'vars'};

  my $user = $vars->{'change_user'};

  if ($save) {
    my $dbh = Bugzilla->dbh;
    $dbh->bz_start_transaction();

    $dbh->do('DELETE FROM bot_watching WHERE user_id = ?', undef, $user->id);
    my $set
      = $dbh->prepare(
      'INSERT INTO bot_watching (user_id, bot_id, squelch, no_minor_updates) VALUES (?, ?, ?, ?)'
      );

    foreach my $bot (@{$user->watched_bots}) {
      ## checkboxes only exists if they are checked
      my $squelch = exists $input->{"squelch_" . $bot->{userid}} ? 1 : 0;
      my $no_minor_updates
        = exists $input->{"no_minor_updates_" . $bot->{userid}} ? 1 : 0;

      if ($squelch || $no_minor_updates) {
        $set->execute($user->id, $bot->{userid}, $squelch, $no_minor_updates);
      }
    }

    $dbh->bz_commit_transaction();

    # Clear this so changes get loaded
    $user->{'watched_bots'} = undef;
  }

  $vars->{'watched_bots'} = $user->watched_bots();

  $$handled = 1;

  return;
}

sub _user_watched_bots {
  my $self = shift;

  # BUGBUG sqlite doesn't like this sql :(
  return if (lc(Bugzilla->localconfig->{db_driver}) eq 'sqlite');

  if (!defined $self->{'watched_bots'}) {

    my $dbh   = Bugzilla->dbh;
    my $regex = Bugzilla->params->{bot_name_regex};

    my $query = <<'EOS';
      SELECT bu.userid, bu.login_name, bw.squelch, bw.no_minor_updates, 
EOS

    $query .= $dbh->sql_group_concat('at.name', "', '", 1) . ' team_names';

    $query .= <<'EOS';
      FROM profiles bu
      LEFT JOIN user_group_map ugm ON ugm.user_id = bu.userid
      LEFT JOIN agile_team at ON ugm.group_id = at.group_id
      LEFT JOIN bot_watching bw ON bw.bot_id = bu.userid AND bw.user_id = ?
      WHERE
      at.group_id IN (
        SELECT group_id FROM user_group_map WHERE user_id = ?
      )
      AND bu.login_name ~ ?
      GROUP BY bu.userid, bu.login_name, bw.squelch, bw.no_minor_updates
EOS

    $self->{'watched_bots'} = $dbh->selectall_arrayref($query, {'Slice' => {}},
      $self->id, $self->id, $regex);
  }

  return $self->{'watched_bots'};
}

sub _user_set_bot_mail_defaults {
  my ($user) = @_;
  my $dbh = Bugzilla->dbh;

  my $sth = $dbh->prepare("
        SELECT 1
          FROM email_setting
         WHERE user_id = ? AND relationship = ?
    ");
  $sth->execute($user->id, REL_BOT_WATCHER);
  return if $sth->fetchrow_array;

  foreach my $event (POS_EVENTS) {
    next if ($event == EVT_ADDED_REMOVED);
    $dbh->do("INSERT INTO email_setting(user_id,relationship,event) VALUES (?,?,?)",
      undef, $user->id, REL_BOT_WATCHER, $event);
  }

  return;
}

sub bugmail_recipients {
  my ($self, $args) = @_;

  my $bug        = $args->{'bug'};
  my $recipients = $args->{'recipients'};
  my $diffs      = $args->{'diffs'};

  # get user ids for
  my @user_ids = ($bug->assigned_to->id);
  push(@user_ids, $bug->qa_contact->id)   if ($bug->qa_contact);
  push(@user_ids, $bug->docs_contact->id) if ($bug->docs_contact);
  push(@user_ids, map { $_->id } @{$bug->cc_users});

  foreach my $change (@{$diffs}) {
    if ($change->{old}) {
      if ($change->{field_name} eq 'cc') {
        foreach my $cc_user (split(/[\s,]+/, $change->{old})) {
          my $uid = login_to_id($cc_user);
          push(@user_ids, $uid) if $uid;
        }
      }
      elsif ($change->{field_name} eq 'qa_contact') {
        my $uid = login_to_id($change->{old});
        push(@user_ids, $uid) if $uid;
      }
      ## REDHAT EXTENSION START 876015
      elsif ($change->{field_name} eq 'docs_contact') {
        my $uid = login_to_id($change->{old});
        push(@user_ids, $uid) if $uid;
      }
      ## REDHAT EXTENSION END 876015
      elsif ($change->{field_name} eq 'assigned_to') {
        my $uid = login_to_id($change->{old});
        push(@user_ids, $uid) if $uid;
      }
    }
  }

  my $watchers = _users_watching_bots(\@user_ids, $args->{'minor_update'});
  foreach my $usr (@$watchers) {
    if (!exists $recipients->{$usr->{userid}}) {
      $recipients->{$usr->{userid}}->{+REL_BOT_WATCHER}
        = Bugzilla::BugMail::BIT_WATCHING();
    }
  }

  return;
}

sub _users_watching_bots {
  my ($bot_ids, $minor_update) = @_;

  # BUGBUG sqlite doesn't like this sql :(
  return if (lc(Bugzilla->localconfig->{db_driver}) eq 'sqlite');

  my $dbh   = Bugzilla->dbh;
  my $regex = Bugzilla->params->{bot_name_regex};

  my $ids = join(',', uniq @$bot_ids);

  my $query = <<EOS;
      SELECT DISTINCT p.userid
      FROM profiles p
      LEFT JOIN user_group_map ugm ON ugm.user_id = p.userid
      LEFT JOIN agile_team at ON at.group_id = ugm.group_id
      LEFT JOIN bot_watching bw ON bw.user_id = p.userid
      WHERE
      at.group_id IN (
        SELECT DISTINCT group_id
        FROM user_group_map bgm
        LEFT JOIN profiles bu ON bu.userid = bgm.user_id
        WHERE user_id IN ($ids)
        AND bu.login_name ~ ?
      )
      AND (bw.squelch IS NULL or bw.squelch = 0)
EOS

  if ($minor_update) {
    $query .= 'AND (bw.no_minor_updates IS NULL OR bw.no_minor_updates = 0)';
  }

  my $watchers = $dbh->selectall_arrayref($query, {'Slice' => {}}, $regex);

  return $watchers;
}

sub bugmail_relationships {
  my ($self, $args) = @_;
  my $relationships = $args->{relationships};
  $relationships->{+REL_BOT_WATCHER} = 'Bot-Watcher';

  return;
}

__PACKAGE__->NAME;

__END__

=head1 NAME

BotWatching Extension

=head1 DESCRIPTION

This extension allows users to watch internal bot accounts that are members of
Agile teams the user is in. Watching a bot causes the user to receive bug mail for
bugs that they can view that the bot is assignee, qa, doc, or cc'd on.

When an internal bot is in a team, members of that team will watch the bot.

On the Email Preferences tab there is a new column for configuring 'Bots', these
are rules for all bots.

On the 'Bot Watching' preferences tab users can disable all mail, or disable
minor update mail, for individual bots.

=head2 BUGBUG TODO

  * bug mail recipients - mailer_before_send
