# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::DependentProducts;
use constant NAME => 'DependentProducts';

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension);
use List::UtilsBy qw( extract_by );
use List::Compare;

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::Token;
use Bugzilla::Util qw(trim trick_taint detaint_natural diff_arrays);
use Bugzilla::Error;
use List::Util qw(any);
use Bugzilla::WebService::Util qw(filter_wants);

use Bugzilla::Extension::DependentProducts::Util;
use Bugzilla::Extension::DependentProducts::Record;
use Bugzilla::Extension::DependentProducts::Bug;

our $VERSION = 0.2;

our %DP_FIELD_MAP = (
  ${\DP_CONTACT}  => 'contacts',
  ${\DP_PROD_MGR} => 'contacts_prod_manager',
  ${\DP_DEV_MGR}  => 'contacts_dev_manager',
  ${\DP_QE_MGR}   => 'contacts_qe_manager',
);

BEGIN {
  *Bugzilla::Product::dependent_product  = \&_dependent_product;
  *Bugzilla::Product::provider_products  = \&_provider_products;
  *Bugzilla::Product::dependent_products = \&_dependent_products;

  *Bugzilla::Bug::dependent_products      = \&_bugs_dependent_products;
  *Bugzilla::Bug::dependent_product_names = \&_bugs_dependent_product_names;

  *Bugzilla::depprod_all_classifications = \&_depprod_all_classifications;
}

sub _depprod_all_classifications {
  require Bugzilla::Classification;

  return [Bugzilla::Classification->get_all];
}

sub _dependent_product {
  my $self = shift;
  if (not defined $self->{dependent_product}) {
    my $record
      = Bugzilla::Extension::DependentProducts::Record->new(
      {condition => 'product_id = ? AND depends_on IS NULL', values => [$self->id()]}
      );

    unless (defined $record) {
      $record = Bugzilla::Extension::DependentProducts::Record->create(
        {product_id => $self->id});
    }

    $self->{dependent_product} = $record;
  }

  return ($self->{dependent_product});
}

sub _provider_products {
  my $self = shift;
  if (not defined $self->{provider_products}) {
    my $dbh = Bugzilla->dbh();
    my $query
      = "SELECT depends_on FROM dependent_products WHERE product_id = ? and depends_on is not NULL";
    my $layered_products = $dbh->selectcol_arrayref($query, undef, $self->id());
    $self->{provider_products}
      = Bugzilla::Product->new_from_list($layered_products);
  }

  return ($self->{provider_products});
}

sub _dependent_products {
  my $self = shift;
  if (not defined $self->{dependent_products}) {
    my $dbh   = Bugzilla->dbh();
    my $query = "SELECT product_id FROM dependent_products WHERE depends_on = ?";
    my $layered_products = $dbh->selectcol_arrayref($query, undef, $self->id());

    $self->{dependent_products}
      = Bugzilla::Product->new_from_list($layered_products);

  }

  return ($self->{dependent_products});
}

sub _bugs_dependent_products {
  my $self = shift;

  if (not defined $self->{dependent_products}) {
    my $dbh   = Bugzilla->dbh();
    my $query = "SELECT value FROM bug_dependent_products WHERE bug_id = ?";
    my $dependent_products_ids
      = $dbh->selectcol_arrayref($query, undef, $self->id());
    $self->{dependent_products} = $dependent_products_ids;
  }

  return ($self->{dependent_products});
}

sub _bugs_dependent_product_names {
  my $self = shift;

  my $prods = Bugzilla::Product->new_from_list($self->dependent_products);
  my $str   = '';
  if (@$prods) {
    $str = join(', ', map { $_->name } @$prods);
  }

  return ($str);
}

sub db_schema_abstract_schema {
  my ($self, $args) = @_;

  my $schema = $args->{schema};

  $schema->{dependent_products} = {
    FIELDS => [
      id         => {TYPE => 'SERIAL', NOTNULL => 1, PRIMARYKEY => 1,},
      product_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'products', COLUMN => 'id', DELETE => 'CASCADE',}
      },
      depends_on => {
        TYPE       => 'INT3',
        REFERENCES => {TABLE => 'products', COLUMN => 'id', DELETE => 'CASCADE',}
      },
      vis_group =>
        {TYPE => 'INT3', REFERENCES => {TABLE => 'groups', COLUMN => 'id',}},
    ],
    INDEXES => [
      dp_product_depends_idx =>
        {FIELDS => ['product_id', 'depends_on'], TYPE => 'UNIQUE',},
    ],
  };

  $schema->{dependent_products_contacts} = {
    FIELDS => [
      id        => {TYPE => 'SERIAL', NOTNULL => 1, PRIMARYKEY => 1,},
      record_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'dependent_products', COLUMN => 'id',}
      },
      contact => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'profiles', COLUMN => 'userid',}
      },
      contact_role => {TYPE => 'INT1', NOTNULL => 1,},
    ],
  };

  $schema->{bug_dependent_products} = {
    FIELDS => [
      id    => {TYPE => 'SERIAL', NOTNULL => 1, PRIMARYKEY => 1},
      value => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'products', COLUMN => 'id', DELETE => 'CASCADE',}
      },
      bug_id => {
        TYPE       => 'INT3',
        NOTNULL    => 1,
        REFERENCES => {TABLE => 'bugs', COLUMN => 'bug_id', DELETE => 'CASCADE',}
      },
      enabled_by =>
        {TYPE => 'INT3', REFERENCES => {TABLE => 'profiles', COLUMN => 'userid',}},
    ],
    INDEXES => [
      dp_bug_product_id_idx => {FIELDS => ['value', 'bug_id'], TYPE => 'UNIQUE',},
    ],
  };

  return;
}

sub install_update_db {
  my ($self, $args) = @_;

  my $dbh = Bugzilla->dbh;

  # Create the field if it does not exist
  if (!Bugzilla::Field->new({name => 'dependent_products'})) {
    Bugzilla::Field->create({
      name        => 'dependent_products',
      description => 'Dependent Products',
      long_desc   => 'Dependent Products affected by this bug',
      mailhead    => 0,
      type        => FIELD_TYPE_MULTI_SELECT,
    });
  }

  $dbh->bz_add_index('dependent_products', 'dp_product_depends_idx',
    {TYPE => 'UNIQUE', FIELDS => [qw(product_id depends_on)]});

  return;
}

sub bugmail_relationships {
  my ($self, $args) = @_;
  my $relationships = $args->{relationships};

  $relationships->{+DP_CONTACT}  = 'DPPManager';
  $relationships->{+DP_PROD_MGR} = 'DPRManager';
  $relationships->{+DP_DEV_MGR}  = 'DPDevManager';
  $relationships->{+DP_QE_MGR}   = 'DPQEManager';

  return;
}

sub bugmail_recipients {
  my ($self, $args) = @_;
  my $recipients = $args->{recipients};
  my $diffs      = $args->{diffs};
  my $bug        = $args->{bug};

  my $dbh = Bugzilla->dbh;

  # if there are DPs on this bug
  # add DP contacts for the bugs product
  # for each DP on this bug add the contacts for each role set on the DP
  foreach my $prod_id (@{$bug->dependent_products()}) {
    my $product = Bugzilla::Product->new($prod_id);
    my $dp      = $product->dependent_product();
    next unless defined $dp;

    foreach my $contact (@{$dp->all_contacts()}) {
      $recipients->{$contact->contact}->{+$contact->contact_role} = 1;
    }
  }

  return;
}

# TODO group
sub group {
  my $self = shift;

  # BUGBUG should be using vis_group
  return ('redhat');
}

sub update_product_add_vars {
  my ($self, $args) = @_;

  my $product = $args->{product};
  my $changes = $args->{changes};
  my $vars    = $args->{vars};
  my $dbh     = Bugzilla->dbh;
  my (@removed, @added);
  $dbh->bz_start_transaction();

  my $record = $product->dependent_product;

  if ($vars->{vis_group} != $record->vis_group) {
    $changes->{dependent_products}{vis_group}
      = [$record->vis_group, $vars->{vis_group}];
    $record->set_vis_group($vars->{vis_group});
  }

  foreach my $role (DP_CONTACT, DP_QE_MGR, DP_DEV_MGR, DP_PROD_MGR) {
    my $field     = $DP_FIELD_MAP{$role};
    my $old_conts = $record->role_contacts_as_email($role);
    my @new_conts = split(/,\s*/, $vars->{$field});
    my $lc        = List::Compare->new($old_conts, \@new_conts);
    my @rem       = $lc->get_unique;
    my @add       = $lc->get_complement;
    if (@add or @rem) {
      push(@{$changes->{dependent_products}{$field}}, \@rem);
      push(@{$changes->{dependent_products}{$field}}, \@add);
    }
    $record->set_role_contacts($vars->{$field}, $role);
  }

  $record->update();

  my $input_dps = $args->{vars}->{dependent_product};
  $input_dps = [$input_dps] if (ref($input_dps) ne 'ARRAY');

  my $existing_dps = $product->provider_products();

  foreach my $dp (@$existing_dps) {
    my @found = extract_by { $_ eq $dp->id } @$input_dps;
    unless (@found) {
      my $sql
        = 'DELETE FROM dependent_products WHERE product_id = ? AND depends_on = ?';
      $dbh->do($sql, undef, $product->id, $dp->id);
      push(@removed, $dp->name);
    }
  }

  foreach my $prod_id (@$input_dps) {
    next unless (defined($prod_id));
    my $entry = Bugzilla::Extension::DependentProducts::Record->create(
      {product_id => $product->id, depends_on => $prod_id});
    $entry->update();
    my $prod = Bugzilla::Product->new($prod_id);

    push(@added, $prod->name);
  }

  $dbh->bz_commit_transaction();
  $changes->{dependent_products}{core}
    = [join(', ', @removed), join(', ', @added)];
  return;
}

## Start Bug List ##
sub buglist_columns {
  my ($self, $args) = @_;

  my $columns = $args->{'columns'};
  state $pg_suffix
    = lc(Bugzilla->localconfig->{db_driver}) eq 'pg' ? '::text' : '';

  $columns->{'dependent_products'} = {
    'name' =>
      Bugzilla->dbh->sql_group_concat('DISTINCT dp_products.name' . $pg_suffix),
    'title' => 'Dependent Products'
  };

  return;
}

sub buglist_column_joins {
  my ($self, $args) = @_;
  my $joins = $args->{'column_joins'};

  # This column is added using the "buglist_columns" hook
  $joins->{'dependent_products'} = {
    from    => 'bug_id',
    to      => 'bug_id',
    table   => 'bug_dependent_products',
    as      => 'dp_bugs',
    then_to => {
      table => 'products',
      as    => 'dp_products',
      from  => 'dp_bugs.value',
      to    => 'id',
    }
  };

  return;
}
## End Bug List ##

sub bug_fields {
  my ($self, $args) = @_;
  my $fields = $args->{fields};
  push(@$fields, 'dependent_products');

  return;
}

sub post_bug_fields {
  my ($self, $args) = @_;
  my $bug_fields = $args->{bug_fields};
  my $input      = Bugzilla->input_params;
  push @$bug_fields, grep {/^dependent_products$/} keys %$input;

  return;
}

sub bug_before_set_all {
  my ($self, $args) = @_;

  my $bug    = $args->{bug};
  my $params = $args->{params};
  my $input  = Bugzilla->input_params;
  my $dbh    = Bugzilla->dbh;
  my $dps    = [];

  # browser mode sets blank if nothing selected
  # other modes only update if set
  if ((Bugzilla->usage_mode == USAGE_MODE_BROWSER)
    || defined($input->{dependent_products}))
  {
    if (ref($input->{dependent_products}) ne 'ARRAY') {
      $dps->[0] = [$input->{dependent_products}];
    }
    else {
      $dps->[0] = $input->{dependent_products};
    }
    if (!defined($dps->[0]->[0]) || $dps->[0]->[0] ne "--ignore--") {
      $bug->{_dependent_products} = $dps;
    }
  }

  return;
}

sub webservice_bug_update {
  my ($self, $args) = @_;

  my $bug     = $args->{bug};
  my $changes = $args->{changes};

  my $dep_changes = $bug->{_dependent_products}->[0]->[0];

  my $old_products = $bug->dependent_products;    # product_ids.
  my @product_set  = @$old_products;

  if (defined($dep_changes->{add})) {
    foreach my $add_product (@{$dep_changes->{add}}) {

      # Check that this product exists.
      my $p = Bugzilla::Product->check({name => $add_product});

      # Skip if it is already set on this bug.
      next if any { $_ == $p->id } @$old_products;

      push @product_set, $p->id;
    }
  }
  if (defined($dep_changes->{remove})) {
    foreach my $rem_product (@{$dep_changes->{remove}}) {

      # Check that this product exists.
      my $p = Bugzilla::Product->check({name => $rem_product});

      # Skip if it is not set on this bug.
      next unless any { $_ == $p->id } @$old_products;

      @product_set = grep { $p->id != $_ } @product_set;
    }
  }
  if (defined($dep_changes->{set})) {
    @product_set = ();
    foreach my $set_product (@{$dep_changes->{set}}) {
      my $p = Bugzilla::Product->check({name => $set_product});

      # Skip if it is already set on this bug.
      next if any { $_ == $p->id } @$old_products;

      push @product_set, $p->id;
    }
  }

  my ($removed, $added) = diff_arrays($old_products, \@product_set);

  # Make sure we have some work to do.
  if ((scalar(@$removed) + scalar(@$added)) == 0) {
    return;
  }

  # Delete the items we want gone.
  foreach my $removed (@$removed) {
    my $entry
      = Bugzilla::Extension::DependentProducts::Bug->new({
      condition => 'value = ? AND bug_id = ?', values => [$removed, $bug->id]
      });
    $entry->delete;
  }
  my @valid_dependent_products
    = map { $_->id } @{$bug->product_obj->dependent_products};

  # And add the new items.
  foreach my $added (@$added) {

    # Before we add a dependent product, make sure it is valid.
    unless (any { $_ == $added } @valid_dependent_products) {
      my $dep_product = Bugzilla::Product->new($added);
      my $product     = $bug->product_obj;
      ThrowUserError("invalid_dep_product",
        {dep => $dep_product->name, product => $product->name});
    }

    my $entry
      = Bugzilla::Extension::DependentProducts::Bug->create({
      value => $added, bug_id => $bug->id
      });
    $entry->update;
  }

  $changes->{dependent_products} = [join(', ', @$removed), join(', ', @$added)];

  return;
}


sub bug_start_of_update {
  my ($self, $args) = @_;

  if (
    any { Bugzilla->usage_mode == $_ }
    (USAGE_MODE_XMLRPC, USAGE_MODE_JSON, USAGE_MODE_REST)
    )
  {

    webservice_bug_update($self, $args);
    return;
  }

  my $bug       = $args->{bug};
  my $timestamp = $args->{timestamp};
  my $changes   = $args->{changes};

  my (@removed, @added);

  return () unless (defined($bug->{_dependent_products}));

  return unless ($bug->check_can_change_field("dependent_products", 0, 1));

  my $input_dps    = $bug->{_dependent_products}->[0];
  my $existing_dps = $bug->dependent_products();

  foreach my $dp (@$existing_dps) {
    my @found = extract_by { $_ eq $dp } @$input_dps;
    unless (@found) {
      my $entry
        = Bugzilla::Extension::DependentProducts::Bug->new({
        condition => 'value = ? AND bug_id = ?', values => [$dp, $bug->id]
        });
      $entry->delete();
      push(@removed, $dp);
    }
  }

  foreach my $prod_id (@$input_dps) {
    next unless (defined($prod_id));
    my $entry = Bugzilla::Extension::DependentProducts::Bug->create(
      {value => $prod_id, bug_id => $bug->id});
    $entry->update();
    push(@added, $prod_id);
  }

  $bug->{dependent_products} = undef if (@added || @removed);    # clear cache

  $changes->{dependent_products} = [join(', ', @removed), join(', ', @added)]
    if (@added || @removed);

  return;
}

sub bug_filter_change {
  my ($self, $args) = @_;

  my $user    = $args->{user};
  my $field   = $args->{field};
  my $visible = $args->{visible};
  my $bug     = $args->{bug};

  $user = defined $user ? $user : Bugzilla->user;

  if ($field eq 'dependent_products') {
    my $vis_group;
    $vis_group = $bug->product_obj->dependent_product->vis_group if ($bug);
    if ($vis_group && !$user->in_group_id($vis_group)) {
      $$visible = 0;
    }
  }

  return;
}

## REDHAT EXTENSIONS START 1348445
# Convert IDs to product name for emails
sub email_before_template {
  my ($self, $args) = @_;
  my $vars = $args->{vars};

  if (any { $_ eq 'dependent_products' } @{$vars->{changedfields}}) {
    foreach my $diff (@{$vars->{diffs}}) {
      if ($diff->{field_name} eq 'dependent_products') {
        foreach my $type (qw{new old}) {
          my $dps = $diff->{$type};
          if ($dps) {
            my @prods  = split /, */, $dps;
            my $aprods = Bugzilla::Product->new_from_list(\@prods);
            $diff->{$type} = join(', ', map { $_->name() } @$aprods);
          }
        }
      }
    }
  }

  return;
}

# Convert IDs to product names for searching when DPs where changed to or from a product
sub search_operator_field_override {
  my ($self, $args) = @_;

  my $operators = $args->{'operators'};

  $operators->{dependent_products} = {
    _non_changed => \&_dependent_products_notchanged,
    changedfrom  => \&_dependent_products_changedfrom_changedto,
    changedto    => \&_dependent_products_changedfrom_changedto,
  };

  return;
}

sub _dependent_products_notchanged {
  my ($invocant, $args) = @_;
  my ($joins) = @$args{qw(joins)};

  my $join = {
    from    => 'bug_id',
    to      => 'bug_id',
    table   => 'bug_dependent_products',
    as      => 'dp_bugs',
    then_to => {
      table => 'products',
      as    => 'dp_products',
      from  => 'dp_bugs.value',
      to    => 'id',
    }
  };

  push(@$joins, $join);
  $args->{full_field} = 'dp_products.name';

  return;
}

sub _dependent_products_changedfrom_changedto {
  my ($invocant, $args) = @_;

  my ($chart_id, $joins, $field, $operator, $quoted)
    = @$args{qw(chart_id joins field operator quoted)};

  my $column = ($operator =~ /from/) ? 'removed' : 'added';
  $column .= '::INTEGER' if Bugzilla->is_pg;

  ## REDHAT EXTENSION 964029: Since we hide fields, throw user error
  my $field_object = $invocant->_chart_fields->{$field}
    || ThrowUserError("invalid_field_name", {field => $field});
  my $field_id  = $field_object->id;
  my $act_table = "act_${field_id}_$chart_id";
  my $dp_table  = "dependent_products_${field_id}_${chart_id}";

  my $join = {
    table   => 'bugs_activity',
    as      => $act_table,
    extra   => ["$act_table.fieldid = $field_id"],
    then_to => {
      table      => 'products',
      as         => $dp_table,
      from       => $column,
      bugs_table => $act_table,
      to         => 'id',
      incomma    => 1
    },
  };

  $args->{term} = "$act_table.bug_when IS NOT NULL AND $dp_table.name = $quoted";
  $invocant->_changed_security_check($args, $join);
  push(@$joins, $join);

  return;
}

## REDHAT EXTENSIONS END 1348445

sub bug_before_create {
  my ($self, $args) = @_;

  my $params         = $args->{params};
  my $deleted_params = $args->{deleted_params};
  my @dps;

  return unless exists $params->{dependent_products};

  if (ref($params->{dependent_products}) ne 'ARRAY') {
    $dps[0] = $params->{dependent_products};
  }
  else {
    @dps = @{$params->{dependent_products}};
  }

  delete($params->{dependent_products});
  $deleted_params->{dependent_products} = \@dps;

  return;
}

sub bug_end_of_create {
  my ($self, $args) = @_;

  my $bug    = $args->{bug};
  my $params = $args->{deleted_params};

  if ($params->{dependent_products}) {
    foreach my $prod_id (@{$params->{dependent_products}}) {
      my $entry = Bugzilla::Extension::DependentProducts::Bug->create(
        {value => $prod_id, bug_id => $bug->id});
      $entry->update();
    }
  }

  return;
}

sub webservice {
  my ($self, $args) = @_;

  my $dispatch = $args->{dispatch};

  $dispatch->{'DependentProducts'}
    = "Bugzilla::Extension::DependentProducts::WebService";

  return;
}

sub bugs_to_hashes {
  my ($self, $args) = @_;

  my $params = $args->{params};
  my $user   = Bugzilla->user;

  my $orig_extra_fields = $params->{'allowed_extra_fields'};

  $params->{'allowed_extra_fields'} = [qw(
    dependent_products
  )];

  for (my $i = 0; $i < scalar @{$args->{bugs}}; $i++) {
    my $bug  = $args->{bugs}->[$i];
    my $item = $args->{hashes}->[$i];

    ## REDHAT EXTENSION START 1372132
    if (filter_wants $params, 'dependent_products', 'extra') {
      $item->{dependent_products}
        = [map { $_->name }
          grep { $user->can_see_product($_->name) }
          @{Bugzilla::Product->new_from_list($bug->dependent_products)}];
    }
    ## REDHAT EXTENSION END 1372132

  }

  $params->{'allowed_extra_fields'} = $orig_extra_fields;

  return;
}

__PACKAGE__->NAME;

__END__

=head1 NAME

Bugzilla::Extension::DependentProducts

=head1 DESCRIPTION

Dependent Products allows you to track when bugs in one product has an impact
on another product.

e.g. The version of Java shipped in an operating system may impact middlewar
in another product.

=head1 Configuration and Environment

=head2 Dependent (Affected) Products

Dependent products is cotrolled on the product administration page for the
product which is not shipping the package.

e.g. If middleware J is dependent on packages in platform R, then to flag bugs in R as
affecting J, you would go to the administration page for middleware J and add R to the
list labeled "This product depends on:"

A product can depend on multiple other products.

There are three conact fields for this section, (Dev Managers, Product Managers,
QE Managers) they take a comma separated list of email address to email when a bug in
another product is flagged as affecting this product.

=head2 Parent (Affecting) Products

There are two fields on the parent product page related to dependent products.

=over 4

=item Notify for dependent bugs

This field takes a comma separated list of people replated to the parent product who will
be notified wehen any bug is flagged as affecting a dependent product.

e.g. If a bug in product R is flagged as affecting middleware J, then the people listed
in this field on product R will be notified.

=item Visibility group for field

This field limits the users who can see the Dependent Product drop down on bugs in this product.

=back

=head1 Author

Jeff Fearn <jfearn@redhat.com>

=cut

