# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Type::GitHub;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension::ExternalBugs::Type);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(correct_urlbase);

use Data::Dumper;
use JSON;
use URI;
use LWP::UserAgent;
use MIME::Base64;
use HTTP::Request;

use constant TYPE => 'GitHub';

my $logger = Bugzilla->logger;

sub get_data {
  my ($self, $ext_bug) = @_;
  my $action = '';

  if ($ext_bug->ext_bz_bug_id =~ '/pull/') {
    $action = 'pulls';
  }
  elsif ($ext_bug->ext_bz_bug_id =~ '/issues/') {
    $action = 'issues';
  }
  else {
    ThrowUserError("ext_bz_cant_sync_id",
      {badvalue => $ext_bug->ext_bz_bug_id, et_type => TYPE});
  }

  my ($owner, $repo, $number)
    = ($ext_bug->ext_bz_bug_id =~ m{^(.*)/(.*)/(?:issues|pull)/(.*)});

  ThrowUserError("ext_bz_cant_sync_id",
    {badvalue => $ext_bug->ext_bz_bug_id, et_type => TYPE})
    unless ($owner && $repo && $number);

  # e.g. https://github.com/openshift/origin/pull/23962
  my $url  = $ext_bug->type->url;
  my $last = chop($url);
  $url .= $last if ($last ne '/');

  # BUGBUG all github sites must match this or they will go boom ...
  $url =~ s{://}{://api.};
  $url .= "/repos/$owner/$repo/$action/$number";

  # e.g. https://api.github.com/repos/openshift/origin/pulls/23962
  # e.g. https://api.github.com/repos/Pylons/webob/issues/403

  my $data = $self->_do_rest_call($url, 'GET');
  my $bug_data;

  # Get the information we require from the result
  if ($data->{merged}) {
    $bug_data->{status} = 'Merged';
  }
  elsif ($data->{draft}) {
    $bug_data->{status} = 'Draft';
  }
  elsif ($data->{state}) {
    $bug_data->{status} = $data->{state};
  }

  if ($data->{title}) {
    $bug_data->{description} = $data->{title};
  }

  return $bug_data;
}

sub can_sync {
  my ($self, $ext_bug) = @_;
  return 1;
}

sub _do_rest_call {
  my ($self, $url, $method) = @_;

  # Send the request
  my $uri = URI->new($url);
  my $ua  = LWP::UserAgent->new;
  $ua->agent("RedHatBugzilla/" . BUGZILLA_VERSION);    # :)

  if ( Bugzilla->params->{external_tracker_proxy}
    && Bugzilla->params->{external_tracker_proxy} ne '')
  {
    $ua->proxy(['http', 'https'], Bugzilla->params->{external_tracker_proxy});
  }

  my $result  = undef;
  my $content = '';

  # Create a request
  my @headers = ('content-type' => 'application/json;charset=UTF-8');
##    push(@headers, ('authorization' => "Basic $auth")) if ($auth);

  my $req = HTTP::Request->new($method, $uri, \@headers, $content,);

  # Pass request to the user agent and get a response back
  my $res = $ua->request($req);

  $logger->debug("REST Calling: " . $uri->as_string);

  unless ($res->is_success) {
    die $res->status_line;
  }

  # Some calls return no content
  $result = $res->content ? decode_json($res->content) : '';

  if (ref $result and $result->{errors}) {
    my $error = Dumper($result->{errors});
    ThrowUserError('ext_bz_rest_error',
      {errtext => "$error", url => $url, et_type => TYPE});
  }

  # Log a comment, except if we were trying to log in
  $logger->debug("REST Success: $url");

  return $result;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::ExternalBugs::Type::GitHub

=head1 DESCRIPTION

Type represents a GitHub review or issue.  It is an implementation of
L<Bugzilla::Extension::ExternalBugs::Type>, and thus provides all methods that
L<Bugzilla::Extension::ExternalBugs::Type> provides.

This type can sync data from remote gitHib sites to the local site.

=cut
