# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ExternalBugs::Type::GitLab;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::Extension::ExternalBugs::Type);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(correct_urlbase);

use Data::Dumper;
use JSON;
use URI;
use LWP::UserAgent;
use HTTP::Request;
use URI::Escape;

use constant TYPE => 'GitLab';

my $logger = Bugzilla->logger;

sub get_data {
  my ($self, $ext_bug) = @_;
  my $action = '';

  if ($ext_bug->ext_bz_bug_id =~ '/issues/') {
    $action = 'issues';
  }
  elsif ($ext_bug->ext_bz_bug_id =~ '/merge_requests/') {
    $action = 'merge_requests';
  }
  else {
    ThrowUserError("ext_bz_cant_sync_id",
      {badvalue => $ext_bug->ext_bz_bug_id, et_type => TYPE});
  }

  my ($group, $project, $number)
    = ($ext_bug->ext_bz_bug_id =~ m{^(.*)/(.*)/-/(?:issues|merge_requests)/(.*)});
  $group   = uri_escape_utf8($group);
  $project = uri_escape_utf8($project);
  $number  = uri_escape_utf8($number);

  ThrowUserError("ext_bz_cant_sync_id",
    {badvalue => $ext_bug->ext_bz_bug_id, et_type => TYPE})
    unless ($group && $project && $number);

  # e.g. https://gitlab.com/cki-project/kernel-webhooks/-/merge_requests/82
  my $url  = $ext_bug->type->url;
  my $last = chop($url);
  $url .= $last if ($last ne '/');
  $url .= "/api/v4/projects/$group%2F$project/$action/$number";

# e.g. https://gitlab.com/api/v4/projects/cki-project%2Fkernel-webhooks/merge_requests/82

  my $data = $self->_do_rest_call($url, 'GET');
  my $bug_data;

  # Get the information we require from the result
  if ($data->{state}) {
    $bug_data->{status} = $data->{state};
  }
  if ($data->{title}) {
    $bug_data->{description} = $data->{title};
  }

  return $bug_data;
}

sub can_sync {
  my ($self, $ext_bug) = @_;
  return 1;
}

sub _do_rest_call {
  my ($self, $url, $method) = @_;

  # Send the request
  my $uri = URI->new($url);
  my $ua  = LWP::UserAgent->new;
  $ua->agent("RedHatBugzilla/" . BUGZILLA_VERSION);    # :)

  if ( Bugzilla->params->{external_tracker_proxy}
    && Bugzilla->params->{external_tracker_proxy} ne '')
  {
    $ua->proxy(['http', 'https'], Bugzilla->params->{external_tracker_proxy});
  }

  my $result  = undef;
  my $content = '';                                    ##defined $callobj ? encode_json($callobj) : '';

  eval {
    # Create a request
    my @headers = ('content-type' => 'application/json;charset=UTF-8');
##    push(@headers, ('authorization' => "Basic $auth")) if ($auth);

    my $req = HTTP::Request->new($method, $uri, \@headers, $content,);

    # Pass request to the user agent and get a response back
    my $res = $ua->request($req);

    $logger->debug("REST Calling: " . $uri->as_string);

    unless ($res->is_success) {
      die $res->status_line;
    }

    # Some calls return no content
    $result = $res->content ? decode_json($res->content) : '';
  };

  if ($@) {
    my $err = $@;
    ThrowUserError('ext_bz_rest_error',
      {errtext => "$err", url => $url, et_type => TYPE});
  }

  if (ref $result and $result->{errors}) {
    my $error = Dumper($result->{errors});
    ThrowUserError('ext_bz_rest_error',
      {errtext => "$error", url => $url, et_type => TYPE});
  }

  # Log a comment, except if we were trying to log in
  $logger->debug("REST Success: $url");

  return $result;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::ExternalBugs::Type::GitLab

=head1 DESCRIPTION

Type represents a GitLab review or issue.  It is an implementation of
L<Bugzilla::Extension::ExternalBugs::Type>, and thus provides all methods that
L<Bugzilla::Extension::ExternalBugs::Type> provides.

This type can sync data from remote GitLab sites to the local site.

=cut
