# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::ListOfBugs::WebService;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Util qw(trick_taint);
use Bugzilla::Error;
use Bugzilla::Constants;

use Bugzilla::Extension::ListOfBugs::BugList;

use List::MoreUtils qw(any uniq);

use constant PUBLIC_METHODS => qw(
  create
  update
  delete
  list
  get
);

sub create {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  unless (defined($args->{name})) {
    ThrowCodeError('params_required',
      {function => 'ListOfBugs.create', params => [qw(name)]});
  }
  unless (defined($args->{bugs}) && scalar(@{$args->{bugs}})) {
    ThrowUserError('buglist_no_bugs');
  }

  my $name = delete $args->{name};
  my $bugs = delete $args->{bugs};
  my $desc = delete $args->{description};

  my $bl_create = {name => $name};

  $bl_create->{description} = $desc if defined $desc;


  Bugzilla->dbh->bz_start_transaction;

  my $list = Bugzilla::Extension::ListOfBugs::BugList->create($bl_create);

  foreach my $n (@$bugs) {
    trick_taint($n);
  }

  my @bug_ids = grep { $_ > 0 } map { int $_ } @$bugs;

  $list->set_bug_ids(\@bug_ids);

  if (defined $args->{edit_groups}) {
    my $groups      = delete $args->{edit_groups};
    my @edit_groups = map { Bugzilla::Group->check({name => $_}); } @$groups;
    $list->set_edit_groups(\@edit_groups);
  }

  if (defined $args->{view_groups}) {
    my $groups      = delete $args->{view_groups};
    my @view_groups = map { Bugzilla::Group->check({name => $_}); } @$groups;
    $list->set_view_groups(\@view_groups);
  }

  $list->update;

  Bugzilla->dbh->bz_commit_transaction;

  return {name => $list->name, id => $list->id};
}

sub update {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $lists = $self->_map_lists($args);

  if (scalar(@$lists) == 0) {
    ThrowCodeError('params_required',
      {function => 'ListOfBugs.update', params => [qw(ids names)]});
  }

  Bugzilla->dbh->bz_start_transaction;

  my $all_changes = {};

  foreach my $list (@$lists) {

    ThrowUserError('buglist_no_edit_permission') unless $list->can_edit;

    # Buglist update.
    if ($args->{bugs}) {
      my $bl_update = $args->{bugs};
      my @newlist   = @{$list->bug_ids};

      if ($bl_update->{set}) {
        foreach my $id (@{$bl_update->{set}}) {
          trick_taint($id);
        }
        @newlist = grep { $_ > 0 } map { int $_ } @{$bl_update->{set}};
      }
      if ($bl_update->{add}) {
        foreach my $id (@{$bl_update->{add}}) {
          trick_taint($id);
          my $nid = int $id;
          push @newlist, $nid if $nid > 0;
        }
      }
      if ($bl_update->{remove}) {
        foreach my $id (@{$bl_update->{remove}}) {
          trick_taint($id);
          $id      = int $id;
          @newlist = grep { $_ != $id } @newlist;
        }
      }
      @newlist = uniq @newlist;
      $list->set_bug_ids(\@newlist);
    }

    # Update owner of the buglist.
    if ($args->{owner}) {
      my $owner = Bugzilla::User->new({name => $args->{owner}})
        // ThrowUserError('invalid_username', {name => $args->{owner}});
      $list->set_owner($owner->id);
    }

    # Update the view groups.
    if ($args->{view_groups}) {
      my $updated_groups
        = $self->_update_groups($list->view_groups, $args->{view_groups});
      $list->set_view_groups($updated_groups);
    }

    if ($args->{edit_groups}) {
      my $updated_groups
        = $self->_update_groups($list->edit_groups, $args->{edit_groups});
      $list->set_edit_groups($updated_groups);
    }

    if ($args->{description}) {
      trick_taint($args->{description});
      $list->set_description($args->{description});
    }

    my ($changes) = $list->update;

    $all_changes->{$list->id} = $changes;
  }

  Bugzilla->dbh->bz_commit_transaction;

  return $all_changes;
}

sub _update_groups {
  my ($self, $groups, $vg) = @_;

  my @new_groups = @$groups;

  if ($vg->{add}) {
    foreach my $group_name (@{$vg->{add}}) {
      unless (any { $_->name eq $group_name } @$groups) {
        push @new_groups, Bugzilla::Group->check({name => $group_name});
      }
    }
  }
  if ($vg->{remove}) {
    my @temp;
    foreach my $group (@$groups) {
    INNER: foreach my $remove_group (@{$vg->{remove}}) {
        unless ($group->name eq $remove_group) {
          push @temp, $group;
          last INNER;
        }
      }
    }
    @new_groups = @temp;
  }
  if ($vg->{set}) {
    @new_groups = map { Bugzilla::Group->check({name => $_}) } @{$vg->{set}};
  }

  return \@new_groups;
}

sub get {
  my ($self, $args) = @_;
  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $lists = $self->_map_lists($args);

  my $results = [];

  foreach my $list (@$lists) {
    next unless $list->can_view;

    my @view_groups = map { $_->name } @{$list->view_groups};
    my @edit_groups = map { $_->name } @{$list->edit_groups};

    push @$results,
      {
      id          => $list->id,
      name        => $list->name,
      description => $list->description,
      owner       => $list->owner->login,
      updated     => $list->updated,
      lastused    => $list->lastused,
      bugs        => $list->bug_ids,
      view_groups => \@view_groups,
      edit_groups => \@edit_groups,
      };
  }

  return $results;
}

sub list {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my @lists = Bugzilla::Extension::ListOfBugs::BugList->get_all;

  my $result = [];

  foreach my $list (@lists) {
    next unless $list->can_view;

    push @$result,
      {
      id          => $list->id,
      name        => $list->name,
      description => $list->description,
      owner       => $list->owner->login,
      };
  }

  return $result;
}

sub delete {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $lists = $self->_map_lists($args);

  if (scalar(@$lists) == 0) {
    ThrowCodeError('params_required',
      {function => 'ListOfBugs.delete', params => [qw(ids names)]});
  }

  Bugzilla->dbh->bz_start_transaction;

  my $removed = [];

  foreach my $list (@$lists) {

    # Let the user edit it if they are the owner or an admin.
    ThrowUserError('buglist_no_edit_permission') unless $list->can_edit;

    push @$removed, {id => $list->id, name => $list->name};

    $list->remove_from_db;
  }

  Bugzilla->dbh->bz_commit_transaction;

  return {removed => $removed};
}

sub _map_lists {
  my ($self, $args) = @_;

  my $list_ids   = $args->{ids}   // [];
  my $list_names = $args->{names} // [];

  my $lists = [];

  foreach my $id (@$list_ids) {
    my $list = Bugzilla::Extension::ListOfBugs::BugList->check({id => $id});
    ThrowUserError('buglist_no_edit_permission', {name => $id})
      unless $list->can_edit;
    push @$lists, $list;
  }
  foreach my $name (@$list_names) {
    my $list = Bugzilla::Extension::ListOfBugs::BugList->check({name => $name});
    ThrowUserError('buglist_no_edit_permission', {name => $name})
      unless $list->can_edit;
    push @$lists, $list;
  }

  return $lists;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::WebService::ListOfBugs - The API for creating, changing
and getting lists of bugs.

=head1 DESCRIPTION

A List of bugs is a saved, static list of bug numbers.  This allows you to
save a snapshot of bug numbers and present it to other users as a fixed
list of bugs that can only be changed manually.

=head1 METHODS

See L<Bugzilla::WebService> for a description of how parameters are passed,
and what B<STABLE>, B<UNSTABLE> and B<EXPERIMENTAL> mean.

=head2 create

=over

=item B<Description>

Create a bug list.

=item B<Params>

=over

=item C<name>

C<string> The name of this list.  Must be unique

=item C<description>

C<string> A description of this list.

=item C<bugs>

C<array> array of bug ids that will be used when running the list search

=item C<view_groups>

C<array> An array of group names to restrict visiblity of this search to members of the named groups

If no view groups are supplied, the list will be public.

=item C<edit_groups>

C<array> An array of group names to restrict editing of this search to members of the named groups

If no edit groups are supplied, only the creator of the list may edit the list.

=back

=item B<Returns>

=over

=item C<id> - int

The ID of the newly created bug list

=item C<name> - string

The name of the newly created bug list

=back

=back

=head2 get

=over

=item B<Description>

Get the list of bugs associated with this bug list, along with other information.

=item B<Params>

You can pass either a list of ids, or names

=over

=item C<ids> (array) - an array of integer list ids.

=item C<names> (array) - an array of string list names.

=back

=item B<Returns>

=over

=item C<id> The ID of this list.

=item C<name> The name of this list.

=item C<description> A description of this list.

=item C<owner> The user that owns this list.

=item C<updated> The time that this list was created or last modified.

=item C<lastused> The time that this list was last used as a search.

=item C<bugs> An array of bug ids

=back

=back

=head2 B<update>

=over

=item B<Description>

Add or remove bugs in the buglist, or change the ownership or visibility of the list.

=item B<Params>

=over

=item C<ids>

C<array> An array of bug list ids to be updated.

=item C<names>

C<array> An array of bug list names to be updated.

=item C<owner>

C<string> Change the owner of the list to the supplied email address.

=item C<description>

C<string> Change the description to the supplied string.

=item C<bugs>

Update the list of bugs.
A hash containing one or more of the following keys - add, remove, set.

=over

=item C<add>

C<array> The listed bug ids will be added to the bug list.

=item C<remove>

C<array> The listed bug ids will be removed from the bug list.

=item C<set>

C<array> The current list will be removed and replaced with the new bug list.

=back

=item C<view_groups>

These groups control the visibility of the bug list.  A list with no view groups is a public list.

A hash containing one or more of the following keys - add, remove, set.

=over

=item C<add>

C<array> The listed group names will be added to the view group list.

=item C<remove>

C<array> The listed group names will be removed from the view group list.

=item C<set>

C<array> The current view groups will be removed and replaced with the new view groups.

=back

=item C<edit_groups>

These groups control the editability of the group list.
A list with no edit groups may only be edited by the owner of the list.

A hash containing one or more of the following keys - add, remove, set.

=over

=item C<add>

C<array> The listed group names will be added to the edit group list.

=item C<remove>

C<array> The listed group names will be removed from the edit group list.

=item C<set>

C<array> The current view groups will be removed and replaced with the new edit groups.

=back

=back

=back

=head2 list

=over

=item B<Description>

Returns a list of all bug lists that you have permission to view.

=item B<Returns>

Returns an array of hashes.  Each hash contains the following keys:

=over

=item C<id>

C<int> An integer id identifying this bug list

=item C<name>

C<string> A the name of this list

=item C<owner>

C<string> The owner of this search

=item C<description>

C<string> A description of this search

=back

=back

=cut
