# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::PlotlyReports::Graphs::BugSeverity;

use strict;
use warnings;
use 5.10.1;

use Bugzilla;
use Data::Dumper;
local $Data::Dumper::Sortkeys = 1;
use List::MoreUtils qw(uniq);

###############################################################################
# Variables
###############################################################################


sub new {
  my $class = shift;
  my $opt   = shift;
  my $self  = bless {
    title             => $opt->{title} // 'Open Defects by Severity',
    total             => 0,
    max               => 0,
    developer         => $opt->{developer} // 0,
    customer          => $opt->{customer} // 0,
    totals            => $opt->{totals} // 0,
    closed            => $opt->{closed} // 0,
    months            => $opt->{months} // 12,
    from              => $opt->{from} // 0,
    to                => $opt->{to} // 0,
    hide_empty        => $opt->{hide_empty} // 0,
    products          => $opt->{products} // '',
    teams             => {},
    deadline          => $opt->{deadline} // 0,
    internal_deadline => $opt->{internal_deadline} // 0,
    all_severities    => $opt->{all_severities} // 0,
  }, $class;

  $self->_init();

  return $self;
}

sub _init {
  my $self   = shift;
  my $months = $self->{months};

  if ($months < 0) {
    $months *= -1;
  }

  my @date = ();

  if ($self->{from} || $self->{to}) {
    if (!$self->{from}) {
      $self->{from} = `date --date='-$months months' +'%Y%m%d'`;
      chomp $self->{from};
    }
    if (!$self->{to}) {
      $self->{to} = `date +'%Y%m%d'`;
      chomp $self->{to};
    }
    my $end_date   = undef;
    my $start_date = undef;
    $end_date = `date --date='+1 months $self->{to}' +'%Y%m'`;
    chomp $end_date;
    $end_date .= '01';
    $start_date = `date --date='-1 month $self->{from}' +'%Y%m'`;
    chomp $start_date;
    $start_date .= '01';

    $months = -2;
    while ($end_date > $start_date) {
      push @date, $end_date;
      $months++;
      $end_date = `date --date='-1 months $end_date' +'%Y%m'`;
      chomp $end_date;
      $end_date .= '01';
    }
  }
  else {
    for (my $i = -1; $i <= $months + 1; $i++) {
      my $offset = $i * -1;
      my $wd     = `date --date='$offset months' +'%Y%m'`;
      chomp $wd;
      $wd .= '01';
      push @date, $wd;
    }
  }

  $self->{date} = \@date;

  return;
}

# Calculate maximum to improve format
sub _SetMax {
  my $self = shift;

  my $data = shift();
  $self->{max} = 0;
  for (my $j = 0; $j < @{$$data[0]}; $j++) {
    my $tot = 0;
    for (my $i = 1; $i < @{$data}; $i++) {
      $tot += $$data[$i][$j];
    }

    if ($tot > $self->{max}) { $self->{max} = $tot; }
  }
}

# set every undefined value in the matrix to 0 and return the totals
sub _complete_matrix {
  my $self    = shift;
  my $param   = shift();
  my @columns = @{$param->{columns}};
  my $matrix  = $param->{matrix};
  my %sum;

  foreach my $pool_id (keys %{$self->{teams}}) {
    $sum{$pool_id} = 0;
  }
  foreach my $column (@columns) {
    $sum{$column} = 0;
  }
  $sum{total} = 0;

  foreach my $column (@columns) {
    foreach my $pool_id (keys %{$self->{teams}}) {
      my $name = $self->{teams}->{$pool_id}->{name};
      my $sev  = $column;
      if (defined $matrix->{$name} && defined $matrix->{$name}->{$sev}) {
        $sum{$pool_id} += $matrix->{$name}->{$sev};
        $sum{$column}  += $matrix->{$name}->{$sev};
        $sum{total}    += $matrix->{$name}->{$sev};
      }
      else {
        $matrix->{$name}->{$sev} = 0;
      }
    }
  }

  return \%sum;
}

sub _get_teams {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  my @tmp;

  my $sql_query = <<SQL;
SELECT agile_sprint.id, agile_team.name
FROM agile_sprint
LEFT JOIN agile_team ON agile_sprint.team_id = agile_team.id
UNION
SELECT agile_backlog.pool_id, agile_team.name
FROM agile_backlog
LEFT JOIN agile_team ON agile_backlog.team_id = agile_team.id
SQL

  my $sth = $dbh->prepare($sql_query);
  $sth->execute();

  while (my $row = $sth->fetchrow_hashref) {
    $self->{teams}->{$row->{id}} = {name => $row->{name}, id => $row->{id}};
  }

  my @all_groups = sort { $a->{name} cmp $b->{name} } @tmp;

  $self->{teams}->{0} = {name => 'No Team', id => 0};

  return;
}

# Graph results
sub generate {
  my $self    = shift();
  my $param   = $self->{params};
  my @columns = @{$param->{columns}};
  my $matrix  = $param->{matrix};
  my @groups  = ();

  if ($self->{hide_empty}) {

    # check for rows with data
    foreach my $pool_id (keys %{$self->{teams}}) {
      my $use = 0;
      foreach my $column (@columns) {
        my $name = $self->{teams}->{$pool_id}->{name};
        my $sev  = $column;
        $use = 1 if (defined $matrix->{$name} && $matrix->{$name}->{$sev});
      }
      push @groups, $self->{teams}->{$pool_id}->{name} if ($use);
    }
  }
  else {
    @groups = map { $_->{name} } values(%{$self->{teams}});
  }

  @groups = sort { $b cmp $a } uniq(@groups);

  my @data = ();

  foreach my $column (@columns) {
    my @rowdata = ();
    foreach my $name (@groups) {
      my $sev = $column;
      push @rowdata, $matrix->{$name}->{$sev};
    }
    push @data, \@rowdata;
  }

  unshift @data, \@groups;

  $self->_SetMax(\@data);
  $self->{data} = \@data;
  return;
}

# adjust data to display bugs per developer
## BUGBUG this hasn't been modified to work with changes to $matrix
sub _developer_adjust {
  my $self       = shift;
  my $param      = shift();
  my @all_groups = @{$param->{rows}};
  my @severities = @{$param->{columns}};
  my $matrix     = $param->{matrix};
  my $sums       = $param->{totals};

  my @groups           = ();
  my $relevant_members = 0;

  if ($self->{hide_empty}) {

    # check for rows with data
    foreach my $row (@all_groups) {
      next if $row->{id} eq 'No Team';
      my $use = 0;
      foreach my $column (@severities) {
        my $id = $row->{id} . '\\' . $column;
        $use = 1 if $matrix->{$id};
      }
      push @groups, $row if $use;
    }

    $param->{rows} = \@groups;
  }
  else {
    @groups = @all_groups;
  }

  for (my $i = 0; $i < @groups; $i++) {

    my @group_members;

#    if ($groups[$i]->{id} ne 'No Group') {
#      my $g_mem_ref = _orgchart_func($orgchart, 'group_members', $groups[$i]->{id});
#      @group_members = @$g_mem_ref;
#    }
#    else {
    @group_members = @{$param->{nogroup}};

#    }

    next if $#group_members < 0;
    my $member_count = $#group_members + 1;

    my $result
      = sprintf(" (%d/%d)", $sums->{$groups[$i]->{id}}, $#group_members + 1);

    my $developer_group      = $groups[$i]->{id} . $result;
    my $developer_group_name = $groups[$i]->{name} . $result;

    foreach my $severity (@severities) {
      my $id  = $groups[$i]->{id} . '\\' . $severity;
      my $did = $developer_group . '\\' . $severity;
      $matrix->{$did} = sprintf("%.3f", $matrix->{$id} / $member_count);
    }
    $sums->{$developer_group}
      = sprintf("%.3f", $sums->{$groups[$i]->{id}} / $member_count);
    @{$param->{rows}}[$i]->{id}   = $developer_group;
    @{$param->{rows}}[$i]->{name} = $developer_group_name;
    $relevant_members += $member_count;
  }

  # if real totals exist, adjust these as well
  if (defined $param->{real_totals}) {
    my $real_sums = $param->{real_totals};

    #   my $mem_ref   = _orgchart_func($orgchart, 'members');
    #   my @members   = @$mem_ref;
    #   $relevant_members = $#members + 1 if !$relevant_members;

    foreach my $severity (@severities) {
      $real_sums->{$severity}
        = sprintf("%.3f", $real_sums->{$severity} / $relevant_members);
    }
    $real_sums->{total} = sprintf("%.3f", $real_sums->{total} / $relevant_members);
  }

  return;
}

# apply bug/counts to the data hash
sub _store_bug {
  my $self   = shift;
  my $bug    = shift;
  my $param  = shift;
  my $matrix = $param->{matrix};


  my $name = $self->{teams}->{$bug->{pool}}->{name};
  my $sev  = $bug->{severity};

  if (defined $matrix->{$name} && defined $matrix->{$name}->{$sev}) {
    $matrix->{$name}->{$sev} += defined $bug->{count} ? $bug->{count} : 1;
  }
  else {
    $matrix->{$name}->{$sev} = defined $bug->{count} ? $bug->{count} : 1;
  }

  # set totals, if referenced
  if (exists $param->{real_totals}) {
    my $totals   = $param->{real_totals};
    my $severity = $bug->{severity};
    if (defined $totals->{$severity}) {
      $totals->{$severity} += defined $bug->{count} ? $bug->{count} : 1;
    }
    else {
      $totals->{$severity} = defined $bug->{count} ? $bug->{count} : 1;
    }
    if (defined $totals->{total}) {
      $totals->{total} += defined $bug->{count} ? $bug->{count} : 1;
    }
    else {
      $totals->{total} = defined $bug->{count} ? $bug->{count} : 1;
    }
  }

  return;
}

sub run {
  my $self = shift;

  my $dbh = Bugzilla->dbh;

  my $sth = $dbh->prepare("SELECT value FROM bug_severity ORDER BY id");
  $sth->execute();

  my @bug_severity;

  while (my ($value) = $sth->fetchrow_array) {
    if ($self->{all_severities} || $value =~ m/(?:urgent|high)/) {
      push @bug_severity, $value;
    }
  }

  $sth->finish();

  my $products = '';

  # restrict to products, if given
  $products
    = " AND bugs.product_id IN (SELECT products.id FROM products WHERE products.name IN ("
    . join(', ', map { $dbh->quote($_) } split(',', $self->{products})) . "))"
    if $self->{products} ne '';

  # restrict clause for customer issues
  my $sfdc_select
    = "SELECT DISTINCT ext_bz_bug_map.bug_id FROM ext_bz_bug_map WHERE ext_bz_bug_map.ext_bz_id=60";
  my $customer_where = '';
  $customer_where = " AND (bugs.bug_id IN ($sfdc_select))" if ($self->{customer});

  my $date_format = $dbh->sql_date_format('bug_when', '%Y%m%d');

  my $activity_table_query = <<STATUS_TEMP_TABLE_QUERY;
          SELECT DISTINCT bug_id
            FROM bugs_activity
           WHERE bugs_activity.fieldid=(SELECT id FROM fielddefs WHERE name='bug_status')
             AND added='CLOSED'
             AND ($date_format > $self->{date}->[$#{$self->{date}}])
             AND ($date_format < $self->{date}->[0])
STATUS_TEMP_TABLE_QUERY

  my $bugs_filter = "";
  if ($self->{closed} && $self->{developer}) {
    $bugs_filter
      = " (bugs.bug_status='CLOSED') AND (bugs.bug_id IN ($activity_table_query))";
  }
  else {
    $bugs_filter = " (bugs.bug_status!='CLOSED')";
  }

  my $deadline = '';
  if ($self->{deadline}) {
    if ($self->{internal_deadline}) {
      $deadline
        = q{ AND ( bugs.cf_deadline IS NOT NULL AND bugs.cf_deadline <= (NOW() + interval }
        . $dbh->quote($self->{deadline}) . q{))};
    }
    else {
      $deadline
        = q{ AND ( bugs.deadline IS NOT NULL AND bugs.deadline <= (NOW() + interval }
        . $dbh->quote($self->{deadline}) . q{))};
    }
  }

  my $sql_query = <<BUG_SEVERITY_QUERY;
          SELECT profiles.login_name AS email,
                 bugs.bug_severity AS severity,
                 COUNT(DISTINCT bugs.bug_id) AS count,
                 bug_agile_pool.pool_id as pool
            FROM profiles
                 LEFT JOIN bugs ON profiles.userid=bugs.assigned_to
                 LEFT JOIN bug_agile_pool ON bugs.bug_id=bug_agile_pool.bug_id
           WHERE
                 $bugs_filter
                 $customer_where
                 $products
                 $deadline
        GROUP BY profiles.login_name,
                 bugs.bug_severity,
                 bug_agile_pool.pool_id
BUG_SEVERITY_QUERY

  $self->_get_teams();

  my %group_severity;
  my %total_sums;

  # create param var for subs
  my $params
    = {columns => \@bug_severity, matrix => \%group_severity, nogroup => [],};

  if ($self->{customer}) {
    $params->{real_totals} = \%total_sums;
  }

  Bugzilla->logger->debug($sql_query . "\n");

  $sth = $dbh->prepare($sql_query);
  $sth->execute();

  while (my $row = $sth->fetchrow_hashref) {

    my $bug = {
      count    => $row->{count},
      member   => $row->{email},
      severity => $row->{severity},
      pool     => ($row->{pool} || 0),
    };

    $self->_store_bug($bug, $params);
  }

  $sth->finish();

  # retrieve real totals
  if (!$self->{customer} && $self->{totals}) {
    my $sql_totals_query = <<BUG_SEVERITY_TOTALS_QUERY;
           SELECT bugs.bug_severity AS severity,
                  COUNT(DISTINCT bugs.bug_id) AS count
             FROM bugs
            WHERE (bugs.bug_status!='CLOSED')
              $customer_where
                  $products
         GROUP BY bugs.bug_severity
BUG_SEVERITY_TOTALS_QUERY

    if ($self->{closed}) {
      $sql_totals_query = <<BUG_SEVERITY_TOTALS_CLOSED_QUERY;
                SELECT bugs.bug_severity AS severity,
                       COUNT(DISTINCT bugs.bug_id) AS count
                  FROM bugs
                 WHERE (bugs.bug_status='CLOSED')
                   $customer_where
                   AND (bugs.bug_id IN ($activity_table_query))
                       $products
              GROUP BY bugs.bug_severity
BUG_SEVERITY_TOTALS_CLOSED_QUERY
    }

    $sth = $dbh->prepare($sql_totals_query);
    $sth->execute();

    my $total = 0;
    while (my ($severity, $no_bugs) = $sth->fetchrow_array) {
      if (defined $total_sums{$severity}) {
        $total_sums{$severity} += $no_bugs;
      }
      else {
        $total_sums{$severity} = $no_bugs;
      }
      $total += $no_bugs
        if ($self->{all_severities} || $severity =~ m/(?:urgent|high)/);
    }
    $total_sums{total} = $total;

    $sth->finish();

    $params->{real_totals} = \%total_sums;
  }

  $dbh->disconnect();

  # adjust values and title

  $params->{totals} = $self->_complete_matrix($params);

  if ($self->{developer}) {
    $self->_developer_adjust($params);
    if ($self->{totals}) {
      $self->{title} .= ": " . $total_sums{total} . " Bugs/Developer";
    }
    else {
      $self->{title} .= " per Developer";
    }
  }
  elsif ($self->{totals}) {
    $self->{title} .= ": " . ($total_sums{total} // 0) . " Total";
  }

  if ($self->{customer}) {
    $self->{title} .= " from customers";
  }

  $self->{params} = $params;

  return;
}

1;

__END__

=head1 Name

Bugzilla::Extension::PlotlyReports::Graphs::BugSeverity

=head1 Description

This module generates reports of the current open defects for a product or
products based on bug severity and grouped by the Agile team on the bug.

=head1 METHODS

=head2 new

=over

=item B<Description>

Creates an object to generate a graph with.

  my %opt = (totals => 1, hide_empty => 0);
  my $gf  =
    Bugzilla::Extension::PlotlyReports::Graphs::BugSeverity->new({%opt});

  $gt->run();
  $gt->generate();

=item B<Params>

=over

=item C<title>

C<string> The start string for the graph title. Defaults to
'Open Defects by Severity'. It will get modified by other parameters.

=item C<developer>

C<bool> display results per developer.

TODO: This feature has not been ported yet.

=item C<customer>

C<bool> Limit results to customer bugs. This is defined as bugs with an SFDC
external tracker.

=item C<totals>

C<bool> Include totals in the output.

=item C<closed>

C<bool> report on closed bugs.

=item C<months>

C<int> When displaying closed bugs set the number of months to show. defaults to
12.

=item C<from>

C<date> When displaying closed bugs set the date to start searching from.

=item C<to>

C<date> When displaying closed bugs set the date to stop searching from.

=item C<hide_empty>

C<bool> Do not show Agile teams that do not have any matching bugs. Defaults off.

=item C<products>

C<string> a comma separated list of product names to restrict the results to.

=item C<deadline>

C<string> A date math string to compare to the deadline field.

e.g. "2weeks" will limit the results to bugs that have a deadline less than 2
weeks from now.

=item C<all_severities>

C<bool> list all severities. Defaults off, which only shows urgent and high
severities.

=back

=item B<Returns>

A C<Bugzilla::Extension::PlotlyReports::Graphs::BugSeverity> object.

=back

=head2 run

=over

=item B<Description>

Run the queries to gather the bug data.

=item B<Params>

None.

=item B<Returns>

Nothing.

=back

=head2 generate

=over

=item B<Description>

Process the bug data and generate the results.

=item B<Params>

None.

=item B<Returns>

Nothing.

=back

=cut
