# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::PlotlyReports::WebService;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Util qw(trick_taint trim);
use Bugzilla::Error;
use Bugzilla::Constants;

use Bugzilla::Extension::PlotlyReports::Graphs::BugStatus;
use Bugzilla::Extension::PlotlyReports::Graphs::BugSeverity;
use Bugzilla::Extension::PlotlyReports::Graphs::NonDefaults;

use List::MoreUtils qw(any uniq);
use JSON;
use Data::Dumper;

use constant PUBLIC_METHODS => qw(
  bug_status
  bug_severity
  non_defaults
  job_queue_stats
);

sub bug_status {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  unless (defined($params->{product}) && defined($params->{product}->{names})) {
    ThrowCodeError('params_required',
      {function => 'PlotlyReports.bug_status', params => [qw(product)]});
  }

  my @products;

  foreach my $pname (@{$params->{product}->{names}}) {

    # This checks you can access the product
    my $product = Bugzilla::Product->check({name => $pname});

    # This checks you can also report on the product
    push(@products, $product) if ($user->can_report_on_product($product));

  }

  ThrowUserError('no_reportable_products') unless (scalar @products);

  my $prod_names = join(',', map { $_->name } @products);

## BUGBUG see ESO/scripts/products.pl to see what graphs were interesting
  my %opt = (
    bz_active_products       => 0,
    bz_all_products          => 1,
    bz_by                    => undef,
    bz_closed_as_open        => 0,
    bz_components            => 0,
    bz_constraint            => undef,
    bz_email                 => undef,
    bz_flag                  => undef,
    bz_keyword               => undef,
    bz_order_by              => undef,
    bz_products              => 0,
    bz_reporter              => 0,
    bz_restrict_by_component => undef,
    bz_restrict_by_product   => $prod_names,
    bz_verified              => undef,
    bz_versions              => 0,
    customer                 => 0,
    date_format              => undef,
    date_from                => $params->{date_from} || 0,
    date_to                  => $params->{date_to}   || 0,
    graph_type               => 'lines',
    help                     => 0,
    interval_count           => $params->{interval_count}  || 12,
    interval_period          => $params->{interval_period} || 'months',
    man                      => 0,
    max_bugs                 => undef,
    orgchart_all_groups      => 0,
    orgchart_by_group_member => 0,
    orgchart_email           => undef,
    orgchart_group           => $params->{group} || undef,
    output_file_prefix       => undef,
    resolution               => undef,
    show_closed              => 1,
    show_new                 => 1,
    show_open                => 1,
    split_new                => 0,
    split_stalled            => 0,
    title                    => "",
    verbose                  => 0,

    bz_by_priority => $params->{by_group} ? (lc($params->{by_group}) eq 'priority')
    : 0,
    bz_by_severity => $params->{by_group} ? (lc($params->{by_group}) eq 'severity')
    : 0,
    bz_by_status => $params->{by_group} ? (lc($params->{by_group}) eq 'status') : 1,
    show_trend   => $params->{by_group} ? (lc($params->{by_group}) eq 'trend')  : 0,

    show_averages => $params->{show} ? (lc($params->{show}) eq 'averages')      : 0,
    show_stalled  => $params->{show} ? (lc($params->{show}) eq 'stalled')       : 0,
    n_day_average => $params->{show} ? (lc($params->{show}) eq 'n_day_average') : 0,
  );

  my $gt;
  my @data;

  my %colors = (
    NEW             => 'rgb(248, 248, 248)',
    CLOSED          => 'rgb(0, 0, 0)',
    RELEASE_PENDING => 'rgb(212, 236, 156)',
    VERIFIED        => 'rgb(234, 245, 205)',
    ON_QA           => 'rgb(154, 228, 243)',
    ON_DEV          => 'rgb(217, 235, 237)',
    MODIFIED        => 'rgb(217, 245, 251)',
    POST            => 'rgb(249, 220, 156)',
    ASSIGNED        => 'rgb(252, 237, 205)',
  );

# BUGBUG above, you should be able to show average, n-day-av, stalled, etc, on trend ...
  $gt = Bugzilla::Extension::PlotlyReports::Graphs::BugStatus::Trend->new({%opt});
  $gt->run();
  $gt->generate();

  my $max = @{$gt->{data}};
  for (my $i = 1; $i < $max; $i++) {
    my %row = (
      x    => $gt->{data}->[0],
      y    => $gt->{data}->[$i],
      type => 'scatter',
      mode => 'lines',
      name => $gt->{legend}->[$i - 1],
##      line      => {color => $colors{uc($gt->{legend}->[$i - 1])},},
    );
    push(@data, \%row);
  }

  return ({
    layout => {
      title     => $gt->chart_title() . '<br><sub>' . $gt->sub_title() . '</sub>',
      hovermode => 'closest',
    },
    data        => \@data,
    annotations => [{
      text => $gt->sub_title(),
      font => {size => 13, color => 'rgb(116, 101, 130)',}
    }],
  });
}

sub bug_severity {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  unless (defined($params->{product}) && defined($params->{product}->{names})) {
    ThrowCodeError('params_required',
      {function => 'PlotlyReports.bug_severity', params => [qw(product)]});
  }

  my @products;
  my $hide_empty        = $params->{hide_empty} // 0;
  my $customer          = 0;
  my $internal_deadline = 0;

  # Only insiders can see customer issues or internal deadlines.
  if ($user->is_insider) {
    $customer          = $params->{customer}          ? 1 : 0;
    $internal_deadline = $params->{internal_deadline} ? 1 : 0;
  }
  my $deadline       = $params->{deadline}       // 0;
  my $all_severities = $params->{all_severities} // 0;

  foreach my $pname (@{$params->{product}->{names}}) {

    # This checks you can access the product
    my $product = Bugzilla::Product->check({name => $pname});

    # This checks you can also report on the product
    push(@products, $product) if ($user->can_report_on_product($product));

  }

  ThrowUserError('no_reportable_products') unless (scalar @products);

  my $prod_names = join(',', map { $_->name } @products);


  my %opt = (
    developer         => 0,
    customer          => $customer,
    totals            => 1,
    closed            => 0,
    months            => 12,
    from              => 0,
    to                => 0,
    hide_empty        => $hide_empty,
    products          => $prod_names,
    deadline          => $deadline,
    internal_deadline => $internal_deadline,
    all_severities    => $all_severities,
  );

  my $gt = Bugzilla::Extension::PlotlyReports::Graphs::BugSeverity->new({%opt});
  $gt->run();
  $gt->generate();

  my %colors = (
    urgent      => 'rgb(204,0,0)',
    heigh       => 'rgb(236,122,8)',
    medium      => 'rgb(240,171,0)',
    low         => 'rgb(100,13,13)',
    unspecified => 'rgb(63,156,53)',
  );

  my @data;
  my $max = @{$gt->{data}};
  for (my $i = 1; $i < $max; $i++) {
    my %row = (
      y           => $gt->{data}->[0],
      x           => $gt->{data}->[$i],
      type        => 'bar',
      orientation => 'h',
      name        => $gt->{params}->{columns}->[$i - 1],
      marker      =>
        {color => $colors{$gt->{params}->{columns}->[$i - 1]}, opacity => 0.8,},
    );
    unless ($row{name} =~ m/(?:urgent|high)/) {
      $row{visible} = 'legendonly';
    }
    push(@data, \%row);
  }

  my $height = 250;
  $height = 300 if ($all_severities);
  my $cal_height = (100 + (scalar @{$gt->{data}->[0]} * 20));

  $height = $cal_height if ($cal_height > $height);

  return ({
    layout => {
      title     => $gt->{title},
      barmode   => 'stack',
      margin    => {l => 200},
      height    => $height,
      hovermode => 'closest',
    },
    data => \@data,
  });
}

sub non_defaults {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  unless (defined($params->{product}) && defined($params->{product}->{names})) {
    ThrowCodeError('params_required',
      {function => 'PlotlyReports.bug_severity', params => [qw(product)]});
  }

  my @products;

  foreach my $pname (@{$params->{product}->{names}}) {

    # This checks you can access the product
    my $product = Bugzilla::Product->check({name => $pname});

    # This checks you can also report on the product
    push(@products, $product) if ($user->can_report_on_product($product));

  }

  ThrowUserError('no_reportable_products') unless (scalar @products);

  my $prod_names = join(',', map { $_->name } @products);


  my %opt = (products => $prod_names,);

  my @args = qw/new closed by_pool title/;
  foreach my $arg (@args) {
    if (exists($params->{$arg})) {
      $opt{$arg} = $params->{$arg};
      trick_taint($opt{$arg});
    }
  }

  my $report
    = Bugzilla::Extension::PlotlyReports::Graphs::NonDefaults->new({%opt});
  $report->run();

  my @data;
  foreach my $prod (keys %{$report->{data}}) {
    my %row = (
      y    => $report->{data}->{$prod}->{'counts'},
      x    => $report->{data}->{$prod}->{'labels'},
      ids  => $report->{data}->{$prod}->{'bugs'},
      type => 'bar',
      name => $prod,
    );
    push(@data, \%row);
  }

#  my $height = 250;
#  $height = 300 if ($all_severities);
#  my $cal_height = (100 + (scalar @{$report->{data}->[0]} * 20));

#  $height = $cal_height if ($cal_height > $height);

  return ({
    layout => {
      title     => $report->{title},
      barmode   => 'group',
      margin    => {l => 200},
      hovermode => 'closest',
    },
    data => \@data,
  });
}

## REDHAT EXTENSION START 1142506
sub job_queue_stats {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->is_insider
    || ThrowUserError('auth_failure',
    {group => 'redhat', action => 'run', object => 'job_queue_stats'});

  my $time_period = $params->{time_period} // 'Day';

  my %time_map = (

    #'Month' => {'trunc' => 'hour',   'subtraction' => '1 month'},
    'Week' => {'trunc' => 'hour',   'subtraction' => '7 day'},
    'Day'  => {'trunc' => 'hour',   'subtraction' => '1 day'},
    'Hour' => {'trunc' => 'minute', 'subtraction' => '1 hour'},
  );

  exists $time_map{$time_period}
    || ThrowUserError('invalid_time_period',
    {time_period => $time_period, periods => [keys(%time_map)]});

  my $dbh = Bugzilla->dbh;

  my $trunc    = $time_map{$time_period}->{trunc};
  my $interval = $time_map{$time_period}->{subtraction};
  my $tz       = $user->timezone->name;

  # BUGBUG postgresql specific ...
  my $query = qq{
    WITH hours AS (
      SELECT generate_series(
        date_trunc('$trunc', now()) - '$interval'::interval,
        date_trunc('$trunc', now()),
        '1 $trunc'::interval
      ) AS dt
    )

    SELECT count(es.jobid), hours.dt at time zone '$tz'
    FROM
      hours
      LEFT JOIN ts_exitstatus es ON date_trunc('$trunc', to_timestamp(es.completion_time)) = hours.dt
      LEFT JOIN ts_note ON es.jobid = ts_note.jobid
    WHERE ts_note.notekey IS NULL OR ts_note.notekey = 'starttime'
    GROUP BY hours.dt
    ORDER BY hours.dt
  };

  my @x;
  my @y;

  foreach my $dbrow (@{$dbh->selectall_arrayref($query)}) {
    push(@x, $dbrow->[1]);
    push(@y, $dbrow->[0]);
  }

  my @data = ({
    x    => \@x,
    y    => \@y,
    type => 'scatter',
    mode => 'lines+markers',
    name => 'Completed',
  });

  my (@x2, @y2);

  $query = qq{
    WITH hours AS (
      SELECT generate_series(
        date_trunc('$trunc', now()) - '$interval'::interval,
        date_trunc('$trunc', now()),
        '1 $trunc'::interval
      ) AS dt
    )

    SELECT count(ts_note.jobid), hours.dt at time zone '$tz'
    FROM
      hours
      LEFT JOIN ts_note ON date_trunc('$trunc', to_timestamp(convert_from(value, 'UTF-8')::bigint)) = hours.dt
    WHERE ts_note.notekey IS NULL OR ts_note.notekey = 'starttime'
    GROUP BY hours.dt
    ORDER BY hours.dt
  };

  foreach my $dbrow (@{$dbh->selectall_arrayref($query)}) {
    push(@x2, $dbrow->[1]);
    push(@y2, $dbrow->[0] == 1 ? 0 : $dbrow->[0]);
  }

  push(
    @data,
    {
      x    => \@x2,
      y    => \@y2,
      type => 'scatter',
      ##mode => 'lines',
      name => 'Created',
    }
  );

  return ({
    layout => {title => "Job queue stats for last $interval using timezone $tz",},
    data   => \@data,
  });
}
## REDHAT EXTENSION END 1142506

1;

__END__
