# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::Job::DeSpamComments;
use strict;
use warnings;
use 5.10.1;

BEGIN {
  ## no critic (ProhibitStringyEval)
# This is safe and avoids having to conditionally use this module if TheSchwartz is not installed
  eval "use parent qw(TheSchwartz::Worker Bugzilla::Extension)";
  ## use critic
}

# The longest we expect a job to possibly take, in seconds.
use constant grab_for => 600;

# We don't want to fail permanently very easily. Retry for 48 hours.
use constant max_retries => 53;

# REDHAT EXTENSION START 1142506
# In seconds, 604800 == 7 days
sub keep_exit_status_for { return 604800; }

# REDHAT EXTENSION END 1142506

# The first few retries happen quickly, but after that we wait an hour for
# each retry.
sub retry_delay {
  my ($class, $num_retries) = @_;
  if ($num_retries < 5) {
    return (10, 30, 60, 300, 600)[$num_retries];
  }

  # One hour
  return 60 * 60;
}

sub work {
  my ($class, $job) = @_;

  Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));

  eval { _insinkerator($job); };

  if ($@) {
    $job->failed($@);
    undef $@;
  }
  else {
    $job->completed;
  }

  return;
}

sub _insinkerator {
  my $job = shift;

  # This is the spammers user_id
  my $spammer    = $job->arg->{spammer};
  my $spam_user  = Bugzilla::User->check($spammer);
  my $spammer_id = $spam_user->id;

  my $actor = Bugzilla->user->id;
  my $dbh   = Bugzilla->dbh;

  $dbh->bz_start_transaction;

  my $new_text
    = 'This comment was flagged a spam, view the edit history to see the original text if required.';

  # Log tag activity on untagged comments
  my $SQL = <<EOS;
INSERT INTO longdescs_tags_activity
  (bug_id, comment_id, who, bug_when, added, removed)
SELECT bug_id, comment_id, ?, NOW(), 'spam', ''
FROM longdescs
WHERE
  who = ?
AND
  comment_id NOT IN (
    SELECT comment_id FROM longdescs_tags WHERE tag = 'spam'
  )
EOS

  $dbh->do($SQL, undef, $actor, $spammer_id);

  # Log edit activity on unedited comments
  $SQL = <<EOS;
INSERT INTO longdescs_activity
  (comment_id, who, change_when, old_comment, security)
SELECT comment_id, ?, NOW(), thetext, 0
FROM longdescs
WHERE
  who = ?
AND
  thetext <> ?
EOS

  $dbh->do($SQL, undef, $actor, $spammer_id, $new_text);

  # Edit text on unedited comments
  $SQL = <<EOS;
UPDATE longdescs SET thetext = ?
WHERE comment_id IN (
SELECT comment_id
WHERE
  who = ?
AND
  thetext <> ?
)
EOS

  $dbh->do($SQL, undef, $new_text, $spammer_id, $new_text);

  # Lastly, tag untagged comments
  $SQL = <<EOS;
INSERT INTO longdescs_tags(comment_id, tag)
SELECT comment_id, 'spam'
FROM longdescs
WHERE
  who = ?
AND
  comment_id NOT IN (
      SELECT comment_id FROM longdescs_tags WHERE tag = 'spam'
);
EOS

  $dbh->do($SQL, undef, $spammer_id);
  $dbh->bz_commit_transaction();

  # _completely_ remove any attachments the spammer created
  # BUGBUG this won't clean the filesystem if attachments are moved there!
  $dbh->bz_start_transaction;
  $dbh->do('DELETE FROM attachments WHERE submitter_id = ?', undef, $spammer_id);
  $dbh->bz_commit_transaction();

  # _completely_ remove any flags the spammer is on
  $dbh->bz_start_transaction;
  $dbh->do('DELETE FROM flags WHERE setter_id = ? OR requestee_id = ?',
    undef, $spammer_id, $spammer_id);
  $dbh->bz_commit_transaction();

  return;
}

1;
