# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::WebService::Bugzilla;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Bug;
use Bugzilla::BugMail;
use Bugzilla::Constants;
use Bugzilla::Group;
use Bugzilla::Util;
use Bugzilla::WebService::Util qw(validate);
use Bugzilla::Error;

use JSON::XS;
use Data::Dumper;
use URI::Escape;
use List::MoreUtils qw(any none);
use Bugzilla::Token qw(issue_hash_token);
use Bugzilla::Report;
use List::Compare;
use Date::Format;

use constant PUBLIC_METHODS => qw(
  product_get_objects
  product_update_objects
  product_create_objects
  product_delete_objects
  search_as
  can_delete_user
  delete_user
  terminate_sessions
  admin_apikeys
  edit_comment
  manage_user_access
  group_get_products
  group_copy_products
  user_search
  run_report
  update_report
  delete_report
  delete_jobs
  view_jobs
  tag_user_comments_as_spam
  reschedule_jobs
);

=pod

=head1 NAME

Bugzilla::Extension::RedHat::WebService::Bugzilla - Red Hat specific XMLRPC methods

=head1 DESCRIPTION

This module provides methods that are Red Hat customisations.

Method calls are normally in the form of namespace.method(arguments).
Where namespace in this case is "RedHat". So for example
RedHat.getBug(bugid, username, password) would return the results
of getBug() for bug id $bugid.

=head1 METHODS

=head2 product_create_objects

Used to create a milestone, release or version object in a product or products

Pass in a single hash named 'data'.  This hash must contain an array of hashes
for each product that will have milestones, releases or versions created against
it

Below is the keys to be included in the array of hashes.

=over

=item name

B<string> The name of the product that will be modified. Either the name or ID of the product must be specified

=item id

B<integer> The ID of the product that will be modified. Either the name or ID of the product must be specified

=item versions

B<array-of-hashes> I<optional> An array of hashes describing each milestone to be created.
The arguments for the versions hash is detailed below

=over

=item name

B<string> The name of the version

=back

=item milestones

B<array-of-hashes> I<optional> An array of hashes describing each milestone to be created.
The arguments for the versions hash is detailed below

=over

=item name

B<string> The name of the milestone

=item sortkey

B<integer> I<optional> The sort order of the milestone

=item default

B<boolean> I<optional> True if you wish this to be the new default milestone

=back

=item releases

B<array-of-hashes> I<optional> An array of hashes describing each release to be created.
The arguments for the release hash is detailed below

=over

=item name

B<string> The name of the release

=item sortkey

B<integer> I<optional> The sort order of this release

=item default

B<boolean> I<optional> True if you want this to be the new default

=back

=back

=cut

sub product_create_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $changes;

  my $data = $params->{data}
    // ThrowCodeError('param_required', {param => 'data'});

  foreach my $product_hash (@$data) {
    my $product;
    if (defined($product_hash->{name})) {
      $product = Bugzilla::Product->check({name => $product_hash->{name}});
    }
    elsif (defined($product_hash->{id})) {
      $product = Bugzilla::Product->new($product_hash->{id});
    }
    else {
      ThrowCodeError('params_required', {params => ['name', 'id']});
    }

    my $product_changes = {name => $product->name, id => $product->id,};

    $user->check_can_admin_product($product->name);

    foreach my $milestone_hash (@{$product_hash->{milestones}}) {
      my $name = $milestone_hash->{name}
        // ThrowCodeError('param_required', {param => 'name'});
      my $sortkey = $milestone_hash->{sortkey} // 0;

      my $milestone
        = Bugzilla::Milestone->create({
        value => $name, sortkey => $sortkey, product => $product
        });

      if (defined($milestone_hash->{default}) && $milestone_hash->{default}) {
        $product->set_default_milestone($milestone->name);
        $product->update;
      }

      push @{$product_changes->{milestones}},
        {
        id      => $milestone->id,
        name    => $milestone->name,
        sortkey => $milestone->sortkey,
        };
    }

    foreach my $release_hash (@{$product_hash->{releases}}) {
      my $name = $release_hash->{name}
        // ThrowCodeError('param_required', {param => 'name'});
      my $sortkey = $release_hash->{sortkey} // 0;

      my $release
        = Bugzilla::Release->create({
        value => $name, sortkey => $sortkey, product => $product
        });

      if (defined($release_hash->{default}) && $release_hash->{default}) {
        $product->set_default_release($release->name);
        $product->update;
      }

      push @{$product_changes->{releases}},
        {id => $release->id, name => $release->name, sortkey => $release->sortkey};
    }

    foreach my $version_hash (@{$product_hash->{versions}}) {
      my $name = $version_hash->{name}
        // ThrowCodeError('param_required', {param => 'name'});

      my $version = Bugzilla::Version->create({value => $name, product => $product,});

      push @{$product_changes->{versions}},
        {id => $version->id, name => $version->name,};
    }

    push @{$changes->{data}}, $product_changes;
  }

  $dbh->bz_commit_transaction;

  return $changes;
}


# RedHat.product_update
# Update information about a products target release, target milestone, target_version

=head2 product_update_objects

Update a set of product milestones, releases and version

=over

=item versions

B<array> I<optional> A hash named versions contaning an array of hashes that describes each version to be updated.
The arguments for each version hash are detailed below

=over

=item id

B<integer> The ID of the version to be updated.

=item name

B<string> I<optional> The new name of the version.

=item is_active

B<boolean> I<optional> True to enable the version, False to disable.

=back

=item milestones

B<array> I<optional> An array of hashes describing each milestone to be updated.  One hash per milestone.
The arguments for each version hash are detailed below

=over

=item id

B<integer> The ID of the milestone to be updated

=item name

B<string> I<optional> The new name of the milestone

=item sortkey

B<integer> I<optional> The sort order of the milestone

=item default

B<boolean> I<optional> set to true to make this the new default milestone.

=item is_active

B<boolean> I<optional> True to enable the milestone, False to disable.

=back

=item releases

B<array> I<optional> An array of hashes describing each release to be updated.  One hash per release
The arguments for each hash are detailed below

=over

=item id

B<integer> The ID of the release to be updated

=item name

B<string> I<optional> The new name of the release

=item sortkey

B<integer> I<optional> The sort order of this release

=item default

B<boolean> I<optional> Set to true to make this the default release.

=item is_active

B<boolean> I<optional> True to enable the release, False to disable.

=back

=back

=cut

sub _name_remap {
  my ($field) = @_;

  if ($field eq 'name') {
    return 'value';
  }
  elsif ($field eq 'is_active') {
    return 'isactive';
  }
  else {
    return $field;
  }
}

sub product_update_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $changes       = {};
  my $product_cache = {};

  foreach my $version_data (@{$params->{versions}}) {
    my $version_id = int $version_data->{id}
      // ThrowCodeError('param_required', {param => 'id'});
    my $version = Bugzilla::Version->check({id => $version_id});

    _check_product($user, $version->product_id, $product_cache);

    foreach my $field (qw(name is_active)) {
      next unless defined $version_data->{$field};

      # Bugzilla 5.0 has changed the setter name on the Bugzilla::Version object.
      # In 4.4 it was set_name, in 5.0 it is set_value.  We need to remap 'name'
      # to 'value' so this API will continue to work. Same for is_active -> isactive
      my $method = 'set_' . _name_remap($field);
      $version->$method($version_data->{$field});
    }

    my $vs_changes = $version->update;

    push @{$changes->{versions}->{$version_id}}, $vs_changes;
  }
  foreach my $milestone_data (@{$params->{milestones}}) {
    my $milestone_id = int $milestone_data->{id}
      // ThrowCodeError('param_required', {param => 'id'});
    my $milestone = Bugzilla::Milestone->check({id => $milestone_id});

    _check_product($user, $milestone->product_id, $product_cache);

    foreach my $field (qw(name is_active sortkey)) {
      next unless defined $milestone_data->{$field};
      my $method = "set_$field";
      $milestone->$method($milestone_data->{$field});
    }

    my $ms_changes = $milestone->update;

    if (defined($milestone_data->{default}) && $milestone_data->{default}) {
      my $product = Bugzilla::Product->check({id => $milestone->product_id});
      $product->set_default_milestone($milestone->name);
      $product->update;
      $ms_changes->{default} = 1;
    }


    push @{$changes->{milestones}->{$milestone_id}}, $ms_changes;
  }
  foreach my $release_data (@{$params->{releases}}) {
    my $release_id = int $release_data->{id}
      // ThrowCodeError('param_required', {param => 'id'});
    my $release = Bugzilla::Release->check({id => $release_id});

    _check_product($user, $release->product_id, $product_cache);

    foreach my $field (qw(name is_active sortkey)) {
      next unless defined $release_data->{$field};
      my $method = "set_$field";
      $release->$method($release_data->{$field});
    }

    if (defined($release_data->{default}) && $release_data->{default}) {
      my $product = Bugzilla::Release->check({id => $release->product_id});
      $product->set_default_release($release->name);
      $product->update;
    }

    my $rs_changes = $release->update;

    push @{$changes->{releases}->{$release_id}}, $rs_changes;
  }


  $dbh->bz_commit_transaction;

  return $changes;
}


# Fetch information about target release, target milestone, versions for a list
# of products or product ids.
# { names => ['Foo', 'Bar', 'Baz'], ids => [1, 2, 3, 4, 5] }

=head2 product_get_objects

Returns all the versions, milestones and releases for the specified products

Arguments to this function are a list of product names and/or product ids

=head3 Params

=over

=item names

B<array-of-strings> I<optional> An array of strings of the product names.

=item ids

B<array-of-integers> I<optional> An array of product ids

=back

=head3 Returns

Returns an array of hashes, one per product describing each version, milestone and component
in the product

=over

=item name

B<string> The product name

=item id

B<integer> The product id

=item versions

A hash named versions that contains an array of hashes describing each
version in the product

=over

=item name

B<string> The version name

=item id

B<integer> The version ID

=item is_active

B<integer> 1 if the version is active, 0 if inactive

=back

=item milestones

A hash named milestones that contains an array of hashes describing each
milestone in the product

=over

=item name

B<string> The milestone name

=item id

B<integer> The milestone ID

=item is_active

B<integer> 1 if the milestone is active, 0 if inactive

=item default

B<integer> 1 if this milestone is the default.  Will not be present if
this is not the default

=back

=item releases

=over

=item name

B<string> The release name

=item id

B<integer> The release ID

=item is_active

B<integer> 1 if the release is active, 0 if inactive

=item default

B<integer> 1 if this release is the default.  Will not be present if
this is not the default

=back

=back

=cut

sub product_get_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  # Convert product ids/names to product objects
  my @product_objs;
  if (defined($params->{names})) {
    push @product_objs,
      map { Bugzilla::Product->check({name => $_}) } @{$params->{names}};
  }
  if (defined($params->{ids})) {
    push @product_objs, @{Bugzilla::Product->new_from_list($params->{ids})};
  }

# For each product, get information about target_release, target_milestone, versions
  my $product_set = [];
  foreach my $product (@product_objs) {

    # Check that they can admin the products.
    $user->check_can_admin_product($product->name);

    my $product_data = {};

    # Target Milestone
    my $milestones = [];
    foreach my $milestone_obj (@{$product->milestones}) {
      my $milestone = {
        id        => $milestone_obj->id,
        name      => $milestone_obj->name,
        sortkey   => $milestone_obj->sortkey,
        is_active => $milestone_obj->is_active,
      };
      if ($product->default_milestone eq $milestone_obj->name) {
        $milestone->{default} = 1;
      }
      push @$milestones, $milestone;
    }

    # Target Release
    my $releases = [];
    foreach my $release_obj (@{$product->releases}) {
      my $release = {
        id        => $release_obj->id,
        name      => $release_obj->name,
        sortkey   => $release_obj->sortkey,
        is_active => $release_obj->is_active,
      };
      if ($product->default_release eq $release_obj->name) {
        $release->{default} = 1;
      }
      push @$releases, $release;
    }

    # Versions
    my $versions = [];
    foreach my $version (@{$product->versions}) {
      push @$versions,
        {
        id        => $version->id,
        name      => $version->name,
        is_active => $version->is_active
        };
    }

    $product_data->{milestones} = $milestones if scalar(@$milestones);
    $product_data->{releases}   = $releases   if scalar(@$releases);
    $product_data->{versions}   = $versions   if scalar(@$versions);

    $product_data->{name} = $product->name;
    $product_data->{id}   = $product->id;

    push @$product_set, $product_data;
  }

  return $product_set;
}

=head2 product_delete_objects

Delete a version, milestone or release.

=head3 Params

=over

=item versions

B<array> I<optional> An array of version ids that should be deleted.

=item milestones

B<array> I<optional> An array of milestone ids that should be deleted.

=item releases

B<array> I<optional> An array of release ids that should be deleted.

=back

=cut

sub product_delete_objects {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar @{$user->get_products_by_permission('editcomponents')}
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'components'});

  my $changes;
  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $product_cache = {};

  foreach my $version_id (@{$params->{versions}}) {
    my $version = Bugzilla::Version->check({id => $version_id});

    _check_product($user, $version->product_id, $product_cache);

    push @{$changes->{versions}}, $version->id;

    $version->remove_from_db;
  }
  foreach my $milestone_id (@{$params->{milestones}}) {
    my $milestone = Bugzilla::Milestone->check({id => $milestone_id});

    _check_product($user, $milestone->product_id, $product_cache);

    push @{$changes->{milestones}}, $milestone->id;

    $milestone->remove_from_db;
  }
  foreach my $release_id (@{$params->{releases}}) {
    my $release = Bugzilla::Release->check({id => $release_id});

    _check_product($user, $release->product_id, $product_cache);

    push @{$changes->{releases}}, $release->id;

    $release->remove_from_db;
  }

  $dbh->bz_commit_transaction;

  return $changes;
}

sub _check_product {
  my ($user, $product_id, $product_cache) = @_;

  unless (defined($product_cache->{$product_id})) {
    my $product = Bugzilla::Product->check({id => $product_id});
    $user->check_can_admin_product($product->name);
    $product_cache->{$product->id} = 1;
  }

  return 1;
}


## REDHAT EXTENSION 1354240 END

## REDHAT EXTENSION START 1219299

=head2 search_as

Convert a URI query string in to an RPC call.

=head3 Params

=over

=item format

B<string> The format required, perl or json.

=item search

B<string> The search string to convert.

=back

=cut

sub search_as {
  my ($self, $params) = @_;

  #  RED HAT EXTENSION 1347097
  return (Bugzilla->search_as($params));
}
## REDHAT EXTENSION END 1219299

=head2 can_delete_user

Report if a user can be deleted and if not, why not.

=head3 Params

=over

=item names

B<array> of user names to check.

=back

=head3 Returns

Returns an array of hashes, one per user reported on. Each hash contains:

=over

=item name

B<string> The user name

=item id

B<int> The user name

=item can

B<bool> Can the user be deleted?

=item not_ok_to_delete

B<hash> A hash of counts of things that can prevent deletion.

=item ok_to_delete

B<hash> A hash of counts of things that cannot prevent deletion.

=back

=cut

sub can_delete_user {
  my $self = shift;
  my ($params) = @_;


  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "users"});

  my $email = trim($params->{names})
    || ThrowCodeError('param_required', {param => 'email'});

  my @user_objects = map { Bugzilla::User->check($_) } @{$params->{names}};

  my @users;

  foreach my $user (@user_objects) {
    push(@users, $user->can_delete());
  }

  return {users => \@users};
}


=head2 delete_user

Delete some users.

Users will only be deleted if the pass L<can_delete_user>.

=head3 Params

=over

=item names

B<array> The user names to delete.

=back

=head3 Returns

Returns an array of hashes, one per user deleted on.

=cut

sub delete_user {
  my $self = shift;
  my ($params) = @_;


  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "users"});

  my $email = trim($params->{names})
    || ThrowCodeError('param_required', {param => 'email'});

  my @user_objects = map { Bugzilla::User->check($_) } @{$params->{names}};

  my @users;

  foreach my $user (@user_objects) {
    push(@users, $user->delete_me());
  }

  return {users => \@users};
}

=head2 terminate_sessions

Terminate sessions for users.

You must be in the I<admin> group to use this API.

=head3 Params

=over

=item sessions

B<array> of hashes with I<name> and I<ids> of sessions to terminate.

  {
    'sessions' =>
      [
        {
          'name' =>'fu@example.com',
          'ids'  => [11,56,78]
        },
        {
          'name' => 'bar@example.com',
          'ids'  => [34, 42]
        },
      ]
  }

=back

=head3 Returns

Returns a hash, with the user name as the key, which contains an array of sessions
terminated.

  {
    'terminated' =>
      {
        'fu@example.com'  => [11,56,78],
        'bar@example.com' => [34, 42],
      }
  }

=cut

sub terminate_sessions {
  my $self = shift;
  my ($params) = @_;

  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "terminate_sessions"});

  my $user_sessions = $params->{sessions}
    || ThrowCodeError('param_required', {param => 'sessions'});

  my %results;
  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  foreach my $user_session (@$user_sessions) {
    my @terminated;

    my $user_name = $user_session->{name}
      || ThrowCodeError('param_required', {param => 'name'});
    my $session_ids
      = $user_session->{ids} || ThrowCodeError('param_required', {param => 'ids'});

    my $user_object = Bugzilla::User->new({name => $user_name});

    my $sessions = Bugzilla::User::Session->new_from_list($session_ids);

    foreach my $session (@$sessions) {
      if ($session->userid == $user_object->id) {
        $session->remove_from_db;
        push(@terminated, $session->id);
      }
    }

    if (@terminated) {
      $results{'terminated'}->{$user_name} = \@terminated;
    }
  }
  $dbh->bz_commit_transaction;

  return (\%results);
}

=head2 admin_apikeys

Ban and revoke APIKEYs for users.

You must be in the I<admin> group to use this API.

=head3 Params

=over

=item apikeys

B<array> of hashes with I<name> and I<ids> of APIKEYs to ban or revoke.

  {
    'apikeys' =>
      [
        {
          'name'        => 'fu@example.com',
          'banned_ids'  => [11,56],
          'revoked_ids' => [78]
        },
        {
          'name'        => 'bar@example.com',
          'banned_ids'  => [34, 42]
        },
      ]
  }

=back

=head3 Returns

Returns an array of hashes for APIKEYs that have been banned or revoked.

  [
    {
      'name'        => 'fu@example.com',
      'banned_ids'  => [11,56],
      'revoked_ids' => [78]
    },
    {
      'name'       => 'bar@example.com',
      'banned_ids' => [34, 42],
    }
  ]

=cut

sub admin_apikeys {
  my $self = shift;
  my ($params) = @_;

  Bugzilla->user->in_group('editusers')
    || ThrowUserError("auth_failure",
    {group => "editusers", action => "run", object => "ban_apikeys"});

  my $user_apikeys = $params->{apikeys}
    || ThrowCodeError('param_required', {param => 'apikeys'});

  my @results;
  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  foreach my $user_apikey (@$user_apikeys) {
    my @banned;
    my @revoked;


    my $user_name
      = $user_apikey->{name} || ThrowCodeError('param_required', {param => 'name'});
    my $banned_ids  = $user_apikey->{banned_ids};
    my $revoked_ids = $user_apikey->{revoked_ids};

    ThrowCodeError('params_required', {params => ['banned_ids', 'revoked_ids']})
      unless (@$banned_ids || @$revoked_ids);

    my %changes = ('name' => $user_name,);

    my $user_object = Bugzilla::User->new({name => $user_name});

    my $apikeys = Bugzilla::User::APIKey->new_from_list($banned_ids);

    foreach my $apikey (@$apikeys) {
      if ($apikey->user_id == $user_object->id) {
        $apikey->set_banned(1);
        $apikey->update();
        push(@banned, $apikey->id);
      }
      else {
        ThrowUserError('api_key_ban_failed',
          {user_name => $user_name, apikey_id => $apikey->id});
      }
    }

    if (@banned) {
      $changes{'banned'} = \@banned;
    }

    $apikeys = Bugzilla::User::APIKey->new_from_list($revoked_ids);

    foreach my $apikey (@$apikeys) {
      if ($apikey->user_id == $user_object->id) {
        $apikey->set_revoked(1);
        $apikey->update();
        push(@revoked, $apikey->id);
      }
      else {
        ThrowUserError('api_key_ban_failed',
          {user_name => $user_name, apikey_id => $apikey->id});
      }
    }

    if (@revoked) {
      $changes{'revoked'} = \@revoked;
    }

    if (@banned || @revoked) {
      push(@results, \%changes);
    }
  }

  $dbh->bz_commit_transaction;

  return (\@results);
}

## REDHAT EXTENSION START 1404885

=head2 edit_comment

Edit a comment on a bug.

The rules for this are

=over 4

=item A security or admin user can edit any comment.

=item A security or admin user can hide previous comment revisions when editing a
comment.

=item A reporter may edit the description, comment 0, on their bugs.

=item A user with I<editbugs> privileges on a product may edit the description comment
of bugs in that product.

=back

When an edit is made an entry is added to the bugs_activity table indicating who
made the change and what comment number was updated. The text of the change is not
included in this entry. When a user views the bug history, either on the bug
history page or when using the inline history feature, they will be able to see
who edited a comment and when, but not the change made.

The texts are stored in a separate table and when entries from this table are
viewed the ACLs are applied. If you lack the authorization to see a change then
the text will be replaced with the string I<REDACTED>.

The security flag is retroactive, it hides all previous revisions of a comment.

Ths feature is accessed in the UI on the bug edit page by clicking the edit icon
in the comment header. If you cannot see this icon on a comment then you do not
have permission to edit that comment.

The security flag is set in the UI by selecting the I<Hide history> checkbox on
the edit form. If you cannot see this checkbox then you do not have access to lock the
comment history.

If you try and edit a comment or set the security flag and you do not have access
to do so the API will throw an error and reject the entire request.

=head3 Params

comment:
  B<int> I<optional> The comment ID. This or bug are required.

bug:
  B<int> I<optional> The bug ID. This or comment are required.

number:
  B<int> I<optional> The comment number on a bug to update.
  This is only used if bug is supplied.
  Defaults to zero.

text:
  B<string> The text to change the comment to.

security:
  B<boolean> I<optional> Hide previous versions of this comment due to security concerns.
  B<Note:> You need to be in the admin or security groups to set this parameter

minor_update:
  B<boolean> I<optional> If set to true, this is considered a minor update and no
  mail is sent to users who do not want minor update emails.

render:
  B<boolean> I<optional> Return the rendered HTML for this comment.

=head3 Returns

Returns a hash, containing:

activity_count:
  B<int> The number of times the comment has been edited.

delta_ts:
  B<bool> The new last changed time stamp for the bug.

longdesc:
  B<array> A array of 2 things. The first is the comment number on the bug. The
  second is what happened to the comment. Currently only 'updated'.
  This text will appear in the bug history and inline history if enabled.

html:
  B<string> If requested the HTML formatted version of the comment.
  This is what a user will see when loading the updated comment in the web UI.

=head3 Examples

=head4 JSONRPC call

=over 4

=item By comment id

  {
      "jsonrpc": "2.0",
      "method": "RedHat.edit_comment",
      "id": 2,
      "params": [
        {
          "comment": 13178344,
          "render": 1,
          "security": 0,
          "text": "Description of problem:\nThe old comment is just no good!\nBug 11\n",
        }
      ]
  }

=item By bug id & comment number

  {
      "jsonrpc": "2.0",
      "method": "RedHat.edit_comment",
      "id": 3,
      "params": [
        {
          "bug": 42,
          "comment": 3,
          "text": "Description of problem:\nExamples are hard\n",
        }
      ]
  }

=back

=head4 JSONRPC reply

=over 4

=item By comment id

  {
      "id": 2,
      "error": null,
      "result": {
          "activity_count": 2,
          "delta_ts": "2020-03-06 03:38:50",
          "html": "Description of problem:\nThe old comment is just no good!\n<a href=\"https://bugzilla.redhat.com/show_bug.cgi?id=11\">https://bugzilla.redhat.com/show_bug.cgi?id=11</a>\n"
          "longdesc": [0, "updated"],
      }
  }

=item By bug id & comment number

  {
      "id": 3,
      "error": null,
      "result": {
          "activity_count": 4,
          "delta_ts": "2020-03-06 03:38:50",
          "longdesc": [3, "updated"],
      }
  }

=back

=cut

sub edit_comment {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $text         = $params->{text} || die "no text";
  my $security     = $params->{security}     // 0;
  my $number       = $params->{number}       // 0;
  my $render       = $params->{render}       // 0;
  my $minor_update = $params->{minor_update} // 0;

  if ($security && !$user->can_edit_comments()) {
    ThrowUserError('only_security_can_hide_comment_history');
  }

  $security = $security == 1 ? 1 : 0;

  my $bug_id     = $params->{bug};
  my $comment_id = $params->{comment};

  if (!$bug_id && !$comment_id) {
    ThrowCodeError('params_required', {params => ['bug', 'comment']});
  }

  my ($comment, $bug);

  if ($comment_id) {
    $comment = Bugzilla::Comment->check({id => $comment_id});
    $bug     = $comment->bug;
  }
  else {
    $bug     = Bugzilla::Bug->check($bug_id);
    $comment = $bug->comments({order => 'oldest_to_newest'})->[$number];

    unless ($comment) {
      ThrowUserError('no_such_comment_number',
        {number => $number, bug_id => $bug->id});
    }

    if ($comment->count() != $number) {
      ThrowUserError('comment_number_mismatch',
        {requested => $number, actual => $comment->count()});
    }
  }

  $bug->check_is_visible();
  $bug->edit_comment($comment->id, $text, $security, $comment->count());

  my ($changes, $delta_ts, $old_bug) = $bug->update();
  $changes->{delta_ts} = $bug->delta_ts;
  delete $comment->{'activity'};

  # skip mail for security changes
  if ($security) {
    Bugzilla->dbh->do('UPDATE bugs SET lastdiffed = ? WHERE bug_id = ?',
      undef, ($delta_ts, $bug->id));
  }
  else {
    $comment->{'thetext'} = $text;
    $bug->send_changes($changes, {edited_comment => $comment}, $minor_update);
  }

  $changes->{activity_count} = scalar @{$comment->activity} - 1;

  if ($render) {
    my $html = Bugzilla::Template::quoteUrls($text, $bug);
    $changes->{html} = $html;
  }

  return ($changes);
}
## REDHAT EXTENSION END 1404885

## REDHAT EXTENSION START 1821023

=head2 manage_user_access

Manage a user's account access.

B<WARNING:> The actions you take here will temporarily prevent a user from
accessing their account, and can permanently disable the user's account and make
I<ALL> the bugs they opened inaccessible. Ensure you have read this documentation
carefully and are sure you know what you are doing before using this functionality.

=head3 Things to check

Before managing a user's access you might wish to do one or more of the following:

View the account history:
  If the user has groups or has notable changes to their privileges it might give
  an indication of the accounts status.

View bug activity:
  This might indicate the account is doing constructive things that may affect
  which action you select.

View administration activity:
  This might indicate the account is doing administration activity that may affect
  which action you select.

=head3 Params

names:
  B<array> The user names to manage. If a name does not have an account it will be
  skipped and a message will be added to the response stating this occurred.

action:
  B<string> What action to take on the accounts. See "The values for 'action' are:"
  below for the accepted values and what they achieve.

replacement:
  B<string> Account to use to replace the user where required. If this name does
  not have an account then the nobody account will be used and a message will be
  added to the response stating this occurred.

message:
  B<string> Message to send to user, to put in to the disabled text field, and to
  enter as a bug comment. This is included in mails sent to the affected user. If
  you don't supply a message then a default message naming the user who made the
  change is used.

=head3 The values for 'action' are:

=head4 reconfirm

This action forces the user to reconfirm their access by using the forgot password
process to reset their password. It takes the following actions:

=over 4

=item Changes their password to a random unknown value

=item Terminates all their sessions

=item Revokes all their api_keys

=item Mails the user about the change

=back

=head4 revoke

This action revokes access, it locks the user's account and forces them to contact
the site administrators to have their account re-enabled. It takes the following
actions:

=over 4

=item Changes their password to a random unknown value

=item Terminates all their sessions

=item Revokes all their api_keys

=item Disables bug mail

=item Sets disabled text

=item Mails the user about the change

=back

=head4 remove

This action terminates the user's account. It takes the following actions:

=over 4

=item Changes their password to a random unknown value

=item Terminates all their sessions

=item Bans all their api_keys

=item Disables bug mail

=item Sets disabled text

=item Removes the user from all groups

=item Removes bless from all groups

=item Removes the user from all product responsibilities. Replaces them with the replacement if specified.

=item Replaces the user with the default contact for assignee/qa/docs on all open bugs

=item Removes user from CC on all open bugs

=back

B<Note:> Bugs are updated using the job queue so happen after the RPC call has 
been completed.

=head4 redact

This action terminates the user's account as per L</remove> above and obfuscates
their name and login. It is B<not reversible> and the account cannot be recovered.
It takes the following actions:

=over 4

=item All actions from L</remove>

=item Obfuscates user's name and login

=back

B<Note:> Bugs are updated using the job queue so happen after the RPC call has 
been completed.

=head4 repudiate

This action terminates the user's account as per L</remove> above, it then moves
B<every> bug the user has ever created in to the Spam product. It tags all
comments the user has ever made with 'spam' and it edits all comments with a
standard replacement text. It removes all CC's and users from roles on these
bugs and makes them inaccessible to most users. No mail will be sent for these
bug changes.

It is B<not reversible> and the account cannot be recovered. It takes the following
actions:

=over 4

=item All actions from L</remove>

=item Moves and locks all bugs the user has created.

=item Marks all comments the user has made as spam.

=item Removes all outstanding flag requests the user made.

=item Removes the user from CC on all bugs.

=item Deletes all attachments the user created.

=back

B<Note:> Bugs are updated using the job queue so happen after the RPC call has 
been completed.

=head3 Returns

Returns an array of hashes, one per user acted on, indexed on the account ID.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.manage_user_access",
      "id": 11,
      "params": [
        {
          "names": ["fu@example.com"],
          "action": "repudiate"
          "message": "No soup for you!",
        }
      ]
  }

=head4 JSONRPC reply

  {
      "error": null,
      "result": {
          "users": {
              "42": {
                  "account_changes": {
                    "cryptpassword": ["", "hashed_with_SHA-256"],
                    "disable_mail":  [0, 1],
                    "disabledtext":  ["", "No soup for you!"]
                    "groups":        ["", "fedora_contrib,packager",
                    "is_enabled":    [1, 0],
                  },
                  "counts": {
                      "bugs": 0,
                      "components": 1,
                      "keys_banned": 2,
                      "keys_revoked": 0,
                      "sessions": 1
                      "spam_bugs": 3,
                  },
                  "messages": [
                    "User has been repudiated"
                  ]
              }
          }
      },
      "id": 11
  }

=cut

sub manage_user_access {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  defined($params->{names})
    || ThrowCodeError('param_required', {param => 'names'});


  defined($params->{action})
    || ThrowCodeError('param_required', {param => 'action'});

  my %args;
  $args{names}       = $params->{names};
  $args{action}      = $params->{action};
  $args{replacement} = $params->{replacement} if (defined $params->{replacement});
  $args{message}     = $params->{message}     if (defined $params->{message});

  return $user->manage_user_access(\%args);
}
## REDHAT EXTENSION END 1821023

## REDHAT EXTENSION START 1699171

=head2 group_get_products

Get a list of all product ACLs for groups.

You must be in the I<editcomponents> group to use this API.

=head3 Params

=over

=item groups

B<array> of group names to fetch.

=back

=head3 Returns

Returns a hash, with the group name as the key, which contains a hash of products
with the group controls.

=head3 Examples

=head4 JSONRPC call

 {
      "jsonrpc": "2.0",
      "method": "RedHat.group_get_products",
      "id": 11,
      "params": [
        {
          "groups": ["group_1", "group_2"]
        }
      ]
  }

=head4 JSONRPC reply

  "groups": {
    "group_1": {
      "product fu": {
        "controls": {
          "canconfirm": 0,
          "canedit": 0,
          "editbugs": 1,
          "editcomponents": 0,
          "entry": 0,
          "membercontrol": 0,
          "othercontrol": 0
        }
      }
    },
    "group_2": {
      "product foo": {
        "controls": {
          "canconfirm": 1,
          "canedit": 0,
          "editbugs": 1,
          "editcomponents": 0,
          "entry": 0,
          "membercontrol": 0,
          "othercontrol": 0
        }
      }
    }
  }

=cut

sub group_get_products {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'groups'});

  defined($params->{groups})
    || ThrowCodeError('param_required', {param => 'groups'});

  my %results;

  foreach my $gname (@{$params->{groups}}) {
    my $group = Bugzilla::Group->check({name => $gname});
    my %grp_acls;

    foreach my $record (@{$group->products}) {
      my $prod = $record->{product}->{name};
      delete $record->{product};
      $grp_acls{$prod} = $record;
    }

    $results{$group->name} = \%grp_acls;
  }

  return ({groups => \%results});
}

=head2 group_copy_products

Copy the product ACLs from one group to another group.

You must be in the I<editcomponents> group to use this API.

This method will not remove ACLs, it will only add ACLs missing from the I<to> group.

=head3 Params

=over

=item from

B<string> of the group name to copy ACLs from.

=item to

B<string> of the group name to copy ACLs to.

=back

=head3 Returns

Returns an array of hashes, with the product name as the key, which contains a
hash of groups control changes.

=head3 Examples

=head4 JSONRPC call

 {
      "jsonrpc": "2.0",
      "method": "RedHat.group_copy_products",
      "id": 11,
      "params": [
        {
          "from": "group_1",
          "to": "group_2"
        }
      ]
  }

=head4 JSONRPC reply

  {
    "changes": [
      {
        "product 1": {
            "group_controls": {
              "group_2": {
                "editbugs": {
                  "added": "1",
                  "removed": "0"
                }
              }
            }
        },
        "id": 35
      }
    ]
  }

=cut

sub group_copy_products {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || ThrowUserError('auth_failure',
    {group => 'editcomponents', action => 'edit', object => 'groups'});


  (defined($params->{from}) && defined($params->{from}))
    || ThrowCodeError('params_required', {params => ['from', 'to']});

  my $from_group = Bugzilla::Group->check({name => $params->{from}});
  my $to_group   = Bugzilla::Group->check({name => $params->{to}});

  my @product_objects;

  my $dbh = Bugzilla->dbh;

  $dbh->bz_start_transaction();
  foreach my $record (@{$from_group->products}) {
    my $prod = $record->{product};

    my %args;
    my @fields
      = qw/entry membercontrol othercontrol canedit editcomponents editbugs canconfirm/;

    foreach my $field (@fields) {
      $args{$field} = $record->{controls}->{$field};
    }

    $prod->set_group_controls($to_group, \%args);
    push(@product_objects, $prod);
  }

  my %changes;
  foreach my $product (@product_objects) {
    my $returned_changes = $product->update();
    $changes{$product->id} = $returned_changes;
  }
  $dbh->bz_commit_transaction();

  my @result;
  foreach my $product (@product_objects) {
    my %hash    = (id => $product->id,);
    my $changed = 0;

    foreach my $field (keys %{$changes{$product->id}}) {
      my $change = $changes{$product->id}->{$field};
      if ($field eq 'group_controls') {
        $changed = 1;
        foreach my $grp (keys %{$change}) {
          foreach my $cntrl (keys %{$change->{$grp}}) {
            $hash{$product->name}{$field}->{$grp}->{$cntrl} = {
              removed => $self->type('string', $change->{$grp}->{$cntrl}->[0]),
              added   => $self->type('string', $change->{$grp}->{$cntrl}->[1])
            };
          }
        }
      }
      else {
        $changed = 1;
        $hash{$product->name}{$field} = {
          removed => $self->type('string', $change->[0]),
          added   => $self->type('string', $change->[1])
        };
      }
    }

    push(@result, \%hash) if ($changed);
  }

  return {changes => \@result};
}

## REDHAT EXTENSION END 1699171

## REDHAT EXTENSION START 1426216

=head2 user_search

A fast way to search for users.

You must be logged in to use this API.

The number of results is limited to the site's max user matches parameter.

=head3 Params

=over

=item match

An B<array> of B<strings> to search for.

These will be compared in order to login and real name.

=back

=head3 Returns

Returns an array of hashes, with the user details.

=head3 Examples

=head4 JSONRPC call

 {
      "jsonrpc": "2.0",
      "method": "RedHat.user_search",
      "id": 11,
      "params": [
        {
          "match": ["dude"],
        }
      ]
  }

=head4 JSONRPC reply

  {
    "users": [
      {
        "id": 11,
        "name": "dude@example.com",
        "real_name": "Dude McDude",
        "email": "dude@example.com",
        "insider": 0
      }
    ]
  }

=cut

sub user_search {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  defined $params->{match}
    || ThrowCodeError('params_required',
    {function => 'RedHat.user_search', params => ['match']});

  my $dbh = Bugzilla->dbh;

  # This SQL tells you if the user is in any group that is an ancestor of the
  # insider group, and thus is an insider even if they aren't directly in the
  # insider group.
  my $sql = <<'EOS';
SELECT userid, login_name, realname, is_enabled,
  (SELECT (count(*) >= 1)
    FROM user_group_map
    WHERE user_id = pf1.userid and group_id IN (
      SELECT id FROM groups
      WHERE id IN (SELECT id FROM groups WHERE name = $2)
      OR id IN (
        WITH RECURSIVE ancestors AS (
          SELECT member_id
            FROM group_group_map
          WHERE grantor_id = (SELECT id FROM groups WHERE name = $2)
          UNION
            SELECT e.member_id
            FROM   group_group_map e
            INNER JOIN ancestors s ON s.member_id = e.grantor_id
        ) SELECT * FROM ancestors
      )
    )
  ) as is_insider,
  (SELECT (count(*) >= 1)
    FROM user_group_map
    WHERE user_id = pf1.userid and group_id IN (
      SELECT id FROM groups
      WHERE id IN (SELECT id FROM groups WHERE name = 'all_partners')
      OR id IN (
        WITH RECURSIVE ancestors AS (
          SELECT member_id
            FROM group_group_map
          WHERE grantor_id = (SELECT id FROM groups WHERE name = 'all_partners')
          UNION
            SELECT e.member_id
            FROM   group_group_map e
            INNER JOIN ancestors s ON s.member_id = e.grantor_id
        ) SELECT * FROM ancestors
      )
    )
  ) as is_partner
FROM profiles pf1
WHERE is_enabled = 1
AND (login_name ilike $1 OR realname ilike $1)
ORDER BY SIMILARITY(pf1.login_name, $3) + SIMILARITY(pf1.realname, $3) desc
LIMIT $4
EOS

  my $str = join '%', @{$params->{match}};
  trick_taint($str);

  my $ilike_str = q{%} . $str . q{%};

  my $users
    = $dbh->selectall_arrayref($sql, undef, $ilike_str,
    Bugzilla->params->{insidergroup},
    $str, Bugzilla->params->{maxusermatches});

  my @results;

  foreach my $match (@$users) {
    my $user_info = {
      id        => $self->type('int',     $match->[0]),
      name      => $self->type('email',   $match->[1]),
      real_name => $self->type('string',  $match->[2]),
      email     => $self->type('email',   $match->[1]),
      active    => $self->type('boolean', $match->[3]),
      insider   => $self->type('boolean', $match->[4]),
    };

    if ($user->is_insider()) {
      $user_info->{partner} = $self->type('boolean', $match->[5]);
    }

    push(@results, $user_info);
  }

  return {'users' => \@results};

}
## REDHAT EXTENSION END 1426216

## REDHAT EXTENSION START 1676959

=head2 run_report

Run a saved report and return the CSV for it.

=head3 Params

name:
  B<string> The name of the saved report.

bot_owner:
  B<string> The name of the bot that owns the report. If not supplied it will use
  the account of the user requesting the run.

  You must have management rights to the bot to set this.

=head3 Returns

Returns a hash containing the CSV output of the report.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.run_report",
      "id": 11,
      "params": [
        {
          "name": "MyImportantReport"
        }
      ]
  }

=head4 JSONRPC reply

  {
      "error": null,
      "result": {
          "report": '"Component": "Bugzilla General""Target Release / Status","ASSIGNED","ON_DEV","ON_QA","VERIFIED"
" ---",0,0,3,1'
          }
      },
      "id": 11
  }

=cut

sub run_report {
  my $self = shift;
  my ($params) = @_;

  my $user  = Bugzilla->login(LOGIN_REQUIRED);
  my $owner = $user;

  defined $params->{name}
    || ThrowCodeError('params_required',
    {function => 'RedHat.run_report', params => ['name']});

  my $bot_owner = $params->{bot_owner} // undef;

  if ($bot_owner) {
    my $bot_user = Bugzilla::User->check($bot_owner);
    $bot_user->is_bot()              || ThrowUserError('not_a_bot');
    $user->can_manage_bot($bot_user) || ThrowUserError('cannot_manage_bot');
    $owner = $bot_user;
  }

  my $report
    = Bugzilla::Report->match({name => $params->{name}, user_id => $owner->id});

  if (scalar @$report == 0) {
    ThrowUserError('no_match');
  }

  if (scalar @$report > 1) {
    ThrowUserError('more_than_one_match', {count => scalar(@$report)});
  }

  my $url = _clean_params($report->[0]->query, qw/format action ctype/);
  $url .= '&format=table&action=wrap&ctype=csv';

  my $output = do {
    my $userenv = "LOCAL_USER=" . $owner->login;
    qx{$userenv ./report.cgi "$url"};
  };

  $output =~ s/^.*charset=UTF-8\r\n\r\n//s;

  return ({report => "$output"});
}

=head2 update_report

Update a saved report. You must either own a report or have management rights to
the bot that owns a report to use this API call.

=head3 Params

name:
  B<string> The name of the saved report.

bot_owner:
  B<string> The name of the bot that owns the report or whom you wish to give
  ownership to. You must have management rights to the bot to set this.

groups:
  B<array> The names of the groups that have view permissions. This allows the
  members of the named groups to run the report, it does not change what data the
  report can access. The list provided is set verbatim. If existing groups are
  missing from the list provided then they will be removed from the report.

=head3 Returns

Nothing.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.update_report",
      "id": 11,
      "params": [
        {
          "name": "MyImportantReport",
          "groups": ["private", "the_secret_group"],
          "bot_owner": "my_bot@bot.bugzilla.redhat.com"
        }
      ]
  }

=cut

sub update_report {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  defined $params->{name}
    || ThrowCodeError('params_required',
    {function => 'RedHat.update_report', params => ['name']});

  defined($params->{groups})
    || ThrowCodeError('param_required', {param => 'groups'});

  my $bot_owner = $params->{bot_owner} // undef;

  my $reports
    = Bugzilla::Report->match({name => $params->{name}, user_id => $user->id});

  if (scalar @$reports == 0) {
    ThrowUserError('no_match');
  }

  if (scalar @$reports > 1) {
    ThrowUserError('more_than_one_match', {count => scalar(@$reports)});
  }

  my $report = $reports->[0];

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my (@added, @removed, @new_groups, @cur_groups);

  @cur_groups = map { $_->id } @{$report->groups()} if (@{$report->groups()});
  @new_groups
    = map { Bugzilla::Group->check({name => $_})->id } @{$params->{groups}};

  my $lc  = List::Compare->new(\@cur_groups, \@new_groups);
  my @rem = $lc->get_unique;
  my @add = $lc->get_complement;

  if (@add) {
    foreach my $gid (@add) {
      my $group = Bugzilla::Group->new($gid);
      $dbh->do("INSERT INTO report_groups (report_id, group_id) VALUES (?,?)",
        undef, $report->id, $group->id);
      push(@added, $group->name);
    }
  }

  if (@rem) {
    foreach my $gid (@rem) {
      my $group = Bugzilla::Group->new($gid);
      $dbh->do("DELETE FROM report_groups WHERE report_id = ? AND group_id = ?",
        undef, ($report->id, $group->id));
      push(@removed, $group->name);
    }
  }

  $report->audit_log({'groups' => [\@removed, \@added]});

  if ($bot_owner) {
    my $bot_user = Bugzilla::User->check($bot_owner);
    $bot_user->is_bot()              || ThrowUserError('not_a_bot');
    $user->can_manage_bot($bot_user) || ThrowUserError('cannot_manage_bot');
    $dbh->do("UPDATE reports SET user_id = ? WHERE id = ?",
      undef, $bot_user->id, $report->id);
    $report->audit_log({'user_id' => [$user->id, $bot_user->id]});
  }

  $dbh->bz_commit_transaction;

  return;
}

sub _clean_params {
  my $url    = shift;
  my @params = shift;

  my $query = $url;
  my $start = '';

  # peel off start of URL if it's included
  if ($url =~ /^(.*\?)(.*)$/) {
    $start = $1;
    $query = $2;
  }

  my @parts = grep {
    my $param = $_;
    none { $param =~ /^$_=/ } @params
  } split('&', $query);

  return $start . join('&', @parts);
}

=head2 delete_report

Delete a saved report.

=head3 Params

name:
  B<string> The name of the saved report.

bot_owner:
  B<string> The name of the bot that owns the report. If not supplied it will use
  the account of the user requesting the run.

  You must have management rights to the bot to set this.

=head3 Returns

Nothing.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.delete_report",
      "id": 11,
      "params": [
        {
          "name": "MyImportantReport"
        }
      ]
  }

=cut

sub delete_report {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  my $uid  = $user->id;

  defined $params->{name}
    || ThrowCodeError('params_required',
    {function => 'RedHat.delete_report', params => ['name']});

  my $bot_owner = $params->{bot_owner} // undef;

  if ($bot_owner) {
    my $bot_user = Bugzilla::User->check($bot_owner);
    $bot_user->is_bot()              || ThrowUserError('not_a_bot');
    $user->can_manage_bot($bot_user) || ThrowUserError('cannot_manage_bot');
    $uid = $bot_user->id;
  }

  my $reports
    = Bugzilla::Report->match({name => $params->{name}, user_id => $uid});


  if (scalar @$reports == 0) {
    ThrowUserError('no_match');
  }

  if (scalar @$reports > 1) {
    ThrowUserError('more_than_one_match', {count => scalar(@$reports)});
  }

  my $report = $reports->[0];
  $report->remove_from_db();

  $user->flush_reports_cache;

  return;
}
## REDHAT EXTENSION END 1676959

## REDHAT EXTENSION START 1991345

=head2 delete_jobs

Delete jobs from the job queue.

You must be in the I<admin> group to use this API.

=head3 Params

ids:
  B<array-of-integers> The ids of the Jobs to delete.

=head3 Returns

A hash containing the number of jobs deleted.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.delete_jobs",
      "id": 11,
      "params": [
        {
          "ids": [11, 12, 42]
        }
      ]
  }

=head4 JSONRPC reply

  {
      "error": null,
      "result": {
          "deleted_count": 3
      },
      "id": 11
  }

=cut

sub delete_jobs {
  my $self = shift;
  my ($params) = @_;

  Bugzilla->user->in_group('admin')
    || ThrowUserError("auth_failure",
    {group => "admin", action => "run", object => "delete_jobs"});

  my $ids
    = $params->{ids}
    || ThrowCodeError('params_required',
    {function => 'RedHat.delete_jobs', params => ['ids']});

  my @ids = map { _clean_id($_) } @$ids;

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;
  my $count = $dbh->do(
    'DELETE FROM ts_job WHERE jobid IN (' . join(',', (('?') x scalar(@ids))) . ')',
    undef, @ids
  );
  $dbh->bz_commit_transaction;

  return {deleted_count => $count};
}

sub _clean_id {
  my $id = shift;
  detaint_natural($id) || ThrowUserError("invalid_id");
  return $id;
}
## REDHAT EXTENSION END 1991345


## REDHAT EXTENSION START 1991345

=head2 view_jobs

Delete jobs from the job queue.

You must be in the I<admin> group to use this API.

=head3 Params

ids:
  B<array-of-integers> The ids of the Jobs to delete.

=head3 Returns

A hash containing the number of jobs deleted.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.view_jobs",
      "id": 11,
      "params": [
        {
          "ids": [42]
        }
      ]
  }

=head4 JSONRPC reply

  {
      "error": null,
      "result": {
          "jobs": [{
            "type": "Bugzilla::Extension::RuleEngine::Job",
            "arg": {
                "bug_ids": [101],
                "delta_ts": "2021-10-20 04:38:17",
                "action": "update"
            },
            "jobid": 42
          }]
      },
      "id": 11
  }

=cut

sub view_jobs {
  my $self = shift;
  my ($params) = @_;

  Bugzilla->user->in_group('admin')
    || ThrowUserError("auth_failure",
    {group => "admin", action => "run", object => "view_jobs"});

  my $ids
    = $params->{ids}
    || ThrowCodeError('params_required',
    {function => 'RedHat.delete_jobs', params => ['ids']});

  my @results;

  foreach my $id (map { _clean_id($_) } @$ids) {
    my @jobs = Bugzilla->job_queue->list_jobs({
      funcname => [
        'Bugzilla::Job::Mailer',
        'Bugzilla::Job::BugMail',
        'Bugzilla::Extension::RedHat::Job::Termination',
        'Bugzilla::Extension::RedHat::Job::DeSpamComments',
        'Bugzilla::Extension::ExternalBugs::Job::SFDC',
        'Bugzilla::Extension::ExternalBugs::Job::JIRA',
        'Bugzilla::Extension::RuleEngine::Job',
      ],
      jobid => $id,
    });

    push(@results, _job_to_hash($jobs[0])) if ($jobs[0]);
  }

  return {jobs => \@results};
}

sub _job_to_hash {
  my $job = shift;

  my %job = (jobid => $job->jobid, arg => $job->arg, type => $job->funcname());

  return (\%job);
}

## REDHAT EXTENSION START 2094129

=head2 reschedule_jobs

Reset the run_after time for jobs.

You must be in the I<admin> group to use this API.

=head3 Params

ids:
  B<array-of-integers> The ids of the Jobs to reset.

=head3 Returns

A hash containing the new run_after time and the count of jobs reset.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.reschedule_jobs",
      "id": 11,
      "params": [
        {
          "ids": [11, 12, 42]
        }
      ]
  }

=head4 JSONRPC reply

  {
      "error": null,
      "result": {
          "run_after": "???",
          "updated_count": 1
      },
      "id": 11
  }

=cut

sub reschedule_jobs {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('admin')
    || ThrowUserError("auth_failure",
    {group => "admin", action => "run", object => "reschedule_jobs"});

  my $ids
    = $params->{ids}
    || ThrowCodeError('params_required',
    {function => 'RedHat.reschedule_jobs', params => ['ids']});

  my @ids      = map { _clean_id($_) } @$ids;
  my $new_time = time() + 60;

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;
  my $count = $dbh->do(
    'UPDATE ts_job SET run_after = ? WHERE jobid IN ('
      . join(',', (('?') x scalar(@ids))) . ')',
    undef, $new_time, @ids
  );
  $dbh->bz_commit_transaction;

  my $f_time = format_unixtime($new_time, '%Y-%m-%d %H:%M %Z');

  return {updated_count => $count, run_after => $f_time};
}
## REDHAT EXTENSION END 2094129

## REDHAT EXTENSION START 2056724

=head2 tag_user_comments_as_spam

Create a 'despam_comments' job to tag all comments a user has made as spam and
edit the comment text to say so.

Creates one job per user name supplied.

Users must be inactive and not be a member of any groups. To run this for current
users use the C<repudiate> action of L</manage_user_access>.

=head3 Params

names:
  B<array-of-strings> The logins of the users who's comments are to be marked as
  spam.

=head3 Returns

Nothing.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "RedHat.tag_user_comments_as_spam",
      "id": 1,
      "params": [
        {
          "names": ["fred@example.com", "barry@example.org"]
        }
      ]
  }

=cut

sub tag_user_comments_as_spam {
  my $self = shift;
  my ($params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('admin')
    || ThrowUserError("auth_failure",
    {group => "admin", action => "run", object => "tag_user_comments_as_spam"});

  my $names
    = $params->{names}
    || ThrowCodeError('params_required',
    {function => 'RedHat.tag_user_comments_as_spam', params => ['names']});

  my @user_objects = map { Bugzilla::User->check($_) } @{$params->{names}};

  my @enabled_users
    = grep { $_->is_enabled || $_->email_enabled || @{$_->groups} } @user_objects;

  if (@enabled_users) {
    ThrowUserError('cannot_process_active_users',
      {method => 'tag_user_comments_as_spam', users => @enabled_users});
  }

  foreach my $spammer (@user_objects) {
    Bugzilla->job_queue->insert('despam_comments', {spammer => $spammer->login});
  }

  return;
}
## REDHAT EXTENSION END 2056724

1;

__END__

=head1 SEE ALSO

L<Bugzilla|Bugzilla> L<Bugzilla::WebService> L<Bugzilla::Extension::RedHat>
