# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RedHat::WebService::Component;

use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::WebService::Component);
use Bugzilla::Product;
use Bugzilla::Component;
use Bugzilla::User;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(trim);
use Bugzilla::WebService::Util qw(validate);

use List::MoreUtils qw(any);

use constant PUBLIC_METHODS => qw(
  get
  update
  create
  delete
);

# function to return component information when passed either,
# component ids or component and product names or both together.
# example of how it will be called:
# $call = $rpc->call( 'Component.get',
#                          {ids => [1,2], names => [{component => 'testingComponent',product => 'TestProduct'},
#                                                   {component => 'testComponent22' ,product => 'TestProduct'}
#                                                  ]
#                          }
#                   );

sub get {
  my ($self, $params) = validate(@_, 'ids', 'names');

  my $accessible_components = _get_accessible_components($params);

  my @components;

  if (!Bugzilla->user->id) {
    @components = map { {
      id                 => $self->type('int',    $_->id),
      name               => $self->type('string', $_->name),
      description        => $self->type('string', $_->description),
      product_id         => $self->type('int',    $_->product_id),
      product_name       => $self->type('string', $_->product->name),
      default_assignee   => $self->type('string', $_->default_assignee->name),
      default_qa_contact => $self->type(
        'string', (defined $_->default_qa_contact ? $_->default_qa_contact->name : '')
      ),
      ## REDHAT EXTENSION START 876015
      default_docs_contact => $self->type(
        'string',
        (defined $_->default_docs_contact ? $_->default_docs_contact->name : '')
      ),
      ## REDHAT EXTENSION END 876015
      default_cc => [map $self->type('string', $_->name), @{$_->initial_cc}],
      ## REDHAT EXTENSION START 460258 1326981
      template_id => $self->type('int', $_->template_id),
      ## REDHAT EXTENSION END 460258 1326981
      is_active      => $self->type('boolean', $_->is_active),
      sub_components => $self->_get_all_subs($_),
    } } @$accessible_components;
  }
  else {
    @components = map { {
      id                 => $self->type('int',    $_->id),
      name               => $self->type('string', $_->name),
      description        => $self->type('string', $_->description),
      product_id         => $self->type('int',    $_->product_id),
      product_name       => $self->type('string', $_->product->name),
      default_assignee   => $self->type('string', $_->default_assignee->login),
      default_qa_contact => $self->type(
        'string', (defined $_->default_qa_contact ? $_->default_qa_contact->login : '')
      ),
      ## REDHAT EXTENSION START 876015
      default_docs_contact => $self->type(
        'string',
        (defined $_->default_docs_contact ? $_->default_docs_contact->login : '')
      ),
      ## REDHAT EXTENSION END 876015
      default_cc => [map $self->type('string', $_->login), @{$_->initial_cc}],
      flags      => _get_component_flags($_),
      ## REDHAT EXTENSION START 460258 1326981
      template_id => $self->type('int', $_->template_id),
      ## REDHAT EXTENSION END 460258 1326981
      ## REDHAT EXTENSION START 1650366
      agile_team => $self->type(
        'string', defined $_->agile_team_obj ? $_->agile_team_obj->name : ''
      ),
      ## REDHAT EXTENSION END 1650366
      is_active          => $self->type('boolean', $_->is_active),
      default_to_private => $self->type('boolean', $_->default_to_private),
      sub_components     => $self->_get_all_subs($_),
    } } @$accessible_components;
  }

  return {components => \@components};
}

sub _get_all_subs {
  my ($self, $component) = @_;

  my $subs = [map { $self->_sub_component_to_hash($_) }
      @{$component->get_sub_components}];

  return $subs;
}

sub _sub_component_to_hash {
  my ($self, $sub_component) = @_;

  # If not logged in, don't reveal the user's e-mail address
  my $func = Bugzilla->user->id ? 'login' : 'name';

  my $field_data = {
    id                  => $self->type('int',    $sub_component->id),
    name                => $self->type('string', $sub_component->name),
    description         => $self->type('string', $sub_component->description),
    default_assigned_to =>
      $self->type('string', $sub_component->default_assignee->$func),
    default_qa_contact => $self->type(
      'string',
      $sub_component->default_qa_contact ? $sub_component->default_qa_contact->$func
      : ''
    ),
    default_docs_contact => $self->type(
      'string',
      $sub_component->default_docs_contact
      ? $sub_component->default_docs_contact->$func
      : ''
    ),
    sort_key           => $self->type('int',     $sub_component->sortkey),
    is_active          => $self->type('boolean', $sub_component->is_active),
    default_to_private =>
      $self->type('boolean', $sub_component->default_to_private),
    default_cc => [map $self->type('string', $_->login), @{$_->initial_cc}],
  };

  $field_data->{default_pool} = $self->type('int', $sub_component->default_pool);

  $field_data->{agile_team} = $self->type('string',
    defined $sub_component->agile_team_obj
    ? $sub_component->agile_team_obj->name
    : '');

  my $children = $sub_component->children;
  if (scalar(@$children)) {
    $field_data->{children}
      = [map { $self->_sub_component_to_hash($_) } @$children];
  }

  return $field_data;
}

# function to update component information when passed either,
# component ids or component and product names or both togther.
# exmaple of how it will be called:
# $call = $rpc->call( 'Component.update',
#                          {ids => [1,2], names => [{component => 'testingComponent',product => 'TestProduct'},
#                                                   {component => 'testComponent22' ,product => 'ProductA'}
#                                                  ]
#                                         updates => { default_assignee => 'testuser@redhat.com', name => 'comp1' }
#                          }
#                   );

sub update {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar(@{$user->get_products_by_permission('editcomponents')})
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "components"});

  my @accessible_components
    = grep { $user->check_can_admin_product($_->product->name) }
    @{_get_accessible_components($params)};

  # if the user tries to change component name for several
  # components of the same product then throw an error
  my %unique_product_comps;
  foreach my $comp (@accessible_components) {
    if ($unique_product_comps{$comp->product_id}
      && (any { $_ =~ /name/ } keys %{$params->{updates}}))
    {
      ThrowUserError("multiple_components_update_not_allowed");
    }
    else {
      $unique_product_comps{$comp->product_id} = 1;
    }
  }

  my %mapped_returns = (
    initialowner     => 'default_assignee',
    initialqacontact => 'default_qa_contact',
    ## REDHAT EXTENSION 876015: Add initialdocscontact
    initialdocscontact => 'default_docs_contact',
    cc_list            => 'default_cc',
  );

  my %changes;

  foreach my $component_obj (@accessible_components) {
    foreach my $field (keys %{$params->{updates}}) {
      my $method = "set_" . $field;
      if ($field eq "default_cc") {
        $method = "set_cc_list";
      }
      $component_obj->$method($params->{updates}{$field});
    }
  }

  # map the field names in the returned changes hash to the
  # right field name for API consistency and convert
  # the user ids in the changes hash to login names
  foreach my $component_obj (@accessible_components) {
    my $returned_changes = $component_obj->update();
    foreach my $changed_field (keys %{$returned_changes}) {
      my $mapped_field = $mapped_returns{$changed_field} || $changed_field;
      ## REDHAT EXTENSION 876015: Add default_docs_contact
      if ( $mapped_field eq 'default_assignee'
        || $mapped_field eq 'default_qa_contact'
        || $mapped_field eq 'default_docs_contact'
        || $mapped_field eq 'default_cc')
      {
        my $user_list    = $returned_changes->{$changed_field};
        my @old_userlist = split(/[,\s]+/, $user_list->[0]);
        my @new_userlist = split(/[,\s]+/, $user_list->[1]);

        my $old_userlist = join(', ',
          (map $_->{login_name}, @{Bugzilla::User->new_from_list(\@old_userlist)}));
        my $new_userlist = join(', ',
          (map $_->{login_name}, @{Bugzilla::User->new_from_list(\@new_userlist)}));

        $changes{$component_obj->id}{$mapped_field} = [$old_userlist, $new_userlist];
      }
      else {
        $changes{$component_obj->id}{$mapped_field}
          = $returned_changes->{$changed_field};
      }
    }
  }

  return {components_updates => \%changes};

}

# function to create a component
# the function can be called like this with the
# component, product, initialowner and description as required arguments
# $call = $rpc->call('Component.create', { component=> 'TestingComponent',
#                     product => 'TestProduct', description => 'testing component',
#                     default_assignee => 'testuser@redhat.com',
#                     default_qa_contact => 'admin@redhat.com',
#                     default_cc => ['testuserb@redhat.com, 'testuserc@redhat.com'] });
#
sub create {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  $user->in_group('editcomponents')
    || scalar(@{$user->get_products_by_permission('editcomponents')})
    || ThrowUserError("auth_failure",
    {group => "editcomponents", action => "edit", object => "components"});

  my $product_name       = trim($params->{product});
  my $component_name     = trim($params->{component});
  my $default_assignee   = trim($params->{default_assignee});
  my $default_qa_contact = trim($params->{default_qa_contact});
  ## REDHAT EXTENSION 876015: Add default_docs_contact
  my $default_docs_contact = trim($params->{default_docs_contact});
  my $description          = trim($params->{description});
  my $default_cc           = $params->{default_cc};

  ## REDHAT EXTENSION START 460258
  my $template_id
    = defined($params->{template_id})
    ? $params->{template_id}
    : Bugzilla->default_template_id;
  ## REDHAT EXTENSION END 460258

  my $product = $user->check_can_admin_product($product_name);

  my $component = Bugzilla::Component->create({
    name             => $component_name,
    product          => $product,
    description      => $description,
    initialowner     => $default_assignee,
    initialqacontact => $default_qa_contact,
    ## REDHAT EXTENSION 876015: Add default_docs_contact
    initialdocscontact => $default_docs_contact,
    ## REDHAT EXTENSION START 460258
    bvp_template_id => $template_id,
    ## REDHAT EXTENSION END 460258
    initial_cc => $default_cc
  });


  return {id => $self->type('int', $component->id)};

}


# subroutine to hold repeated code between Component.get and
# Component.update, this function returns the unique accessible
# Component objects to the user from list of component ids and
# Component/product names.
sub _get_accessible_components {
  my $params     = shift;
  my @components = ();

  if (defined $params->{ids}) {
    push @components, @{Bugzilla::Component->new_from_list($params->{ids})};
  }

  if (defined $params->{names}) {

    # to get the component objects for product/component combination
    # first obtain the product object from the passed product name
    foreach my $name_hash (@{$params->{names}}) {
      push @components,
        @{Bugzilla::Component->match({
          product_id => Bugzilla::Product->check({name => $name_hash->{product}})->id,
          name       => $name_hash->{component}
        })
        };
    }
  }

  my %seen_component_ids = ();
  my %accessible_product_ids
    = map { $_->id => 1 } @{Bugzilla->user->get_accessible_products};

  my @accessible_components;
  foreach my $component (@components) {

    # Skip if we already included this component
    next if $seen_component_ids{$component->id}++;

    # Can the user see this product?
    next unless $accessible_product_ids{$component->product_id};

    push @accessible_components, $component;
  }

  return \@accessible_components;
}

sub _get_component_flags {
  my $comp_obj = $_;

  my @flag_list;

  foreach my $f (@{$comp_obj->flag_types->{bug}}) {
    next unless $f->is_active;
    push @flag_list, $f->id;
  }
  foreach my $f (@{$comp_obj->flag_types->{attachment}}) {
    next unless $f->is_active;
    push @flag_list, $f->id;
  }

  return \@flag_list;
}

1;

__END__

=head1 NAME

Bugzilla::WebService::Component - The Component API

=head1 DESCRIPTION

This part of the Bugzilla API allows you to get info about accessible components,
create new components and also edit the attributes of existing components.

=head1 METHODS

See L<Bugzilla::WebService> for a description of what B<STABLE>, B<UNSTABLE>,
and B<EXPERIMENTAL> mean, and for more information about error codes.

=head2 Get Components

=over

=item C<get> B<EXPERIMENTAL>

=over

=item B<Description>

Returns information about L<Bugzilla::Component|Components> in accessible
L<Bugzilla::Product|Products>.

=item B<Params> 

At least one of the following two parameters must be specified:

=over

=item C<ids> (array) - An array of integers, representing components ids.

=item C<names> (array) - An array of hashes, each hash has product name and component name.

=back

=item B<Returns>    

Information about all accessible product components passed. Components
that do not exist or are not accessible are silently ignored.

The return value will be A hash containing one item, C<components>, 
that is an array of hashes. Each hash describes a component,
and has the following items:

=over

=item id

C<int> The unique integer ID that Bugzilla uses to identify this component. 
Even if the name of the component changes, this ID will stay the same.

=item name

C<string> The name of the component.

=item description

C<string> The description of the component.

=item product_id

C<int> The id of the product that the component belongs to.

=item product_name

C<string> The name of the product that the component belongs to.

=item default_assignee 

C<string> The login name of the default owner of the component if
the user is logged in, if the user is not logged in then only the
realname of the default owner will be returned.

=item default_qa_contact 

C<string> The login name of the default qa contact of the component if
the user is logged in, if the user is not logged in then only the
realname of the default qa contact will be returned.

=item default_docs_contact 

C<string> The login name of the default docs contact of the component if
the user is logged in, if the user is not logged in then only the
realname of the default docs contact will be returned.

=item default_cc

C<array> The default CC list for this component. It is an array 
of login names if the user is logged in. If the user is not logged in 
then only the realname of the default cc list members  will be returned.

=item flags

C<array> List of component flags, only returned for logged in users.

=item template_id

C<string> The template text ID that is used when creating new bugs for
this component.

=item agile_team

C<string> the Agile Team bugs in this component will default to.

=item C<is_active>

C<boolean> A boolean indicating if the component is active.  Inactive
components are not enabled for new bugs.

This feature is a Red Hat customisation.

=item C<default_to_private>

C<boolean> A boolean indicating if bugs for the component default to private.

This feature is a Red Hat customisation.

=back

=item B<Errors>

=over

=item 51 (Invalid Object)

A non existing product or component name was passed to the function, as a result no 
product or component object existed for that invalid name.

=item 106 (Invalid Product)

You were required to specify a product, and either you didn't, or you
specified an invalid product (or a product that you can't access).

=back

=back

=back

=head2 Create Component

=over

=item C<create> B<EXPERIMENTAL>

=over

=item B<Description>

Creates new component for an existing product.

=item B<Params> 

A hash containing the following component fields:

=over

=item product

C<string> B<Required> The name of the product that the component 
will be added to.

=item component

C<string> B<Required> The name of the new component.

=item description

C<string> B<Required> The description of the new component.

=item default_assignee

C<string> B<Required> The initial owner of the new component.

=item default_qa_contact

C<string> B<Optional> The initial qa contact for the new component.

=item default_docs_contact

C<string> B<Optional> The initial docs contact for the new component.

=item default_cc 

C<array> B<Optional> Array of the initial cclist for the new component.

=item template_id

C<string> B<Optional> The ID of the default template to use when creating
a bug with this component.

=back

=item B<Returns>    

A hash containing one item, C<id>, which is the id of the newly 
created component

=item B<Errors>

=over

=item 304 (Authorization Required)
Logged-in user not authorized to create components.

=item 106 (Product Access Denied)
Logged-in user doesn't have access to the specified product to create 
components for it.

=item 50 (Product Not Specified)
Logged-in user didn't specify any product names for the component they want 
to create.

=item 106 (Product Doesn't Exist)
The specified product doesn't exist.

=item 105 (Component Name Is Too Long)
The name of the component is longer than 64 characters.

=item 105 (Component Requires Default Assignee)
User didn't specify a default assignee for the new component.

=item 105 (Blank Component Description Not Allowed)
User didn't specify a description for the new component.

=item 105 (Blank Component Name Not Allowed)
User didn't specifiy name for the new component.

=item 105 (Component Already Exists)
Trying to create a component which has similar name to an existing component.

=item 51 (Invalid Username)
Invalid login name was provided for the component's default assignee,
default qa contact, default docs contact or one of the default cc.

=back

=back

=back

=head2 Update Components

=over

=item C<update> B<EXPERIMENTAL>

=over

=item B<Description>

Updates single or multiple components information.

=item B<Params> 

A hash containing either C<ids>, that is an array of integers or 
a single integer representing the components ids, 
or C<names>, that is an array of hashes, each hash has 
product name and component name, or both together, and C<updates> 
that is a hash containg the component fields that will be updated,
the update has may contain any of the following items:

=over

=item name

C<string> The name of the component.

=item description

C<string> The description of the component.

=item default_assignee 

C<string> The login name of the default owner of the component.

=item default_qa_contact 

C<string> The login name of the default qa contact of the component.

=item default_docs_contact 

C<string> The login name of the default docs contact of the component.

=item default_cc

C<array> A list of the login names in the default cc list of the component.

=item template_id

C<string> B<Optional> The ID of the default template to use when creating
a bug with this component.

=back

=item B<Returns>    

A hash containing one item, C<components_updates>, which is a hash of hashes,
with component ids as the hash keys pointing to hashes that contains 
the changes that were made to the component information. The hash may contain
any of the following:

=over

=item name

C<array> An array of (strings) that holds the old and the new 
names of the component.

=item description

C<array> An array of (strings) that holds the old and the new
descriptions of the component.

=item default_assignee 

C<array> An Array of (strings) that holds the old and the new 
login names of the default owner of the component.

=item default_qa_contact 

C<array> An array of (strings) that holds the old and the new  
login names of the default qa contact of the component.

=item default_docs_contact 

C<array> An array of (strings) that holds the old and the new  
login names of the default docs contact of the component.

=item default_cc

C<array> An array of (strings) that holds the old and the new 
lists of the login names in the default cc list of the component.

=back

=item B<Errors>

=over

=item 50 (Invalid Product)

You were required to specify a product, and either you didn't, or you
specified an invalid product (or a product that you can't access).

=item 304 (Authorization Required)
Logged-in user not authorized to edit components.

=item 105 (Component Already Exists)
Trying to change component name to similar name of an existing component
of the same product.

=item 51 (Invalid Username)
Invalid login name was provided for the component's default assignee,
default qa contact, default docs contact or one of the default cc.

=item 105 (Component Name Is Too Long)
The name of the component is longer than 64 characters.

=item 105 (Multiple Components Update Not Allowed)
Can not do multiple components update when changing the name
of multiple components of the same product.

=back

=back

=back

=head2 Delete Components

=over

=item C<delete> B<EXPERIMENTAL>

=over

=item B<Description>

This allows you to delete one or more components in Bugzilla.

=item B<Params>

B<Note:> The following parameters specify which components you are deleting.
You must set one or both of these parameters.

=over

=item C<ids>

C<array> of C<int>s. Numeric ids of the components that you wish to delete.

=item C<names>

C<array> of C<hash>es. Names of the components that you wish to delete. The
hash keys are C<product> and C<component>, representing the name of the product
and the component you wish to delete.

=back

=item B<Returns>

A C<hash> with a single field "components". This points to an array of hashes
with the following field:

=over

=item C<id>

C<int> The id of the component that was deleted.

=back

=item B<Errors>

=over

=item 106 (Product access denied)

The product you are trying to modify does not exist or you don't have access to it.

=item 706 (Product admin denied)

You do not have the permission to delete components for this product.

=item 1202 (Component has bugs)

The component you are trying to delete currently has bugs assigned to it.
You must move these bugs before trying to delete the component.

=back

=item B<History>

=over

=item Added in Bugzilla B<5.0>

Backported to RHBZ 4.4.

=back

=back

=back

