# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Releases::ACKCustomSearch;

use warnings;
use strict;
use 5.10.1;

use Carp;
use URI;

sub new {
  my $class  = shift;
  my $params = shift || {joiner => 'AND'};

  if ( ref($params) ne "HASH"
    || !defined $params->{joiner}
    || $params->{joiner} !~ /^(AND|OR)$/)
  {
    croak __PACKAGE__ . " requires a valid argument\n";
  }

  my $self = {};
  $self->{joiner}   = delete $params->{joiner};
  $self->{children} = [];
  $self->{query}    = {};

  bless $self, $class;

  return $self;
}

sub add {
  my ($self, $clause) = @_;

  croak "Clause is not defined\n"
    if (!$clause
    || !ref($clause)
    || (ref($clause) ne "HASH" && !$clause->isa(__PACKAGE__)));

  if (ref($clause) eq "HASH") {
    $self->_validate($clause);
  }

  push @{$self->{children}}, $clause;

  return;
}

sub _validate {
  my ($self, $clause) = @_;

  croak "Invalid clause format\n" if (ref $clause ne "HASH");

  my $checks = {field => 1, operator => 1, value => 1};

  if ($clause->{negate} && $clause->{negate} !~ /^(1|0)$/) {
    croak "negate value is not 1 or 0.\n";
  }

  foreach my $check (keys %{$checks}) {
    croak "Missing $check in clause.\n" if (!defined $clause->{$check});
  }

  return;
}

sub get_query_text {
  my ($self) = @_;

  $self->{query}{'query_format'} = 'advanced';
  my $uri = URI->new;
  $uri->query_form(%{$self->{query}});
  return $uri->query();
}

sub get_query_params {
  my ($self) = @_;

  return $self->{query};
}

sub set {
  my $self = shift;
  my $id   = shift || 0;

  # As the following format
  # ( col1 = val1 AND ( col2 = val2 or col3 = val3 ) )
  # f1=OP, j1=AND,
  #     f2=col1, o2='=', v2=val1,
  #     f3=OP, j3=OR,
  #         f4=col2, o4='=', v4=val2,
  #         f5=col3, o5='=', v5=val3,
  #     f6=CP,
  # f7=CP

  if (@{$self->{children}}) {

    # open clause
    $self->_set_op($id);
    $id++;

    foreach my $child (@{$self->{children}}) {
      if (ref $child eq "HASH") {
        $self->_set_field($child, $id);
        $id++;
      }
      elsif (ref $child && $child->isa(__PACKAGE__)) {
        $id = $child->set($id);
        my $query = $child->{query};
        $self->{query} = {%{$self->{query}}, %{$query}};
      }
    }

    # close clause
    $self->_set_cp($id);
    $id++;
  }

  return $id;
}

sub _set_field {
  my $self   = shift;
  my $clause = shift;
  my $id     = shift || 0;

  my $maps = {field => 'f', operator => 'o', value => 'v', negate => 'n'};

  foreach my $param (keys %{$maps}) {
    next if (!defined $clause->{$param} || $clause->{$param} eq '');
    $self->{query}{"$maps->{$param}$id"} = $clause->{$param};
  }

  return;
}

sub _set_op {
  my $self = shift;
  my $id   = shift || 0;

  $self->{query}{"f$id"} = 'OP';
  $self->{query}{"j$id"} = $self->{joiner};

  return;
}

sub _set_cp {
  my $self = shift;
  my $id   = shift || 0;

  $self->{query}{"f$id"} = 'CP';

  return;
}
1;

__END__

=head1 NAME

BugzillaMetrics::WebApp::BugzillaCustomSearch - Create bugzilla custom search query


=head1 VERSION

This document describes BugzillaMetrics::WebApp::BugzillaCustomSearch


=head1 SYNOPSIS

    use BugzillaMetrics::WebApp::BugzillaCustomSearch;

    # Create a main clause with an AND joiner
    my $main_clause = BugzillaMetrics::WebApp::BugzillaCustomSearch->new();

    # Create an 'OR' sub clause
    my $or_clause = BugzillaMetrics::WebApp::BugzillaCustomSearch->new( { joiner => 'OR'} );

    my $c1 = { field => 'field1', operator => 'equals', value => "value1" };
    $main_clause->add($c1);

    my $sub1 = { field => 'field2', operator => 'anyexact',  value => "value2" };
    my $sub2 = { field => 'field2', operator => 'isnotnull', value => "" };
    $or_clause->add($sub1);
    $or_clause->add($sub2);

    # Join all clauses together
    $main_clause->add($or_clause);

    # Generate bugzilla custom search query format
    $main_clause->set();

    # print out the query
    print $main_clause->get_query_text();

=head1 DESCRIPTION

BugzillaMetrics::WebApp::BugzillaCustomSearch - This module is use to create bugzilla custom query format.

=head1 CONSTRUCTOR

=head2  BugzillaMetrics::WebApp::BugzillaCustomSearch->new( { joiner => AND|OR } )

    Create a clause object with either 'AND' or 'OR' joiner.

=head1 METHODS

=head2 add( HASH_REF|BugzillaMetrics::WebApp::BugzillaCustomSearch )

    Join a column or sub clause.

=head2 set( id )

    Generate bugzilla custom query. Accepts optional field id as argument. Default is 0.

=head2 get_query_text

    Print generated query in url text format.

=head2 get_query_params

    Return generated query in hash format.

=head1 SEE ALSO

BugzillaMetrics::WebApp

=head1 AUTHOR

Hao Chang Yu  C<< <hyu@redhat.com> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2013, Red Hat, Inc. All rights reserved.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
