# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Releases::Util;

use strict;
use warnings;
use 5.10.1;

use Bugzilla;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Group;
use JSON::XS;
use List::MoreUtils qw(any uniq);
use base qw(Exporter);
@Bugzilla::Extension::Releases::Util::EXPORT = qw(
  log_activity
  get_activity
  get_releases
  untaint
  is_empty_array
  get_later_types
  ack_config
  group_user_email
  groups_selection
);

=pod

=head1 NAME

Bugzilla::Extension::Releases::Util - Generic utility functions for bugzilla release support.

=head1 SYNOPSIS

  use Bugzilla::Extension::Releases::Util;

  # Function for logging all release activity
  log_activity($activity_data);

  # Function for retrieving release activity for a release type
  get_activity($params_data);

=head1 DESCRIPTION

This package contains various utility functions which do not belong anywhere
else.

B<It is not intended as a general dumping group for something which
people feel might be useful somewhere, someday>. Do not add methods to this
package unless it is intended to be used for a significant number of files,
and it does not belong anywhere else.

=head1 FUNCTIONS

=over

=item C<log_activity>

=over

=item B<Description>

Log component activity to show the edit history

=item B<Params>

=item B<Returns>

None

=back

=cut

sub log_activity {
  my ($act_ref) = @_;
  my $dbh = Bugzilla->dbh;

  $act_ref->{'act_when'}
    = $act_ref->{'act_when'} ? $act_ref->{'act_when'} : 'now()';

  $dbh->do(
    "INSERT INTO rh_release_activity 
              (release_id, act_type, who, act_when, what, added, removed)
              VALUES (?, ?, ?, ?, ?, ?, ?)",
    undef,
    (
      $act_ref->{'release_id'}, $act_ref->{'act_type'}, $act_ref->{'who'},
      $act_ref->{'act_when'},   $act_ref->{'what'},     $act_ref->{'added'},
      $act_ref->{'removed'}
    )
  );

  return;
}

=item C<get_activity>

=over

=item B<Description>

Return list of changes made to this release

=item B<Params>

=item B<Returns>

None

=back

=cut

sub get_activity {
  my ($params) = @_;
  my $dbh = Bugzilla->dbh;
  my @criteria;
  my @values;

  # Only consider changes since $starttime, if given.
  if (defined $params->{starttime} && $params->{starttime}) {
    trick_taint($params->{starttime});
    push(@values,   $params->{starttime});
    push(@criteria, "rh_release_activity.act_when > ?");
  }

  if (defined $params->{release_id} && $params->{release_id}) {
    trick_taint($params->{release_id});
    push(@values,   $params->{release_id});
    push(@criteria, "rh_release_activity.release_id = ?");
  }

  if (defined $params->{act_type} && $params->{act_type}) {
    trick_taint($params->{act_type});
    push(@values,   $params->{act_type});
    push(@criteria, "rh_release_activity.act_type = ?");
  }

  ThrowUserError('release_no_valid_criteria') unless scalar(@criteria);

  my $query = "
        SELECT rh_release_activity.what, "
    . $dbh->sql_date_format('rh_release_activity.act_when', '%Y.%m.%d %H:%i:%s')
    . ", rh_release_activity.removed, rh_release_activity.added,
               profiles.login_name
          FROM rh_release_activity
    INNER JOIN profiles
            ON profiles.userid = rh_release_activity.who
         WHERE " . join(" AND ", @criteria) . "
      ORDER BY rh_release_activity.act_when";

  my $list = $dbh->selectall_arrayref($query, undef, @values);

  my @operations;
  my $operation = {};
  my $changes   = [];

  foreach my $entry (@$list) {
    my ($what, $when, $removed, $added, $who) = @$entry;
    my %change;

    # An operation, done by 'who' at time 'when', has a number of
    # 'changes' associated with it.
    # If this is the start of a new operation, store the data from the
    # previous one, and set up the new one.
    if ($operation->{'who'}
      && ($who ne $operation->{'who'} || $when ne $operation->{'when'}))
    {
      $operation->{'changes'} = $changes;
      push(@operations, $operation);

      $operation = {};
      $changes   = [];
    }

    $operation->{'who'}  = $who;
    $operation->{'when'} = $when;

    $change{'what'}    = $what;
    $change{'removed'} = $removed;
    $change{'added'}   = $added;

    push(@$changes, \%change);
  }

  if ($operation->{'who'}) {
    $operation->{'changes'} = $changes;
    push(@operations, $operation);
  }

  return \@operations;
}

########################################################
# sub to retrieve approved releases from bugs database

sub get_releases {
  my ($param)                = @_;
  my $flag_name              = $param->{flag_name};
  my $show_inactive_releases = $param->{show_inactive_releases};

  my $match_flag_name_sql = "";
  if ($flag_name) {
    my @flags = split(' ', $flag_name);

    for (my $i = 0; $i < scalar(@flags); $i++) {

      # Trim trialing and ending space
      $flags[$i] =~ s/(^\s+)|(\s+$)//g;

      $flags[$i] =~ s/\*/\%/g;
      $flags[$i] = qq|flagtypes.name LIKE '$flags[$i]'|;
    }

    $match_flag_name_sql = " AND (" . join(' OR ', @flags) . ") ";
  }

  my $show_inactive = $show_inactive_releases;

  if (!defined $show_inactive || $show_inactive eq '') {
    $show_inactive = 0;
  }

  my $sortkey_first = 0;

  my $dbh = Bugzilla->dbh();

  my $releases_sql
    = "SELECT flagtypes.name FROM flagtypes WHERE flagtypes.category like "
    . $dbh->quote('%release');
  $releases_sql .= " AND flagtypes.is_active=1" if !$show_inactive;
  $releases_sql .= $match_flag_name_sql;
  $releases_sql
    .= $sortkey_first
    ? " ORDER BY flagtypes.sortkey, flagtypes.name"
    : " ORDER BY flagtypes.name";

  my $res;
  eval { $res = $dbh->selectcol_arrayref($releases_sql); };
  if ($@) {
    use Carp;
    croak $@;
  }
  return $dbh->selectcol_arrayref($releases_sql);
}

sub untaint {
  my $value  = shift;
  my $return = undef;

  if ((defined $value) && ($value =~ /^([_\/\w\d\. @()&*%+-?']+)$/)) {
    $return = $1;
  }

  return $return;
}

sub is_empty_array {
  my $array = shift;

  return (
        !defined $array
      || ref $array ne 'ARRAY'
      || scalar(@{$array}) <= 0
      || (scalar(@{$array}) == 1
      && (!defined $array->[0] || $array->[0] =~ /^(undef|)$/))
  ) ? 1 : 0;
}

########################################################
# sub to retrieve resolved types after given type

sub get_later_types {
  my $resolved_type      = shift;
  my @ACK_RESOLVED_TYPES = @_;

  my @later_types = ();

  my $index = 0;

  my $length = scalar @ACK_RESOLVED_TYPES;

  for (my $i = 0; $i < @ACK_RESOLVED_TYPES; $i++) {
    if ($resolved_type eq $ACK_RESOLVED_TYPES[$i]) {
      $index = $i;
      last;
    }
  }

  @later_types = splice @ACK_RESOLVED_TYPES, $index, $length;

  return @later_types;

}

sub group_user_email {
  my $gids = shift;

  my $groups = Bugzilla::Group->new_from_list($gids);
  my @mails;

  foreach my $grp (@$groups) {
    my @members = map { $_->{login_name} } @{$grp->members_non_inherited()};
    push(@mails, @members);
  }

  @mails = uniq(@mails);
  return (\@mails);
}

sub groups_selection {
  my $selected_items = shift;

  my $groups = Bugzilla::Group->match({isactive => 1, category => 'Agile'});

  my @vals;

  foreach my $group (@$groups) {
    my %grp = (
      value    => $group->{id},
      name     => $group->{name},
      selected => any { $_ == $group->{id} } @$selected_items,
    );

    push @vals, \%grp;
  }

  return (\@vals);
}

my $cfg = q(
{
    "IMAGE_PATH": "/extensions/Releases/web/images/",
    "bugzilla_url": "/",
    "BZ_LINK": "/show_bug.cgi?id=%s",
    "bugzilla_path": "/",
    "cookie_path": "/",
    "BLOCKER_COLUMN_ASSIGNED": "bugs.bug_status='ASSIGNED'",
    "BLOCKER_COLUMN_MODIFIED": "bugs.bug_status='MODIFIED'",
    "BLOCKER_COLUMN_Enhancements": "keyworddefs.name='FutureFeature'",
    "BLOCKER_COLUMN_Bugs": "keyworddefs.name!='FutureFeature' AND keyworddefs.name!='Regression'",
    "BLOCKER_COLUMN_NEW": "bugs.bug_status='NEW'",
    "BLOCKER_COLUMN_NEEDINFO": "bugs.bug_status='NEEDINFO'",
    "BLOCKER_COLUMN_POST": "bugs.bug_status='POST'",
    "BLOCKER_COLUMN_Regressions": "keyworddefs.name='Regression'",
    "BLOCKER_COLUMN_Regression": "keyworddefs.name='Regression'",
    "BLOCKER_COLUMN_ON_QA": "bugs.bug_status='ON_QA'",
    "BLOCKER_COLUMN_Enhancement": "keyworddefs.name='FutureFeature'",
    "BLOCKER_COLUMN_Security": "keyworddefs.name='Security'",
    "BLOCKER_COLUMN_Bug": "keyworddefs.name!='FutureFeature' AND keyworddefs.name!='Regression' AND keyworddefs.name!='Security'",
    "BLOCKER_COLUMN_ON_DEV": "bugs.bug_status='ON_DEV'",
    "DEFAULT_ACK_RESOLVED_TYPE": "ON_QA",
    "BLOCKER_COLUMNS": ["Bugs", "Enhancements", "Regressions"],
    "BLOCKER_COLORS": ["red.png", "blue.png", "pink.png", "green.png"],
    "BLOCKER_security_COLORS": ["red.png", "pink.png", "pink.png", "pink.png", "green.png"],
    "BLOCKER_status_COLORS": ["red.png", "red.png", "red.png", "red.png", "red.png", "red.png", "red.png", "green.png"],
    "BLOCKER_status_COLUMNS": ["NEW", "ASSIGNED", "MODIFIED", "POST", "NEEDINFO", "ON_QA", "ON_DEV"],
    "BLOCKER_security_COLUMNS": ["Security", "Bug", "Enhancement", "Regression"],
    "contact": "bugzilla-owner@redhat.com",
    "ACK_FLAGS": ["pm_ack", "devel_ack", "qa_ack"],
    "AVAILABLE_FLAGS": ["blocker", "exception", "needinfo", "hot_fix_requested", "hss_hot_fix"],
    "FLAG_LIST": ["BLOCKER-BETA", "BLOCKER", "EXCEPTION", "FAST"],
    "SORT_BY": ["bug_id", "assigned_to,component,bug_id", "component,assigned_to,bug_id", "bug_severity,component,bug_id", "bug_status,component,assigned_to,bug_id", "component,bug_id", "pm_score", "component,pm_score"],
    "errata_url": "https://errata.devel.redhat.com",
    "sfdc_url": "https://c.na7.visual.force.com/apex/Case_View?sbstr=",
    "c_cf_fixed_in": "Fixed In",
    "c_cf_conditional_nak": "DevNAK",
    "c_cf_qe_conditional_nak": "QeNAK",
    "def_cols": {
        "acks": "Acks",
        "flag_blocker": "Blocker",
        "flag_hot_fix_requested": "Hot Fix Requested",
        "flag_exception": "Exception",
        "flag_needinfo": "Needinfo",
        "cf_conditional_nak": "Devel Conditional NAK",
        "cf_qe_conditional_nak": "QE Conditional NAK",
        "blocked": "Blocks",
        "cf_pm_score": "PM Score",
        "rep_platform": "Platform",
        "bug_severity": "Severity",
        "priority": "Priority",
        "keywords": "Keywords",
        "assigned_to": "AssignedTo",
        "qa_contact": "QAContact",
        "bug_status": "Status",
        "component": "Component",
        "short_short_desc": "Short Summary",
        "flag_hss_hot_fix": "Hss Hot Fix",
        "external": "Links",
        "dependson": "Depends on",
        "cf_partner": "Partner",
        "reporter": "ReportedBy",
        "cf_fixed_in": "Fixed In Version",
        "cf_verified": "Verified",
        "target_release": "Target Release",
        "target_milestone": "Target Milestone",
        "flagtypes.name": "Flags",
        "estimated_time": "Estimated Hours",
        "remaining_time": "Remaining Hours",
        "status_whiteboard": "Status Whiteboard",
        "cf_internal_whiteboard": "Internal Whiteboard",
        "cf_devel_whiteboard": "Devel Whiteboard",
        "short_desc": "Full Summary",
        "cf_internal_target_release" : "Internat Target Release"
    },
    "VIEW": [ 
        "acks",
        "flag_blocker",
        "flag_hot_fix_requested",
        "flag_hss_hot_fix=0",
        "flag_exception",
        "flag_needinfo",
        "cf_conditional_nak",
        "cf_qe_conditional_nak",
        "external=1",
        "cf_internal_target_release",
        "blocked",
        "dependson=0",
        "cf_partner=0",
        "cf_pm_score",
        "rep_platform",
        "bug_severity",
        "priority",
        "keywords",
        "assigned_to",
        "qa_contact",
        "reporter=0",
        "bug_status",
        "component",
        "cf_fixed_in=0",
        "cf_verified=0",
        "target_release=0",
        "target_milestone=0",
        "flagtypes.name=0",
        "estimated_time=0",
        "remaining_time=0",
        "status_whiteboard=0",
        "cf_internal_whiteboard=0",
        "cf_devel_whiteboard=0",
        "short_short_desc",
        "short_desc=0"
    ]
}
);

sub ack_config {

  my $config = eval { JSON::XS->new->decode($cfg) };
  if ($@) {
    die "Error parsing JSON: $@";
  }

  # Make sure we grab data from params
  $config->{"IMAGE_PATH"} = correct_urlbase() . "extensions/Releases/web/images/";
  $config->{"bugzilla_url"}  = correct_urlbase();
  $config->{"bugzilla_path"} = Bugzilla->params->{'cookiepath'};
  $config->{"cookie_path"}   = Bugzilla->params->{'cookiepath'};
  $config->{"contact"}       = Bugzilla->params->{maintainer};
  $config->{"sfdc_url"}
    = "https://" . Bugzilla->params->{sfdc_hostname} . '/apex/Case_View?sbstr=';

  return ($config);
}

=pod

=back

=head1 SEE ALSO

L<Bugzilla> 

=cut
