# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Releases::WebService;

use strict;
use warnings;
use 5.10.1;

use base qw(Bugzilla::WebService);

use Bugzilla::FlagType;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::WebService::Constants;
use Bugzilla::WebService::Util qw(filter validate);

use Bugzilla::Extension::Releases::Components;
use Bugzilla::Extension::Releases::Release;
use Bugzilla::Extension::Releases::Schedule;

use constant PUBLIC_METHODS => qw(
  getReleaseComponents
  updateReleaseComponents
  getReleaseSchedule
  updateReleaseSchedule
  getRelease
  updateRelease
);

=pod

=head1 NAME

Bugzilla::Extensions::Releases::WebService - XMLRPC Methods for obtaining and setting Bugzilla Release attributes. 

=head1 DESCRIPTION

This module provides methods that can be used to get information 
about a bug report in contained in the Bugzilla Bug Tracking System. 
Methods are also provided to update certain attributes of an existing 
bug report or create a new report. The username and password for a 
Bugzilla account can be passed in as parameters or cookies can be 
passed with the HTTP request that correspond to a valid Bugzilla 
login.

Method calls are normally in the form of namespace.method(arguments). 
Where namespace in this case is "bugzilla". So for example 
bugzilla.getBug(bugid, username, password) would return the results 
of getBug() for bug id $bugid.

=head1 METHODS

=over

=item C<getReleaseComponents($params)>

Returns release components for the passed release name in a set of lists. The lists include
'approved', 'capacity', 'ack', and 'nack'.

    $params = { 
        release => "<release>"
            # Release name or flag to return list of approved components for
    }

=cut

sub getReleaseComponents {
  my ($self, $params) = validate(@_);

  $params->{'release'} || ThrowCodeError('param_required', {param => 'release'});

  my $dbh = Bugzilla->dbh;

# Connect to the shadow database if this installation is using one to improve performance
  Bugzilla->switch_to_shadow_db();

  my $release = Bugzilla::Extension::Releases::Release->new(
    {release => $params->{'release'}});

  $release || ThrowUserError('release_invalid_flag_name');

  my $release_components = Bugzilla::Extension::Releases::Components->new(
    {release => $release->flag_type->name});

  return $release_components->get_release_components();
}

=item C<updateReleaseComponents($params)>

Returns release components for the passed flag name in a set of lists. The lists include
'approved', 'capacity', 'ack', and 'nack'.

    $params = {
        "release" => "<release>",                
            # Name of release (flag) to update components for.
        "approved"  => [],          
            # List of approved components to add, remove or make exact.
        "approved_action" => "", 
            # Action to take on approved components.
            #   add = Add the provided components to the existing list
            #   remove = Remove the provided components from the existing list
            #   makeexact = Make the existing list match the list provided.
        "capacity"  => [],          
            # List of capacity priority components to add, remove or make exact
        "capacity_action" => "", 
            # Action to take on capacity components.
            #   add = Add the provided components to the existing list
            #   remove = Remove the provided components from the existing list
            #   makeexact = Make the existing list match the list provided.
        "allow_empty" => 1|0,
            # Allow either capacity or approved lists to be empty.
    }

=cut

sub updateReleaseComponents {
  my ($self, $params) = validate(@_);

  my $dbh = Bugzilla->dbh;

  Bugzilla->user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "components"});

  $params->{'release'} || ThrowCodeError('param_required', {param => 'release'});

  if (!defined $params->{'approved'} && !defined $params->{'capacity'}) {
    ThrowCodeError('param_required', {param => 'approved'});
  }

  my $release = Bugzilla::Extension::Releases::Release->new(
    {release => $params->{'release'}});

  $release || ThrowUserError('release_invalid_flag_name');

  my $change_data_ref = {
    release => $release,    # pass $release instead of $id to save db calls
    approved        => $params->{'approved'},
    approved_action => $params->{'approved_action'},
    capacity        => $params->{'capacity'},
    capacity_action => $params->{'capacity_action'},
    allow_empty     => $params->{'allow_empty'} // 0,

  };

  my $release_components = Bugzilla::Extension::Releases::Components->new(
    {release => $release->flag_type->name});

  $release_components->set_release_components($change_data_ref);

  return $release_components->get_release_components();
}

=item C<getReleaseSchedule($params)>

Returns release schedule for the passed release name

    $params = {
        "release" => "<release>",                
            # Name of release (flag).
    }

=cut

sub getReleaseSchedule {
  my ($self, $params) = validate(@_);

  my $dbh = Bugzilla->dbh;

  Bugzilla->user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "list", object => "schedule"});

  $params->{'release'} || ThrowCodeError('param_required', {param => 'release'});

# Connect to the shadow database if this installation is using one to improve performance
  Bugzilla->switch_to_shadow_db();

  my $release = Bugzilla::Extension::Releases::Release->new(
    {release => $params->{'release'}});

  $release || ThrowUserError('release_invalid_flag_name');

  my $schedule = Bugzilla::Extension::Releases::Schedule->new(
    {release => $release->flag_type->name});

  $schedule->phases if ($schedule);

  return $schedule;
}

=item C<updateReleaseSchedule($params)>

Updates the release schedule for the passed release name.

    $params = {
        "release" => "<release>",                
            # Name of release (flag)
        "phases"  => {},
            # A hash reference containing phase names which are references
            # to other hashes containing their start and end dates. NOTE: This is a make 
            # exact type operation. Any phases not passed in will be removed.
            # { development => { date_start => 'YYYY-MM-DD', date_end => 'YYYY-MM-DD' } }  
    }

=cut

sub updateReleaseSchedule {
  my ($self, $params) = validate(@_);

  my $dbh = Bugzilla->dbh;

  Bugzilla->user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "schedule"});

  foreach my $param ("release", "phases") {
    defined $params->{$param}
      || ThrowCodeError('param_required', {param => $param});
  }

  my $release = Bugzilla::Extension::Releases::Release->new(
    {release => $params->{'release'}});

  $release || ThrowUserError('release_invalid_flag_name');

  my $schedule = Bugzilla::Extension::Releases::Schedule->new(
    {release => $release->flag_type->name});

  my $changes = {};
  if (!$schedule) {
    $schedule = Bugzilla::Extension::Releases::Schedule->create(
      {release_id => $release->id, phases => $params->{'phases'}});
    $changes->{'phases'} = {added => [keys %{$params->{'phases'}}]};
  }
  elsif ($schedule && !%{$params->{'phases'}}) {
    $changes->{'phases'} = {removed => [keys %{$schedule->phases}]};
    $schedule->remove_from_db();
    $schedule = undef;
  }
  else {
    $schedule->set_phases($params->{'phases'});
    $changes = $schedule->update();
  }

  return $changes;
}

=item C<getRelease($params)>

Retrieve information about a specific release or a full list of current releases.

    $params = {
        "release" => "<release>",
        # Name of the release to retrieve information for (optional)                
    }

=cut

sub getRelease {
  my ($self, $params) = validate(@_);

  my $dbh = Bugzilla->dbh;

  Bugzilla->user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "release"});

# Connect to the shadow database if this installation is using one to improve performance
  Bugzilla->switch_to_shadow_db();

  if (defined $params->{'release'}) {
    my $release = Bugzilla::Extension::Releases::Release->new(
      {release => $params->{'release'}});
    $release || ThrowUserError('release_invalid_flag_name');
    return $release;
  }
  else {
    return Bugzilla::Extension::Releases::Release->match();
  }
}

=item C<updateRelease($params)>

Update information about a specific release

    $params = {
        "release" => "<release>",
        # Name of the release to update information for (required)
        "description" => "<description>",
        # Description of release
    }

=cut

sub updateRelease {
  my ($self, $params) = validate(@_);

  my $dbh = Bugzilla->dbh;

  Bugzilla->user->can_edit_releases()
    || ThrowUserError("auth_failure",
    {group => 'program_management', action => "edit", object => "release"});

  foreach my $param ("release", "description") {
    defined $params->{$param}
      || ThrowCodeError('param_required', {param => $param});
  }

  my $release = Bugzilla::Extension::Releases::Release->new(
    {release => $params->{'release'}});

  if (!$release) {
    return Bugzilla::Extension::Releases::Release->create({
      release => $params->{'release'}, description => $params->{'description'}
    });
  }
  elsif ($params->{'remove'}) {
    $release->remove_from_db();
    return 1;
  }
  else {
    $release->set_description($params->{'description'});
    return $release->update();
  }
}

1;

=back

=head1 SEE ALSO

L<Bugzilla|Bugzilla> L<Bugzilla::WebService|Bugzilla::WebService>
