# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::RuleEngine::WebService;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);
use Bugzilla::Product;
use Bugzilla::Component;
use Bugzilla::User;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util qw(trim diff_arrays);
use Bugzilla::WebService::Util qw(validate);

use Bugzilla::Extension::RuleEngine::RuleGroup;
use Bugzilla::Extension::RuleEngine::Rule;
use JSON qw(to_json from_json);

use constant PUBLIC_METHODS => qw(
  create
  get
  list
  update
  update_order
);

sub list {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $is_active = $args->{is_active} ? $args->{is_active} : 0;

  my @rule_groups
    = Bugzilla::Extension::RuleEngine::RuleGroup->all_user_can_view($user);

  my $group_data = [];
  my $rule_data  = [];

  foreach my $group (@rule_groups) {
    next if ($is_active && !$group->is_active);

    push @$group_data,
      {
      name        => $self->type('string',  $group->name),
      is_active   => $self->type('boolean', $group->is_active),
      id          => $self->type('int',     $group->id),
      description => $self->type('string',  $group->description),
      };

    my $params = {rule_group_id => $group->id, inactive => 1,};

    if ($is_active) {
      $params->{isactive} = 1;
    }

    my @rules = @{Bugzilla::Extension::RuleEngine::Rule->match_rules($params)};

    foreach my $rule (@rules) {
      push @$rule_data,
        {
        id            => $self->type('int',     $rule->id),
        name          => $self->type('string',  $rule->name),
        description   => $self->type('string',  $rule->description),
        rule_group_id => $self->type('int',     $rule->rule_group_id),
        is_active     => $self->type('boolean', $rule->current->is_active),
        };
    }
  }

  my $data = {rules => $rule_data, rule_groups => $group_data,};

  return $data;
}

sub get {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $rule_ids   = $args->{ids}   // [];
  my $rule_names = $args->{names} // [];

  my @rules;

  push @rules, map { Bugzilla::Extension::RuleEngine::Rule->new($_) } @$rule_ids;
  push @rules,
    map { Bugzilla::Extension::RuleEngine::Rule->new({name => $_}) } @$rule_names;

  my @rule_defns;

  foreach my $rule (@rules) {
    my $current = $rule->current;
    ThrowUserError('cannot_view_rule_group')
      unless ($rule->rule_group->user_can_view($user));

# The rule definition is internally stored as two json encoded text strings.
# we deserialize them here so they can be neatly reencoded into json as a single
# piece of text for the user to pass around.
# This also makes deserialization on our end easier.
    my $details = {
      match_json  => from_json($current->match_json),
      action_json => from_json($current->action_json),
    };

    my $data = {
      id          => $self->type('int',    $rule->id),
      name        => $self->type('string', $rule->name),
      description => $self->type('string', $rule->description),
      owners => [map { $self->type('string', $_->login) } @{$rule->rule_owners}],
      rule_group_id    => $self->type('int', $rule->rule_group_id),
      is_active        => $self->type('boolean', $current->is_active),
      last_modified    => $self->type('string', $current->creation_ts),
      is_periodic      => $self->type('boolean', $rule->is_periodic),
      minor_update     => $self->type('boolean', $rule->minor_update),
      definition       => $self->type('string', to_json($details)),
      last_used        => $self->type('string', $rule->last_used),
      use_count        => $self->type('int', $rule->use_count),
      use_since_change => $self->type('int', $rule->use_since_change),
      usage_7days      => $self->type('int', $rule->last7days),
      usage_24hrs      => $self->type('int', $rule->last24hours),
      details          => $details,
    };

    push @rule_defns, $data;
  }

  return \@rule_defns;
}

sub update {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $rules = $args->{rules}
    // ThrowCodeError('param_required', {param => 'rules'});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $changes = [];

  foreach my $rule_data (@$rules) {
    ThrowCodeError('param_required', {param => 'id'})
      unless defined $rule_data->{id};

    my $rule = Bugzilla::Extension::RuleEngine::Rule->new(int $rule_data->{id});
    ThrowUserError('cannot_edit_rule_group')
      unless ($rule->rule_group->user_can_edit($user));

    my $rule_changes = {id => $rule->id};

    if (defined($rule_data->{name})) {
      my $old_name = $rule->name;
      my $new_name = $rule_data->{name};
      if ($old_name ne $new_name) {
        $rule_changes->{name} = [$old_name, $new_name];
        $rule->set_name($new_name);
      }
    }

    if (defined $rule_data->{description}) {
      my $old_desc = $rule->description;
      my $new_desc = $rule_data->{description};
      if ($old_desc ne $new_desc) {
        $rule_changes->{description} = [$old_desc, $new_desc];
        $rule->set_description($new_desc);
      }
    }

    if (defined($rule_data->{owners})) {
      my $old_owners = [map { $_->login } @{$rule->rule_owners}];
      my $owners     = $rule_data->{owners};

      # set_rule_owners will split and validate the string.
      $rule->set_rule_owners(join(',', @$owners));

      my $new_owners = [map { $_->login } @{$rule->rule_owners}];

      my ($removed, $added) = diff_arrays($old_owners, $new_owners);

      $rule_changes->{owners} = [$removed, $added]
        if scalar(@$removed) || scalar(@$added);
    }

    if (defined $rule_data->{rule_group_id}) {
      my $old_rule_group_id = $rule->rule_group_id;
      my $new_rule_group_id = int $rule_data->{rule_group_id};

      if ($old_rule_group_id != $new_rule_group_id) {
        $rule->set_rule_group_id($new_rule_group_id);
        $rule_changes->{rule_group_id} = [$old_rule_group_id, $new_rule_group_id];
      }
    }

    if (defined $rule_data->{is_periodic}) {
      my $old_value = $rule->is_periodic;
      my $new_value = int $rule_data->{is_periodic};

      if ($old_value != $new_value) {
        $rule->set_is_periodic($new_value);
        $rule_changes->{is_periodic} = [$old_value, $new_value];
      }
    }

    if (defined $rule_data->{minor_update}) {
      my $old_value = $rule->minor_update;
      my $new_value = int $rule_data->{minor_update};

      if ($old_value != $new_value) {
        $rule->set_minor_update($new_value);
        $rule_changes->{minor_update} = [$old_value, $new_value];
      }
    }

    # If the definition or the active state of the rule has changed,
    # we have to to create a new RuleDetail object.
    if (defined($rule_data->{definition}) || defined($rule_data->{is_active})) {
      my ($match, $action);

      # If a new definition has been supplied use that.
      # If not, use the existing rule defintion.
      if (defined $rule_data->{definition}) {
        my $defn = from_json($rule_data->{definition});
        $match  = $defn->{match_json}  // ThrowCodeError();
        $action = $defn->{action_json} // ThrowCodeError();
        $rule_changes->{definition} = 1;
      }
      else {
        $match  = $rule->current->match;
        $action = $rule->current->action;
      }

      # If a new active state has been supplied, use that.
      my $is_active
        = defined $rule_data->{is_active}
        ? int $rule_data->{is_active}
        : $rule->current->is_active;

      if ($is_active != $rule->current->is_active) {
        $rule_changes->{is_active} = [$rule->current->is_active, $is_active];
      }

      my $rule_detail = Bugzilla::Extension::RuleEngine::RuleDetail->create({
        match      => $match,
        action     => $action,
        creator_id => Bugzilla->user->id,
        rule_id    => $rule->id,
        isactive   => $is_active,
      });

      $rule->set_current_detail_id($rule_detail->id);

      push @$changes, $rule_changes;
    }
    $rule->update;
  }

  $dbh->bz_commit_transaction;

  return $changes;
}

sub create {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $rules = $args->{rules}
    // ThrowCodeError('param_required', {param => 'rules'});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  my $changes = [];

  foreach my $rule (@$rules) {
    my $name = $rule->{name} // ThrowCodeError('param_required', {param => 'name'});
    my $description = $rule->{description}
      // ThrowCodeError('param_required', {param => 'description'});
    my $owners        = $rule->{owners} // [];
    my $rule_group_id = int($rule->{rule_group_id})
      // ThrowCodeError('param_required', {param => 'rule_group_id'});
    my $is_periodic  = int($rule->{is_periodic}  // 0);
    my $minor_update = int($rule->{minor_update} // 0);
    my $definition   = $rule->{definition}
      // ThrowCodeError('param_required', {param => 'definition'});
    my $is_active = $rule->{is_active}
      // ThrowCodeError('param_required', {param => 'is_active'});

    my $defn_obj = from_json($definition);
    my $match    = $defn_obj->{match_json} // ThrowCodeError('rule_malformed_defn');
    my $action = $defn_obj->{action_json} // ThrowCodeError('rule_malformed_defn');

    my $rule_group
      = Bugzilla::Extension::RuleEngine::RuleGroup->new($rule_group_id);
    ThrowUserError('cannot_edit_rule_group')
      unless ($rule_group->user_can_edit($user));

    my $rule_obj = Bugzilla::Extension::RuleEngine::Rule->create({
      name          => $name,
      description   => $description,
      rule_group_id => $rule_group_id,
      is_periodic   => $is_periodic,
      minor_update  => $minor_update,
    });

    my $rule_detail = Bugzilla::Extension::RuleEngine::RuleDetail->create({
      match      => $match,
      action     => $action,
      creator_id => Bugzilla->user->id,
      rule_id    => $rule_obj->id,
      isactive   => $is_active,
    });

    $rule_obj->set_rule_owners(join(',', @$owners)) if scalar(@$owners);
    $rule_obj->set_current_detail_id($rule_detail->id);

    $rule_obj->update;

    my $change = {
      id            => $rule_obj->id,
      name          => $rule_obj->name,
      description   => $rule_obj->description,
      rule_group_id => $rule_obj->rule_group_id,
      is_periodic   => $rule_obj->is_periodic,
      is_active     => $rule_obj->current->is_active,
      minor_update  => $rule_obj->minor_update,
    };

    push @$changes, $change;
  }
  $dbh->bz_commit_transaction;

  return $changes;
}

sub update_order {
  my ($self, $params) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  my $changes;

  my $rules
    = delete $params->{rules}
    || ThrowCodeError('params_required',
    {function => 'RuleGroup.update_order', params => ['rules']});

  my $dbh = Bugzilla->dbh;
  $dbh->bz_start_transaction;

  foreach my $rule (@$rules) {
    my $id      = $rule->{id} // ThrowCodeError('param_required', {param => 'id'});
    my $sortkey = int $rule->{sortkey}
      // ThrowCodeError('param_required', {param => 'sortkey'});

    my $rule = Bugzilla::Extension::RuleEngine::Rule->new($id);
    ThrowUserError('cannot_edit_rule_group')
      unless ($rule->rule_group->user_can_edit($user));

    my $old_val = $rule->sortkey();

    if ($sortkey != $old_val) {
      $rule->set_sortkey($sortkey);
      $rule->update();
      push @$changes, {id => $rule->id, sortkey => [$old_val, $sortkey]};
    }
  }

  $dbh->bz_commit_transaction;

  return $changes;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::RuleEngine::WebService - The RuleEngine API

=head1 DESCRIPTION

This API allows you to upload, download, enable and disable rules.

The API does not allow full editing of rules, but rather allows rules to
be copied between testing and production servers.

Also possible is fetching usage stats of rules, as well as switching them on and off.

=head1 METHODS

=head2 list

=over

=item B<Description>

Returns basic information about all the rules, and rule groups in the rules engine

=item B<Params>

=over

=item is_active

C<boolean> I<optional> If set then only active groups and rules will be returned.

=back

=item B<Returns>

Returns a list of all rules, and a list of all rule groups in the rules engine.

=over

=item rules

C<array-of-hashes> Returns an array of hashes listing all the rules in the rules engine.
The format of the hash is explained below

=over

=item id

C<int> The ID of the rule

=item name

C<string> The name of the rule

=item description

C<string> The description of the rule

=item rule_group_id

C<int> The ID of the rule group that this rule belongs to

=item is_active

C<boolean> True if the rule is enabled.  False if the rule is disabled

=back

=item rule_groups

C<array-of-hashes> Returns an array of hashes listing all the rule groups in the rules
engine.  The format of the hash is explained below.

=over

=item id

C<int> The ID of the rule group

=item name

C<name> The name of the rule group

=item description

C<string> The description of the rule group

=item is_active

C<boolean> True if the rule group is active.  False if inactive.

=back

=back

=back

=head2 get

=over

=item B<Description>

Fetches detailed information about a list of rules

=item B<Params>

=over

=item ids

C<array> I<optional> An array of rule ids.

=item names

C<array> I<optional> An array of rule names

=back

=item B<Returns>

Returns an array of hashes.  Each hash describes a rule. The format of each has is explained below.

=over

=item id

C<int> The ID of the rule

=item name

C<string> The name of the rule

=item description

C<string> The rule description

=item definition

C<string> The rule data itself.  Note that this format is not documented and is subject to change.
Manual editing of this is discouraged as the format may change without notice.  This should be
passed to the update or create functions unmodified.

=item is_active

C<boolean> True if the rule is active and running.  False if the rule is inactive.

=item rule_group_id

C<int> The rule group that this rule belongs to.

=item owners

C<array-of-strings> An array of email addresses.  Each address is a registered bugzilla user that
will receive email notifications upon modification of the rule, or if any errors occour.

=item last_modified

C<string> A timestamp of when the current version of the rule was created

=item last_used

C<string> A timestamp of then the rule was last run

=item use_count

C<int> A counter of how many times the rule has run since it was created

=item use_since_change

C<int> A counter of how many times the rule has run since it was last modified.

=item usage_7days

C<int> A counter of how many times the rule has run in the last 7 days.

=item usage_24hrs

C<int> A counter of how many times the rule has run in the last 24 hours.

=item is_periodic

C<boolean> True if the rule runs once per 24 hours.  False if it runs on bug change

=item minor_update

C<boolean> True if changes by this rule should be flagged as minor updates.

=back

=back

=head2 create

=over

=item B<Description>

Create a new rules engine rule using a rule definition copied from an existing rule

=item B<Params>

The function takes a hash with a singular key, 'rules'.  The value of the rules item
is an array of hashes.  The format of each hash within this array is described below.

=over

=item name

B<string> I<mandatory> The name of the new rule

=item description

B<string> I<mandatory> The description of the new rule

=item owners

B<array-of-strings> I<optional> An array of email addresses of registered bugzilla users.

These users will receive notifications when the rule is modified, or if there are any
run time errors causing the rule to halt.

=item rule_group_id

B<integer> I<mandatory> The ID if the rule group that this rule belongs to.

A list of valid rule group ids can be obtained from the B<list> RPC call.

=item definition

B<string> I<mandatory> The rule definition.

This can be obtained from the B<get> RPC call.

=item is_active

B<boolean> I<mandatory>  If true the rule will be created in a running state.
If false, the rule will be created and disabled.

=item is_periodic

B<boolean> I<optional> If True the rule will be run once per 24 hours.
If False, the rule will run on bug change.

Default: B<False>.

=item minor_update

B<boolean> I<optional> If True then changes made to bugs by this rule will be
marked as minor updates.

=back

=item B<Returns>

Returns an array of hashes.  Each hash describes the new rule that has been created.

The contents of the hash are identical to the above, with the addition of the ID parameter.

=over

=item id

B<int> The ID of the newly created rule

=back

=back

=head2 update

=over

=item B<Description>

Update the definition of an existing rule, Also allow a rule to be enabled or disabled

=item B<Params>

The function takes a hash with a singular key, 'rules'.  The value of the rules item
is an array of hashes.  The format of each hash within this array is described below.

If you do not wish to update a particular paramter, do not supply a value for it.

=over

=item id

B<int> I<mandatory> The ID of the rule to update.

=item name

B<string> I<optional> The new name of the rule

=item description

B<string> I<optional> The new description of the rule

=item owners

B<array-of-strings> I<optional> An array of email addresses of the new rule owners.

A rule owner must be a registered bugzilla user.  Rule owners will receive notifications
if there are any errors with their rules.

=item rule_group_id

B<int> I<optional> The new group that this rule should belong to

=item is_periodic

B<boolean> I<optional> True if the rule should run once every 24 hours.
False if the rule should run on bug change.

=item minor_update

B<boolean> I<optional> True if changes by this rule should be flagged as minor
updates.

=item definition

B<string> I<optional> A new rule definition.  Rule definitions can be obtained from the B<get> RPC call.

=item is_active

B<boolean> I<optional> True to enable the rule.  False to disable the rule.

=back

=item B<Returns>

Returns an array of hashes.  Each hash describes the changes to one rule. Only fields that were changed
will be returned.

Each field will be returned as an array.
Elemenet 0 will be the old value that was replaced.
Element 1 will be the new value for that field.

=back

=head2 update_order

=over

=item B<Description>

Update the order of multiple rules.

=item B<Params>

The function takes an array of hashes. Each entry must contain 'id' and 'sortkey'. 

=over

=item id

B<int> I<mandatory> The ID of the rule to update.

=item sortkey

B<int> I<mandatory> The order of the rule.

=back

=item B<Returns>

An array of hashes with the changes made.

=back
