# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::SAML2Auth::Login;
use 5.10.1;
use strict;
use warnings;
use base   qw(Bugzilla::Auth::Login);
use fields qw(saml2auth_failure);
use URI;
use Net::LDAP;

use Bugzilla;
use Bugzilla::Constants qw(AUTH_NODATA AUTH_ERROR USAGE_MODE_BROWSER);
use Bugzilla::Util
  qw(trick_taint correct_urlbase generate_random_password validate_email_syntax);
use Bugzilla::Extension::SAML2Auth::Util;
use Bugzilla::Extension::SAML2Auth::IDP;
use Bugzilla::Error;
use Bugzilla::Token;
use URI::Escape;

use Bugzilla::Extension::BayotBase::Util qw(clean_params);

use constant {
  requires_verification   => 1,
  user_can_create_account => 1,
  extern_id_used          => 1,
  can_change_email        => 0
};

sub get_login_info {
  my ($self) = @_;
  my $cgi = Bugzilla->cgi;

  my $logincookie = $cgi->cookie('Bugzilla_logincookie');
  my $saml_choice = $cgi->cookie('Bugzilla_saml_choice');

  # Skip SAML if user and pass is included
  my $login    = $cgi->param('Bugzilla_login');
  my $password = $cgi->param('Bugzilla_password');
  my $cur_url  = $cgi->url();

  if ( !$login
    && !$password
    && !$logincookie
    && $saml_choice
    && $cur_url !~ m/saml2_[^.]*\.cgi/)
  {
    my $idp = Bugzilla::Extension::SAML2Auth::IDP->new({name => $saml_choice});
    if ($idp) {
      my $urlbase = correct_urlbase();
      my $target  = clean_params($cgi->url("-relative" => 1, "-query" => 1));
      my $url     = "$urlbase/saml2_login.cgi?idp=$saml_choice&target="
        . uri_escape_utf8($target);
      $cgi->remove_cookie('Bugzilla_saml_choice');
      print($cgi->redirect(-uri => $url));
      exit;
    }
  }

## TODO do we need to check an attribute in the assertation?
  return ({failure => AUTH_NODATA})
    unless (Bugzilla->request_cache->{saml2auth_assertation});

  my $assertation = delete(Bugzilla->request_cache->{saml2auth_assertation});

  # Verify InResponseTo attribute matches a token
  my $tk = $assertation->in_response_to;
  my (undef, undef, $data, $type) = Bugzilla::Token::GetTokenData($tk);

  return ({
    failure    => AUTH_ERROR,
    user_error => 'saml_no_such_token',
    details    => {in_response_to => $tk}
  })
    unless ($data);

  delete_token($tk);

  my $idp_name = $data;

  return ({failure => AUTH_ERROR, user_error => 'saml_no_idp'})
    unless ($idp_name);

  my ($idp, $user, $email, $extern_id);

  eval { $idp = Bugzilla::Extension::SAML2Auth::IDP->new({name => $idp_name}); };
  return ({
    failure    => AUTH_ERROR,
    user_error => 'saml_no_such_idp',
    details    => {name => $idp_name}
  })
    unless ($idp);

  my $extern_id_field = $idp->saml_extern_id_attr();
  my $extern_id_set   = 0;

  if ($extern_id_field eq '') {
    $email
      = exists($assertation->{attributes}{Email})
      ? $assertation->{attributes}{Email}[0]
      : $assertation->{attributes}{email}[0];
    $email = $assertation->{attributes}{'urn:oid:1.2.840.113549.1.9.1'}[0]
      if (exists($assertation->{attributes}{'urn:oid:1.2.840.113549.1.9.1'}));

    $user = Bugzilla::User->new({name => $email});
  }
  else {
    return ({
      failure    => AUTH_ERROR,
      user_error => 'saml_failed_uuid',
      details    => {extern_id_field => $extern_id_field},
    })
      unless (exists $assertation->{attributes}{$extern_id_field});

    $extern_id = $assertation->{attributes}{$extern_id_field}[0];
    $user      = Bugzilla::User->new({extern_id => $extern_id});

    # No match, check LDAP
    if ($@ || !$user) {
      foreach my $adr (_ldap_search(
        $idp->ldap_extern_id_attr(), $extern_id, $idp->ldap_mail_attrs()))
      {
        $email = $adr unless ($email);                  # save the preferred address
        $user  = Bugzilla::User->new({name => $adr});
        if ($user && $user->is_enabled()) {
          $email = $adr;
          last;
        }
      }
    }
    else {
      $email         = $user->login;
      $extern_id_set = 1;
    }
  }

  my $regexp = $idp->userregexp;

## FIXME what to do if there is no email address in the response? Some customers don't have one apparently ...

  return ({
    failure    => AUTH_ERROR,
    user_error => 'saml_failed_regex',
    details    => {email => $email, regexp => $regexp, comment => $idp->comment}
  })
    unless ($regexp eq '' || $email =~ /$regexp/gi);

  if (!$user) {

    # BUGBUG when we do this we must ensure that disabled accounts cannot be opened
    # you have to contact an admin for that because you may have been banned
    $self->{saml2auth_failure} = {
      template => 'account/auth/saml2auth-verify-account.html.tmpl',
      vars     => {
        saml2email        => $email,
        saml2_ldap_fields => $idp->ldap_mail_attrs(),
        saml2_idp_name    => $idp->name,
      },
    };
    return ({failure => AUTH_NODATA});
  }
  else {
    if ($user->is_bot()) {
      return ({
        failure => AUTH_ERROR, user_error => 'no_login_for_bots', details => {},
      });
    }

    my $group_skip_attr  = $idp->group_skip_attr();
    my $group_skip_value = $idp->group_skip_value();

    my $skip
      = $group_skip_attr ne ''
      && $group_skip_value ne ''
      && exists($assertation->{attributes}{$group_skip_attr})
      && $assertation->{attributes}{$group_skip_attr}[0] eq $group_skip_value;

    unless ($skip) {
      ## IDP groups specify who is NOT allowed to use it!
      foreach my $gid (@{$idp->groups()}) {
        my $group = Bugzilla::Group->new($gid);
        return ({
          failure    => AUTH_ERROR,
          user_error => 'saml_failed_group',
          details => {idp => $idp->name, group => $group->name, comment => $idp->comment}
        })
          if ($user->in_group($group->name));
      }
    }

    my %ret = (saml2_auth => 1, username => $email,);

    if ($extern_id_set) {
      $ret{user_id} = $user->id;
      unless ($user->is_enabled) {
        $ret{force_enable} = 1;
      }
    }
    elsif ($user->is_enabled) {
      $ret{extern_id} = $extern_id;
    }

    Bugzilla->request_cache->{saml2_idp_name} = $idp_name;
    return (\%ret);
  }

  return ({failure => AUTH_ERROR});
}

sub login_uri {
  my ($self) = @_;
  my $cgi = Bugzilla->cgi;

  my $idp_name = $cgi->param('idp');
  trick_taint($idp_name) if ($idp_name);

  return ({failure => AUTH_ERROR, user_error => 'saml_no_idp'}) unless $idp_name;

  my $idp = idp($idp_name);
  my $sp  = sp();

  my $authn
    = $sp->authn_request(
    $idp->sso_url('urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'),
    $idp->format);
  my $authnreq = $authn->as_xml;
  my $bt       = Bugzilla->params->{'bugzilla_term'};
  $authnreq =~ s/My SP's human readable name./$bt/g;
  my $src_url = $sp->url();

  ## FIXME Net::SAML2::Protocol::AuthnRequest should allow optional metadata
  my $acsurl = $src_url . 'saml2_acs.cgi';
  $authnreq
    =~ s{IssueInstant}{AssertionConsumerServiceURL="$acsurl" IssueInstant};

  my $redirect = $sp->sso_redirect_binding($idp, 'SAMLRequest');

  my $cb_url = URI->new($cgi->param('target') || 'index.cgi');
  my $path   = $cb_url->as_string;

  # we only use relative URLs so if it has a host it's fake
  if ($cb_url->authority()) {
    ThrowUserError('saml_target_url_attack', {target => $path});
  }

  $path =~ s{^/}{};           # remove leading slash
  $path =~ s/\?logout=1//;    # remove logout command

  # we never want to redirect to a saml2 page
  $path = 'index.cgi' if ($path =~ m/saml2_[^.]*\.cgi/);

  my $url = $redirect->sign($authnreq, "$src_url$path");
  Bugzilla->logger->info("redirecting to: $url");

  # create a token with the AuthnRequest ID as the token.
  $self->_create_token($authn->id, $idp_name);

  print($cgi->redirect(-uri => $url));
  exit;
}

sub fail_nodata {
  my ($self)   = @_;
  my $cgi      = Bugzilla->cgi;
  my $template = Bugzilla->template;

  ThrowUserError('login_required')
    if (Bugzilla->usage_mode != USAGE_MODE_BROWSER);

  my $file = $self->{saml2auth_failure}{template}
    // "account/auth/login.html.tmpl";
  my $vars = $self->{saml2auth_failure}{vars}
    // {target => $cgi->url(-relative => 1)};

  print($cgi->header());
  $template->process($file, $vars) or ThrowTemplateError($template->error());
  exit;
}

sub _create_token {
  my ($self, $token, $idp) = @_;

  my $dbh = Bugzilla->dbh;

  $dbh->bz_start_transaction();

  $dbh->do(
    "INSERT INTO tokens (issuedate, token, tokentype, eventdata)
        VALUES (NOW(), ?, ?, ?)", undef, ($token, 'saml2auth_login', $idp)
  );

  $dbh->bz_commit_transaction();

  return;
}

sub _ldap_search {
  my ($match_attr, $value, $mail_attrs) = @_;
  my @addrs;

  # BUGBUG these are in the redhat extension, there are other
  # LDAP settings, but they don't allow a password AFAICT
  my $ldap_host   = Bugzilla->params->{LDAPMXserver};
  my $ldap_binddn = Bugzilla->params->{LDAPMXBaseDN};
  my $ldap_pass   = Bugzilla->localconfig->{LDAPMXpass};
  my $ldap_basedn = Bugzilla->params->{LDAPBaseDN};
  my $ldap_attr   = Bugzilla->params->{LDAPMXmailattribute};

  my $ldap = Net::LDAP->new($ldap_host, scheme => 'ldaps')
    || ThrowUserError('ldap_error', {error => $@});
  my $mesg = $ldap->bind($ldap_binddn, password => $ldap_pass);
  $mesg->code && ThrowUserError('ldap_error', {error => $mesg->error});
  $mesg = $ldap->search(    # perform a search
    base   => $ldap_basedn,
    filter => "($match_attr=$value)"
  );
  $mesg->code && ThrowUserError('ldap_error', {error => $mesg->error});

  foreach my $entry ($mesg->entries) {

    foreach my $atr (split /[,\s]+/, $mail_attrs) {
      my $adr = $entry->get_value($atr);
      push(@addrs, $adr) if ($adr && validate_email_syntax($adr));
    }
  }

  $mesg = $ldap->unbind;

  return (@addrs);
}

1;

__END__

=head1 Description

Bugzilla::Extension::SAML2Auth::Login - A module for logging in to an IDP.

=head1 Methods

=over 4

=item get_login_info

Maps IDP Login data to Bugzilla accounts.

=item login_uri

Generates a SAML authentication request, signs it, and returns the IDPs login redirect URL with the SAML Request embedded.

=item fail_nodata

Handles authentication errors from the IDP.

=back

