# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::SubComponents::RHSubComponent;

use strict;
use warnings;
use 5.10.1;
use base qw(Bugzilla::Object);

use Bugzilla::Constants;
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::User;
use Bugzilla::Field;
use Bugzilla::Field::Choice;
use Bugzilla::FlagType;
use Bugzilla::Series;

use Scalar::Util qw(blessed);
use List::MoreUtils qw(any);

###############################
####    Initialization     ####
###############################

use constant DB_TABLE => 'rh_sub_components';

# This is mostly for the editfields.cgi case where ->get_all is called.
use constant LIST_ORDER => 'level, sortkey, name, id';

use constant DEFAULT_SORTKEY => 0;

use constant DB_COLUMNS => qw(
  id
  name
  component_id
  parent_id
  initialowner
  initialqacontact
  initialdocscontact
  description
  isactive
  sortkey
  level
  full_name
  default_to_private
);

use constant UPDATE_COLUMNS => qw(
  name
  parent_id
  initialowner
  initialqacontact
  initialdocscontact
  description
  isactive
  sortkey
  level
  full_name
  default_to_private
);

use constant REQUIRED_FIELD_MAP => {component_id => 'component',};

## REDHAT EXTENSION 876015: Add initialdocscontact
use constant VALIDATORS => {
  create_series      => \&Bugzilla::Object::check_boolean,
  component          => \&_check_component,
  initialowner       => \&_check_initialowner,
  initialqacontact   => \&_check_initialqacontact,
  initialdocscontact => \&_check_initialdocscontact,
  description        => \&_check_description,
  initial_cc         => \&_check_cc_list,
  name               => \&_check_name,
  isactive           => \&_check_isactive,
  parent_id          => \&_check_parent_id,
  sortkey            => \&_check_sortkey,
  default_to_private => \&Bugzilla::Object::check_boolean,
};

use constant VALIDATOR_DEPENDENCIES =>
  {name => ['component', 'parent_id'], parent_id => ['component'],};

###############################

sub create {
  my $class = shift;
  my $dbh   = Bugzilla->dbh;

  $dbh->bz_start_transaction();

  $class->check_required_create_fields(@_);
  my $params    = $class->run_create_validators(@_);
  my $cc_list   = delete $params->{initial_cc};
  my $component = delete $params->{component};
  $params->{component_id} = $component->id;

  my $sub_component = $class->insert_create_data($params);
  $sub_component->{component} = $component;

  # We still have to fill the component_cc table.
  $sub_component->_update_cc_list($cc_list) if $cc_list;

  # We also need to update the two meta columns (level and full_name)
  $sub_component->update();

  $dbh->bz_commit_transaction();
  return $sub_component;
}

sub update {
  my $self = shift;
  $self->update_meta();
  my $changes = $self->SUPER::update(@_);

  # Update the component_cc table if necessary.
  if (defined $self->{cc_ids}) {
    my $diff = $self->_update_cc_list($self->{cc_ids});
    ## REDHAT EXTENSION START 1104055
    if (defined $diff) {
      $changes->{cc_list} = $diff;
      $self->audit_log({initialcc => $diff}) if $self->AUDIT_UPDATES;
    }
    ## REDHAT EXTENSION END 1104055
  }
  return $changes;
}

sub remove_from_db {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  $dbh->bz_start_transaction();

  if ($self->bug_count) {
    if (Bugzilla->params->{'allowbugdeletion'}) {
      require Bugzilla::Bug;
      foreach my $bug_id (@{$self->bug_ids}) {

        # Note: We allow admins to delete bugs even if they can't
        # see them, as long as they can see the product.
        my $bug = new Bugzilla::Bug($bug_id);
        $bug->remove_from_db();
      }
    }
    else {
      ThrowUserError('component_has_bugs', {nb => $self->bug_count});
    }
  }

  $self->SUPER::remove_from_db();

  $dbh->bz_commit_transaction();

  return;
}

################################
# Validators
################################

sub _check_name {
  my ($invocant, $name, undef, $params) = @_;
  my $component
    = blessed($invocant) ? $invocant->component : $params->{component};
  my $parent_id
    = blessed($invocant) ? $invocant->parent_id : $params->{parent_id};

  $name = trim($name);
  $name || ThrowUserError('component_blank_name');

  if (length($name) > MAX_COMPONENT_SIZE) {
    ThrowUserError('component_name_too_long', {'name' => $name});
  }

  my $subcomponents = Bugzilla::Extension::SubComponents::RHSubComponent->match({
    component_id => $component->id,
    parent_id    => ($parent_id || IS_NULL),
    name         => $name,
  });

  if (scalar(@$subcomponents)
    && (!ref $invocant || $subcomponents->[0]->id != $invocant->id))
  {
    ThrowUserError('sub_component_already_exists',
      {name => $name, comp => $component});
  }
  return $name;
}

sub _check_description {
  my ($invocant, $description) = @_;

  $description = trim($description);
  $description || ThrowUserError('component_blank_description');
  return $description;
}

sub _check_initialowner {
  my ($invocant, $owner) = @_;

  $owner || ThrowUserError('component_need_initialowner');
  my $owner_id = Bugzilla::User->check($owner)->id;
  return $owner_id;
}

sub _check_initialqacontact {
  my ($invocant, $qa_contact) = @_;

  my $qa_contact_id;
  if (Bugzilla->params->{'useqacontact'}) {
    $qa_contact_id = Bugzilla::User->check($qa_contact)->id if $qa_contact;
  }
  elsif (ref $invocant) {
    $qa_contact_id = $invocant->{initialqacontact};
  }
  return $qa_contact_id;
}

## REDHAT EXTENSION START 876015
sub _check_initialdocscontact {
  my ($invocant, $docs_contact) = @_;

  my $docs_contact_id;
  if (Bugzilla->params->{'usedocscontact'}) {
    $docs_contact_id = Bugzilla::User->check($docs_contact)->id if $docs_contact;
  }
  elsif (ref $invocant) {
    $docs_contact_id = $invocant->{initialdocscontact};
  }
  return $docs_contact_id;
}
## REDHAT EXTENSION END 876015

sub _check_component {
  my ($invocant, $component) = @_;
  $component
    || ThrowCodeError('param_required',
    {function => "$invocant->create", param => 'component'});
  Bugzilla->user->check_can_admin_product($component->product->name);
  return $component;
}

sub _check_cc_list {
  my ($invocant, $cc_list) = @_;

  my %cc_ids;
  foreach my $cc (@$cc_list) {
    my $id = login_to_id($cc, THROW_ERROR);
    $cc_ids{$id} = 1;
  }
  return [keys %cc_ids];
}

sub _check_isactive {
  my ($invocant, $isactive, undef, $params) = @_;

  return 1 if defined $isactive and $isactive;

  # It is inactive. Check that it doesn't have any children
  if (blessed($invocant)) {
    if (any { $_->is_active } @{$invocant->children}) {
      ThrowUserError('sub_component_inactive_children');
    }
  }

  return 0;
}

sub _check_parent_id {
  my ($invocant, $parent_id, undef, $params) = @_;

  return if ((not defined $parent_id) or ($parent_id eq ''));

  # Keep a copy in case detaint_signed() clears the parent_id
  my $component
    = blessed($invocant) ? $invocant->component : $params->{component};
  my $stored_parent_id = $parent_id;

  if ( !detaint_signed($parent_id)
    || $parent_id < MIN_SMALLINT
    || $parent_id > MAX_SMALLINT)
  {
    ThrowUserError('sub_component_parent_id_invalid',
      {parent_id => $stored_parent_id});
  }

  # Check that the parent is not itself
  if (blessed($invocant) && $invocant->id == $parent_id) {
    ThrowUserError('sub_component_parent_not_itself');
  }

  # Check that the parent exists and belongs to the same component
  my $parent = Bugzilla::Extension::SubComponents::RHSubComponent->new($parent_id)
    || ThrowUserError('sub_component_parent_not_exists');

  if ($parent->component_id != $component->id) {
    ThrowUserError('sub_component_parent_not_same');
  }

  unless ($parent->is_active) {
    ThrowUserError('sub_component_parent_not_active');
  }

  # Check that we are not creating a loop (A is parent of B, B of C, C of A)
  my %seen = ();
  my $loop = 0;

  if (blessed($invocant)) {
    $seen{$invocant->id}++;
  }
  while (++$loop) {
    if ($seen{$parent->id}++) {
      ThrowUserError('sub_component_loop');
    }
    elsif ($loop >= 50) {
      ThrowCodeError('sub_component_possible_loop', {sub_component => $parent});
    }

    $parent = $parent->parent;
    last unless defined $parent;
  }

  return $parent_id;
}

sub _check_sortkey {
  my ($invocant, $sortkey) = @_;

  # Keep a copy in case detaint_signed() clears the sortkey
  my $stored_sortkey = $sortkey;

  if ( !detaint_signed($sortkey)
    || $sortkey < MIN_SMALLINT
    || $sortkey > MAX_SMALLINT)
  {
    ThrowUserError('sub_component_sortkey_invalid', {sortkey => $stored_sortkey});
  }
  return $sortkey;
}

###############################
####       Methods         ####
###############################

sub _update_cc_list {
  my ($self, $cc_list) = @_;
  my $dbh = Bugzilla->dbh;

  my $old_cc_list = $dbh->selectcol_arrayref(
    'SELECT user_id FROM rh_sub_component_cc
                                WHERE rh_sub_component_id = ?', undef, $self->id
  );

  my ($removed, $added) = diff_arrays($old_cc_list, $cc_list);
  my $diff;
  if (scalar @$removed || scalar @$added) {
    $diff = [join(', ', @$removed), join(', ', @$added)];
  }

  $dbh->do('DELETE FROM rh_sub_component_cc WHERE rh_sub_component_id = ?',
    undef, $self->id);

  my $sth = $dbh->prepare(
    'INSERT INTO rh_sub_component_cc
                             (user_id, rh_sub_component_id) VALUES (?, ?)'
  );
  $sth->execute($_, $self->id) foreach (@$cc_list);

  return $diff;
}

sub set_name               { $_[0]->set('name',               $_[1]); return; }
sub set_description        { $_[0]->set('description',        $_[1]); return; }
sub set_is_active          { $_[0]->set('isactive',           $_[1]); return; }
sub set_sortkey            { $_[0]->set('sortkey',            $_[1]); return; }
sub set_parent_id          { $_[0]->set('parent_id',          $_[1]); return; }
sub set_default_to_private { $_[0]->set('default_to_private', $_[1]); return; }

sub set_default_assignee {
  my ($self, $owner) = @_;

  $self->set('initialowner', $owner);

  # Reset the default owner object.
  delete $self->{default_assignee};

  return;
}

sub set_default_qa_contact {
  my ($self, $qa_contact) = @_;

  $self->set('initialqacontact', $qa_contact);

  # Reset the default QA contact object.
  delete $self->{default_qa_contact};

  return;
}

## REDHAT EXTENSION START 876015
sub set_default_docs_contact {
  my ($self, $docs_contact) = @_;

  $self->set('initialdocscontact', $docs_contact);

  # Reset the default Docs contact object.
  delete $self->{default_docs_contact};

  return;
}
## REDHAT EXTENSION END 876015

sub set_cc_list {
  my ($self, $cc_list) = @_;

  $self->{cc_ids} = $self->_check_cc_list($cc_list);

  # Reset the list of CC user objects.
  delete $self->{initial_cc};

  return;
}

sub update_meta {
  my $self      = shift;
  my $name_tree = $self->name_tree;

  $self->set('full_name', join(' ', @$name_tree));
  $self->set('level',     scalar(@$name_tree) - 1);

  return;
}

sub bug_count {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{'bug_count'}) {
    ## REDHAT EXTENSION BEGIN 584957
    $self->{'bug_count'} = $dbh->selectrow_array(
      qq{
                SELECT COUNT(*)
                FROM bug_rh_sub_components
                WHERE rh_sub_component_id = ?
            }, undef, $self->id,
    ) || 0;
    ## REDHAT EXTENSION END 584957
  }
  return $self->{'bug_count'};
}

sub bug_ids {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{'bugs_ids'}) {
    $self->{'bugs_ids'} = $dbh->selectcol_arrayref(
      q{
            SELECT bug_id FROM bug_rh_sub_components
            WHERE rh_sub_component_id = ?}, undef, $self->id
    );
  }
  return $self->{'bugs_ids'};
}

sub default_assignee {
  my $self = shift;

  if (!defined $self->{'default_assignee'}) {
    $self->{'default_assignee'} = new Bugzilla::User($self->{'initialowner'});
  }
  return $self->{'default_assignee'};
}

sub default_qa_contact {
  my $self = shift;

  return if !$self->{'initialqacontact'};

  if (!defined $self->{'default_qa_contact'}) {
    $self->{'default_qa_contact'} = new Bugzilla::User($self->{'initialqacontact'});
  }
  return $self->{'default_qa_contact'};
}

## REDHAT EXTENSION START 876015
sub default_docs_contact {
  my $self = shift;

  return if !$self->{'initialdocscontact'};

  if (!defined $self->{'default_docs_contact'}) {
    $self->{'default_docs_contact'}
      = new Bugzilla::User($self->{'initialdocscontact'});
  }
  return $self->{'default_docs_contact'};
}
## REDHAT EXTENSION END 876015

sub initial_cc {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (!defined $self->{'initial_cc'}) {

    # If set_cc_list() has been called but data are not yet written
    # into the DB, we want the new values defined by it.
    my $cc_ids = $self->{cc_ids} || $dbh->selectcol_arrayref(
      'SELECT user_id FROM rh_sub_component_cc
                                                  WHERE rh_sub_component_id = ?',
      undef, $self->id
    );

    $self->{'initial_cc'} = Bugzilla::User->new_from_list($cc_ids);
  }
  return $self->{'initial_cc'};
}

sub component {
  my $self = shift;
  if (!defined $self->{'component'}) {
    require Bugzilla::Component;    # We cannot |use| it.
    $self->{'component'} = new Bugzilla::Component($self->component_id);
  }
  return $self->{'component'};
}

###############################
####      Accessors        ####
###############################

sub description        { return $_[0]->{'description'}; }
sub component_id       { return $_[0]->{'component_id'}; }
sub parent_id          { return $_[0]->{'parent_id'}; }
sub sortkey            { return $_[0]->{'sortkey'}; }
sub is_active          { return $_[0]->{'isactive'}; }
sub level              { return $_[0]->{'level'}; }                # Level 0 = top level
sub full_name          { return $_[0]->{'full_name'}; }
sub default_to_private { return $_[0]->{'default_to_private'}; }

sub full_name_pretty {
  my $self = shift;
  return (join('<i class="fas fa-arrow-right"></i>',
    ($self->component->name, @{$self->name_tree})));
}

sub cc_list_string {
  my $self = shift;

  my $str = '';
  $str = join(', ', map { $_->login } @{$self->initial_cc})
    if ($self->initial_cc);
  return ($str);
}

sub type {
  return 'sub_component';
}

sub product_id { return $_[0]->component->product_id; }

###############################
####      Subroutines      ####
###############################

sub new_from_list {
  my $self = shift;
  my $list = $self->SUPER::new_from_list(@_);
  return [sort { $a->sortkey <=> $b->sortkey || $a->name cmp $b->name } @$list];
}

sub children {
  my $self = shift;
  my $dbh  = Bugzilla->dbh;

  if (not defined $self->{children}) {
    my $query = qq{
            SELECT id
            FROM rh_sub_components
            WHERE component_id = ? AND parent_id = ?
        };

    my $ids
      = $dbh->selectcol_arrayref($query, undef, $self->component_id, $self->id);
    $self->{children} = $self->new_from_list($ids);
  }

  return $self->{children};
}

sub parent {
  my $self = shift;
  return $self->parent_id ? $self->new($self->parent_id) : undef;
}

sub name_tree {
  my $self = shift;

  my $here = $self;
  my @tree = ($self->name);
  my $loop = 0;
  while (++$loop) {
    $here = $here->parent;
    last unless defined $here;
    unshift @tree, $here->name;

    if ($loop >= 50) {
      ThrowCodeError('sub_component_possible_loop', {sub_component => $self});
    }
  }

  return \@tree;
}

1;

__END__

=head1 NAME

Bugzilla::Extension::SubComponents::RHSubComponent - Bugzilla sub component class.

=head1 SYNOPSIS

    use Bugzilla::Extension::SubComponents::RHSubComponent;

    my $sub_component = new Bugzilla::Extension::SubComponents::RHSubComponent($comp_id);

    my $bug_count            = $sub_component->bug_count();
    my $bug_ids              = $sub_component->bug_ids();
    my $id                   = $sub_component->id;
    my $name                 = $sub_component->name;
    my $description          = $sub_component->description;
    my $component_id         = $sub_component->component_id;
    my $default_assignee     = $sub_component->default_assignee;
    my $default_qa_contact   = $sub_component->default_qa_contact;
    my $default_docs_contact = $sub_component->default_docs_contact;
    my $initial_cc           = $sub_component->initial_cc;
    my $is_active            = $sub_component->is_active;
    my $parent_id            = $sub_component->parent_id;
    my $level                = $sub_component->level;
    my $full_name            = $sub_component->full_name;
    my $sortkey              = $sub_component->sortkey;
    my $parent               = $sub_component->sortkey;
    my $name_tree            = $sub_component->sortkey;
    my $children             = $sub_component->sortkey;
    my $component            = $sub_component->component;

    my $sub_component =
      Bugzilla::Extension::SubComponents::RHSubComponent->create({
                                    name             => $name,
                                    component        => $component,
                                    initialowner     => $user_login1,
                                    initialqacontact => $user_login2,
                                    initialdocscontact => $user_login3,
                                    description      => $description});

    $sub_component->set_name($new_name);
    $sub_component->set_description($new_description);
    $sub_component->set_default_assignee($new_login_name);
    $sub_component->set_default_qa_contact($new_login_name);
    $sub_component->set_default_docs_contact($new_login_name);
    $sub_component->set_cc_list(\@new_login_names);
    $sub_component->set_sortkey($new_sortkey);
    $sub_component->update();

    $sub_component->remove_from_db;

=head1 DESCRIPTION

RHSubComponent.pm represents a Sub Component object.

=head1 METHODS

=over 4

=item C<new($id)>

 Description: The constructor is used to load an existing sub component
              by passing a sub component id.

 Params:      $id - The id of the sub component you want to load

 Returns:     A Bugzilla::Extension::SubComponents::RHSubComponent object.

=item C<bug_count()>

 Description: Returns the total of bugs that belong to the sub component.

 Params:      none.

 Returns:     Integer with the number of bugs.

=item C<bug_ids()>

 Description: Returns all bug IDs that belong to the sub component.

 Params:      none.

 Returns:     A reference to an array of bug IDs.

=item C<default_assignee()>

 Description: Returns a user object that represents the default assignee for
              the sub component.

 Params:      none.

 Returns:     A Bugzilla::User object.

=item C<default_qa_contact()>

 Description: Returns a user object that represents the default QA contact for
              the sub component.

 Params:      none.

 Returns:     A Bugzilla::User object if the default QA contact is defined for
              the sub component. Otherwise, returns undef.

=item C<default_docs_contact()>

 Description: Returns a user object that represents the default Docs contact for
              the sub component.

 Params:      none.

 Returns:     A Bugzilla::User object if the default Docs contact is defined for
              the sub component. Otherwise, returns undef.

=item C<initial_cc>

 Description: Returns a list of user objects representing users being
              in the initial CC list.

 Params:      none.

 Returns:     An arrayref of L<Bugzilla::User> objects.

=item C<parent_id()>

 Description: Returns the id of the parent sub component of this sub component.

 Params:      none.

 Returns:     The id of the parent sub component, or undef if there is no parent.

=item C<parent()>

 Description: Returns the parent sub component the sub component belongs to.

 Params:      none.

 Returns:     A Bugzilla::Extension::SubComponents::RHSubComponent object,
              or undef if there is no parent.

=item C<children()>

 Description: Returns the child sub components of this sub component.

 Params:      none.

 Returns:     A arrayref of Bugzilla::Extension::SubComponents::RHSubComponent
              objects, or an empty arrayref if there are no children.

=item C<full_name()>

 Description: Returns the full name of the sub component.

 Params:      none.

 Returns:     A string. Levels are separated with a space

=item C<name_tree()>

 Description: Returns the full name of the sub component.

 Params:      none.

 Returns:     An array ref of the full name of the tree

=item C<level()>

 Description: Returns the level of the sub component

 Params:      none.

 Returns:     An integer. 0 is the top level, 1 means it has one parent, and
              so on.

=item C<component()>

 Description: Returns the component the sub component belongs to.

 Params:      none.

 Returns:     A Bugzilla::Component object.

=item C<set_name($new_name)>

 Description: Changes the name of the sub component.

 Params:      $new_name - new name of the sub component (string). This name
                          must be unique within the component and sub
                          component level.

 Returns:     Nothing.

=item C<set_description($new_desc)>

 Description: Changes the description of the sub component.

 Params:      $new_desc - new description of the sub component (string).

 Returns:     Nothing.

=item C<set_default_assignee($new_assignee)>

 Description: Changes the default assignee of the sub component.

 Params:      $new_owner - login name of the new default assignee of
                           the sub component (string). This user account
                           must already exist.

 Returns:     Nothing.

=item C<set_default_qa_contact($new_qa_contact)>

 Description: Changes the default QA contact of the sub component.

 Params:      $new_qa_contact - login name of the new QA contact of
                                the sub component (string). This user
                                account must already exist.

 Returns:     Nothing.

=item C<set_default_docs_contact($new_docs_contact)>

 Description: Changes the default Docs contact of the sub component.

 Params:      $new_docs_contact - login name of the new Docs contact of
                                the sub component (string). This user
                                account must already exist.

 Returns:     Nothing.

=item C<set_cc_list(\@cc_list)>

 Description: Changes the list of users being in the CC list by default.

 Params:      \@cc_list - list of login names (string). All the user
                          accounts must already exist.

 Returns:     Nothing.

=item C<set_sortkey($new_sortkey)>

 Description: Changes the sortkey of the sub component.

 Params:      $new_sortkey - new sortkey of the sub component (signed integer).

 Returns:     Nothing.

=item C<update()>

 Description: Write changes made to the sub component into the DB.

 Params:      none.

 Returns:     A hashref with changes made to the sub component object.

=item C<remove_from_db()>

 Description: Deletes the current sub component from the DB. The object itself
              is not destroyed. Any bugs with this sub component will have
              their sub component removed.

 Params:      none.

 Returns:     Nothing.

=back

=head1 CLASS METHODS

=over

=item C<create(\%params)>

 Description: Create a new sub component for the given product.

 Params:      The hashref must have the following keys:
              name             - name of the new sub component (string). This
                                 name must be unique within the parent
                                 sub component (or component at the top level)
              component        - a Bugzilla::Component object to which
                                 the Sub component is being added.
              description      - description of the new sub component (string).
              initialowner     - login name of the default assignee (string).
              The following keys are optional:
              initialqacontact - login name of the default QA contact (string),
                                 or an empty string to clear it.
              initialdocscontact - login name of the default Docs contact (string),
                                   or an empty string to clear it.
              initial_cc       - an arrayref of login names to add to the
                                 CC list by default.

 Returns:     A Bugzilla::Extension::SubComponents::RHSubComponent object.

=back

=cut
