# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::APIKeyExpiry;

use strict;
use warnings;
use 5.10.1;

use parent qw(Exporter);

our @EXPORT = qw(
  api_key_expiry_bug_update
  api_key_expiry_key_update
);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util;
use Bugzilla::Extension::Workflows::Util;

use Data::Dumper;
use Carp qw/longmess/;

BEGIN {
  *Bugzilla::User::APIKey::expiry_bug = \&_apikey_expiry_bug;
}

sub api_key_expiry_bug_update {
  my ($args) = @_;
  my $dbh = Bugzilla->dbh;

  my ($bug, $old_bug, $timestamp, $changes)
    = @$args{qw(bug old_bug timestamp changes)};

  my $subject = $bug->short_desc;
  $subject =~ m/APIKey expiry for key (\d+)/;
  my $apikey_id = $1;
  my $apikey    = Bugzilla::User::APIKey->check({id => $apikey_id});

  return if ($apikey->banned);

  my $apikey_descr = $apikey->description // 'blank';
  my $apikey_login = $apikey->user->login;

  # if we have reached expiry
  if ($bug->bug_status eq 'CLOSED' && $bug->resolution eq 'COMPLETED') {

    $apikey->set_banned(1);
    $apikey->update();

    unless ($apikey->revoked) {
      my $msg
        = "The APIKey $apikey_id '$apikey_descr' for $apikey_login has been banned. To continue using the API you will need to create a new APIKey.";
      my $params = {comment => {body => $msg}};

      $bug->set_all($params);
      $bug->update();
    }
  }

  return;

}

sub _apikey_expiry_bug {
  my $self = shift();
  my $dbh  = Bugzilla->dbh;

  my $sql = q{
      SELECT bug_id
        FROM bugs
       WHERE bug_status <> 'CLOSED'
         AND short_desc LIKE ?
    };

  my $bug_ids = $dbh->selectcol_arrayref($sql, undef,
    'APIKey expiry for key ' . $self->id . ' %');
  my $bugs = Bugzilla::Bug->new_from_list($bug_ids);
  my $bug;
  $bug = $bugs->[0] if (@$bugs);

  return $bug;
}

sub api_key_expiry_key_update {
  my $args = shift;

  my ($object, $old_object, $changes) = @$args{qw(object old_object changes)};

  my $bug = $object->expiry_bug();

  if ($bug) {
    my $user = Bugzilla->login(LOGIN_REQUIRED);

    Bugzilla->set_user(Bugzilla::User->check(Bugzilla->params->{system_user}));

    my $login = $user->login;

    my %params = (
      bug_status   => 'CLOSED',
      resolution   => 'COMPLETED',
      minor_update => 1,
      comment      => {body => "This key was revoked by $login"},
    );

    $bug->set_all(\%params);
    $bug->update();

    Bugzilla->set_user($user);
  }

  return;
}

1;


__END__

=head1 NAME

Bugzilla::Extension::Workflows::APIKeyExpiry

=head1 DESCRIPTION

This workflow enables automated revocation and expiration of APIKeys.

=head2 The Banning Workflow

Banning is permanent, a banned APIKey can never be used again.

=over 4

=item The APIKey checking cron job detects an APIKey approaching expiry

=item A tracker bug is opened for each APIKey

For bot accounts the users in the Agile team managing the bot are CC'd on the
bug.

For other accounts the account is CC'd on the bug.

=item either:

=over 4

=item An owner revokes the key.

=over 4

=item The key is banned

=item The tracking bug has a comment added that the key was revoked

=item The tracker bug is closed

=back

=item The deadline is reached

=over 4

=item The key is banned

=item The tracking bug has a comment added that the key was banned

=item The tracking bug is closed

=back

=back

=back

=head2 The Revoking Workflow

Revocation is not permanent, a revoked key can be reenabled on the account's
APIKey preferences tab.

When the APIKey checking cron job is run it will automatically revoke any APIKey
that has not been used in 30 days and is not being processed by the Banning
Workflow.

An email notification is sent for this event.

For bot accounts the email is sent to all the users in the Agile team managing
the bot.

For other users the email is sent to the account's email address.

=cut
