# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::CreateBotAccount;

use strict;
use warnings;
use 5.10.1;

use parent qw(Exporter);

our @EXPORT = qw(
  page_create_bot
);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util;
use Bugzilla::Extension::Workflows::Util;
use Bugzilla::Token;

sub page_create_bot {
  my ($vars) = @_;
  my $dbh    = Bugzilla->dbh;
  my $input  = Bugzilla->input_params;
  my $cgi    = Bugzilla->cgi;
  my $action = $cgi->param('action');

  Bugzilla->login(LOGIN_REQUIRED);
  my $user = Bugzilla->user;
  $user->is_insider || ThrowUserError(
    "auth_failure",
    {
      group  => Bugzilla->params->{'insidergroup'},
      action => "Workflows",
      object => "create_bot"
    }
  );

  if ($action) {
    if ($action eq 'create') {
      my $token = $input->{'token'};
      check_token_data($token, 'create_bot_account');

      my $login = $cgi->param('login');
      my $name  = $cgi->param('name') // "";
      my $team  = $cgi->param('team') || ThrowUserError('team_required');

      my $admin_group = Bugzilla::Group->check({name => $team});

      $admin_group->category() eq 'Agile'
        || ThrowUserError("group_not_a_team", {name => $admin_group->name});

      $user->in_group($admin_group)
        || ThrowUserError("user_not_in_team", {name => $admin_group->name});

      my $login_name = $login . Bugzilla->params->{bot_name_regex};
      $login_name =~ s{\\}{}g;
      $login_name =~ s{\$}{}g;

      # Switch to system user as it requires elevated privileges to do this
      Bugzilla->set_user(
        Bugzilla::User->new({name => Bugzilla->params->{system_user}}));

      $dbh->bz_start_transaction(1);

      my $new_user
        = Bugzilla::User->create({
        login_name => $login_name, realname => $name, cryptpassword => '*',
        });

      $new_user->set_prefs_admin_group($admin_group->id);

      my @add_groups = $cgi->param('groups');
      if (@add_groups) {
        foreach my $group_name (@add_groups) {
          $user->in_group($group_name)
            || ThrowUserError("user_not_in_group", {name => $group_name});
        }

        $new_user->set_groups({add => \@add_groups});
      }

      $new_user->update();

      # Switch back to real user
      Bugzilla->set_user($user);

      my @add_teams = $cgi->param('managed_teams');
      if (@add_teams) {
        foreach my $team_name (@add_teams) {
          my $add_team
            = Bugzilla::Extension::AgileTools::Team->check({name => $team_name});
          $add_team->add_member($new_user);
        }
      }

      $dbh->bz_commit_transaction();

      # clean out bot list as there is a new one
      delete $user->{editable_bots};

      delete_token($token);

      $vars->{'message'}  = 'bot_account_created';
      $vars->{'new_user'} = $new_user;
    }
  }

  $vars->{'token'} = issue_session_token('create_bot_account');

  my @teams;
  foreach my $team (@{$user->agile_teams}) {
    push(@teams, $team->group());
  }

  my $gr_regex = Bugzilla->params->{bot_groups_regex};
  my @groups   = grep {
    $_->category() ne 'Agile'
      && ($_->{direct} || $_->name =~ /$gr_regex/i)
  } @{$user->groups};

  $vars->{'teams'}         = \@teams;
  $vars->{'groups'}        = \@groups;
  $vars->{'managed_teams'} = $user->editable_agile_teams;

  return;
}


1;


__END__

=head1 NAME

Bugzilla::Extension::Workflows::CreateBotAccount

=head1 DESCRIPTION

This workflow allows insiders to create bot accounts. These accounts can only
use the API and are owned by an Agile team.

While it is possible for teams to create numerous bots it is often a better idea
to have a separate API_KEY for each use of the account. If the API_KEYs have
suitable names access for external services can be retracted as required.

To start this workflow go to the B<My Links> -> B<Workflows> -> B<Create Bot Account> menu.

To manage bot accounts, create API_KEYs, etc, you go to the B<Administration>
page and click the B<Edit Bot Preferences> link.

=cut
