# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::WebService::GroupInvite;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::WebService::Util qw(validate);
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Object;
use Bugzilla::Extension::Workflows::GroupInvite;

use Bugzilla::WebService::Bug;

use constant PUBLIC_METHODS => qw(
  invite_users
);


=head1 NAME

Bugzilla::Extension::Workflows::WebService::GroupInvite

=head1 DESCRIPTION

This part of the Bugzilla API allows you to invite users to groups and teams.

=head1 METHODS

=head2 invite_users

Invite users to groups and teams.

=head3 Params

groups:
  An array of B<strings>. The names of the groups and teams to add users to.

users:
  An array of B<strings>. The names of the users to affect.

=head3 Returns

Returns an array of hashes, one per user changed, with the
following fields:

groups:
  C<array> An array of B<strings>. If the user was added directly to any groups they will be listed here.

teams:
  C<array> An array of B<strings>. If the user was added to any teams they will be listed here.

bug:
  C<hash> If a Group Membership Workflow bugs was created it will be detailed here.

=head3 Examples

=head4 JSONRPC call

  {
      "jsonrpc": "2.0",
      "method": "GroupInvite.invite_users",
      "id": 1,
      "params": [
        {
          "users": ["fu@example.com", "bar@example.com"],
          "groups": ["TheBigGroup", "TheLittleGroup", "A-Team"]
        }
      ]
  }

=head4 JSONRPC reply

  {
      "id": 1,
      "error": null,
      "result": [
        "fu@example.com":  {
            "groups": ["TheLittleGroup"],
            "teams": ["A-Team"],
            "bug": {
              "id": 111,
              "summary": "bz_NEW bz_unspecified bz_unspecified bz_secure bz_requested",
              "bug_classes": ""Group membership request for fu@example.com"
            }
          },
         "bar@example.com": {
            "groups": ["TheLittleGroup"],
            "teams": ["A-Team"],
            "bug": {
              "id": 112,
              "summary": ""Group membership request for bar@example.com",
              "bug_classes": "bz_NEW bz_unspecified bz_unspecified bz_secure bz_requested"
            }
          }
      ]
  }


=cut

## BUGBUG fix FAQ

sub invite_users {
  my ($self, $args) = validate(@_);

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $dbh = Bugzilla->dbh;

  my $group_names = $args->{groups}
    // ThrowCodeError('param_required', {param => 'groups'});

  my $user_names = $args->{users}
    // ThrowCodeError('param_required', {param => 'users'});

  my @group_objects = map { Bugzilla::Group->check({name => $_}); } @$group_names;

  my @user_objects
    = map { Bugzilla::User->check($_) } grep { $_ ne '' } @$user_names;

  my @add_groups;
  my @add_teams;

  foreach my $group (@group_objects) {
    if ($group->my_team()) {
      push(@add_teams, $group->my_team());
    }
    else {
      push(@add_groups, $group->name);
    }
  }

  # All or nothing
  $dbh->bz_start_transaction();
  my $changes;

  foreach my $otherUser (@user_objects) {
    if (@add_groups) {
      $otherUser->set_groups({add => \@add_groups});
      if ($otherUser->{_group_changes}) {
        my $grp_changes = $otherUser->update();
        if (defined($grp_changes->{groups})) {
          $changes->{$otherUser->login}->{groups} = $grp_changes->{groups};
        }
      }

      if ($otherUser->{_workflow_bug}) {
        $changes->{$otherUser->login}->{bug} = {
          id          => $otherUser->{_workflow_bug}->id,
          bug_classes => join(" ", @{$otherUser->{_workflow_bug}->classes}),
          summary     => $otherUser->{_workflow_bug}->short_desc,
        };
      }
    }

    foreach my $team (@add_teams) {
      if ($team->add_member($otherUser)) {
        push(
          @{$changes->{$otherUser->login}->{teams}},
          {name => $team->name, id => $team->id}
        );
      }
    }
  }
  $dbh->bz_commit_transaction();

  return $changes;
}

1;

__END__


