# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

package Bugzilla::Extension::Workflows::WebService::GroupRequest;

use 5.10.1;
use strict;
use warnings;
use base qw(Bugzilla::WebService);

use Bugzilla;
use Bugzilla::Constants;
use Bugzilla::WebService::Util qw(validate);
use Bugzilla::Util;
use Bugzilla::Error;
use Bugzilla::Object;
use Data::Dumper;
use Bugzilla::Extension::Workflows::GroupRequest;

use constant PUBLIC_METHODS => qw(
  update_group
  get_groups
  invite_manager
);

sub update_group {
  my ($self, $args) = @_;

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  $user->can_admin_group_workflow();

  my $group_name = $args->{group}
    // ThrowCodeError('param_required', {param => 'group'});
  my $managed     = $args->{managed}     // 0;
  my $regex       = $args->{regex}       // '';
  my $invite_only = $args->{invite_only} // 0;
  my $approval_group_name = $args->{approval_group};
  my $cc_list             = $args->{cc_list} // '';

  my $dbh = Bugzilla->dbh;

  $dbh->bz_start_transaction();

  my $group = Bugzilla::Group->check({name => $group_name});

  my $gwf = $group->group_workflow;

  my %params = (
    managed     => $managed,
    invite_only => $invite_only,
    regex       => $regex,
    group_id    => $group->id,
  );

  if ($gwf) {
    $gwf->set_all(\%params);
  }
  else {
    $gwf = Bugzilla::Extension::Workflows::GroupRequest->create(\%params);
  }

  ThrowUserError('approvals_approval_group_required')
    if ($invite_only && (!$approval_group_name || $approval_group_name eq ''));

  my $bu_regex = Bugzilla->params->{bot_or_user_regex};

  if (
       $managed
    && ((!$cc_list) || $cc_list eq '')
    && (
      (!$regex)
      || ( $regex !~ m{(?:\@redhat[\\]+\.com\$|^\!\@)$}
        && $regex ne $bu_regex)
    )
    && ((!$approval_group_name) || $approval_group_name eq '')
    )
  {
    ThrowUserError('approvals_missing_approval_group', {group => $group});
  }

  $gwf->set_approval_group($approval_group_name);
  $gwf->set_cc_list($cc_list);
  $gwf->update();

  $dbh->bz_commit_transaction();

  return ("Group " . $group->name . " updated");
}

sub get_groups {
  my ($self, $args) = validate(@_);

  my $user = Bugzilla->login(LOGIN_REQUIRED);
  $user->can_admin_group_workflow();

  my $cat = $args->{category}
    // ThrowCodeError('param_required', {param => 'category'});

  my $groups = Bugzilla::Group->match({category => $cat});

  my @rows;

  foreach my $group (@$groups) {
    my $gwf = $group->group_workflow;

    my %row = (name => $group->name,);

    if ($gwf) {
      $row{managed}        = $gwf->managed;
      $row{invite_only}    = $gwf->invite_only;
      $row{regex}          = $gwf->regex;
      $row{ccs}            = $gwf->cc_list;
      $row{approval_group} = $gwf->approval_group ? $gwf->approval_group->name : '';
    }

    push(@rows, \%row);
  }

  return (\@rows);

}

sub invite_manager {
  my ($self, $args) = validate(@_);

  my $user = Bugzilla->login(LOGIN_REQUIRED);

  my $manager = $args->{manager}
    // ThrowCodeError('param_required', {param => 'manager'});

  Bugzilla::validate_redhat_addr($manager . '@redhat.com')
    || ThrowUserError('invalid_manager_uid', {'uid' => $manager});

  Bugzilla->request_cache->{invite_manager} = 1;
  $user->check_and_send_account_creation_confirmation($manager . '@redhat.com');
  delete Bugzilla->request_cache->{invite_manager};

  return ("An account invitation has been mailed to $manager\@redhat.com");
}

1;

__END__

=head1 NAME

Bugzilla::Extension::Workflows::WebService::GroupRequest - API for handling
Adminsitration actions for the L<Bugzilla::Extension::Workflows> extension.

=head1 DESCRIPTION

This part of the Bugzilla API allows you to manage how groups are exposed in the
request forms.

  {
    "method": 'GroupRequest.update_group',
    "params": {
      "group": "thegroup",
      "managed": 1,
      "regex": ".*@example.net$",
      "ccs": "theboss@example.net,anotherboss@example.net",
      "approval_group": "admin"
    }

  }

=head1 METHODS

See L<Bugzilla::WebService> for a description of what B<STABLE>, B<UNSTABLE>,
and B<EXPERIMENTAL> mean, and for more description about error codes.

=head2 C<update_group> B<EXPERIMENTAL>

=head3 Description

Update a group's approvals settings.

=head3 Params

group:
  B<string> The name of the group being modified.
managed:
  B<boolean> Should this group be exposed on the request page?
invite_only:
  B<boolean> This group cannot be directly applied for by users, instead a user has to be invited by someone in the approval group.
regex:
  B<string> The regular expression to match logins on.
approval_group:
  B<string> name of the group to use for approvals.
ccs:
  B<string> A comma separated list of users to CC when this group is requested.

=head2 C<get_groups> B<EXPERIMENTAL>

=head3 Description

Get the approvals settings for a category of groups.

=head3 Params

category:
  B<string> The name of the group category to fetch.

=head3 Returns

An B<array> of hashes, one for each group in the category, containing the
following.

name:
  B<string> The name of the group being modified.
managed:
  B<boolean> Should this group be exposed on the request page?
invite_only:
  B<boolean> This group cannot be directly applied for by users, instead a user has to be invited by someone in the approval group.
regex:
  B<string> The regular expression to match logins on.
approval_group:
  B<string> name of the group to use for approvals.
ccs:
  B<string> A comma separated list of users to CC when this group is requested.

=cut
